(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: compile_move.ml,v 1.11 2004/10/20 15:55:46 julia Exp $ *)

(* among previous_needed, and process_stmt code duplication
is possible.  To fix... *)

module B = Objects
module CS = Class_state
module CSel = Compile_select

let in_handler = ref false

let error ln str =
  Error.update_error();
  Printf.printf "%s: %s \n" (B.loc2c ln) str

(* -------------------------- Useful syntax ------------------------- *)

let state_names states =
  List.map (function CS.STATE(id,_,_,_) -> id) states

let switch_error states attr =
  match Aux.subtract (CS.all_states()) states with
    [] -> []
  | rest ->
      let rest_names = state_names rest in
      [Ast.SEQ_CASE(rest_names,rest,
		    Ast.ERROR(Printf.sprintf
				"unexpected process state: %s"
				(String.concat " or "
				   (List.map B.id2c rest_names)),
			      attr),
		    attr)]

let make_state_cst id =
  B.mkId(Printf.sprintf "%s_STATE" (String.uppercase (B.id2c id)))

let make_class_cst id =
  let cls =
    let rec lookup x = function
	[] -> raise (Error.Error "internal error: no class for state")
      | ((y,CS.STATE(_,cls,_,_))::r) -> if x = y then cls else lookup x r in
    lookup id (!CS.state_env) in
  B.mkId(CS.class2c cls)

(* ---------------- Reorganization of move statements --------------- *)

type source_type =
    PROC of CS.state_info | KEY of CS.state_info |
    QUEUE of CS.state_info list | NOWHERE of CS.state_info list

let classify_sources key_sorted states aspects dst :
    (source_type *
       ((Ast.vardecl * Ast.stmt * (B.identifier * B.attribute) list) list *
	  (Ast.vardecl * Ast.stmt * (B.identifier * B.attribute) list) list))
    list =
  let (procs,key,queues) =
    List.fold_left
      (function (procs,key,queues) ->
	function
	  CS.STATE(_,_,CS.PROC,_) as x -> (x::procs,key,queues)
	| CS.STATE(_,_,CS.QUEUE(CS.SELECT(_)),_) as x ->
	    if key_sorted then (procs,x::key,queues) else (procs,key,x::queues)
	| CS.STATE(_,_,CS.QUEUE(_),_) as x -> (procs,key,x::queues))
      ([],[],[]) states in
  let (queues,nowhere) =
    List.partition
      (function
	  CS.STATE(_,CS.NOWHERE,CS.QUEUE(_),_) -> false
	| _ -> true)
      queues in
  let rec get_aspect_loop = function
      [] -> ([],[])
    | src::srcs ->
	let (a,nona) = get_aspect_loop srcs in
	(try
	  (((src,List.assoc (src,dst) aspects) :: a),nona)
	with Not_found -> (a,src::nona)) in
  let (procsa,procsnona) = get_aspect_loop procs in
  let (keya,keynona) = get_aspect_loop key in
  let (queuesa,queuesnona) = get_aspect_loop queues in
  let (nowherea,nowherenona) = get_aspect_loop nowhere in
  (List.map (function (p,a) -> (PROC p,a)) procsa) @
  (List.map (function p -> (PROC p,([],[]))) procsnona) @
  (List.map (function (p,a) -> (KEY p,a)) keya) @
  (List.map (function p -> (KEY p,([],[]))) keynona) @
  (List.map (function (p,a) -> (QUEUE [p],a)) queuesa) @
  (if queuesnona = [] then [] else [(QUEUE queuesnona, ([],[]))]) @
  (List.map (function (p,a) -> (NOWHERE [p],a)) nowherea) @ 
  (if nowherenona = [] then [] else [(NOWHERE nowherenona, ([],[]))])

let classify_dst dynamic_sort state override =
  match (state,override) with
    (CS.STATE(_,_,CS.PROC,_),_) -> "proc"
  | (CS.STATE(_,_,CS.QUEUE(CS.FIFO),_),_) -> "queue"
  | (CS.STATE(_,_,CS.QUEUE(CS.SELECT(_)),_),Ast.HEAD) ->
      (match dynamic_sort with
	CSel.SORTED -> "sorted_lifo"
      |	CSel.UNSORTED -> "lifo"
      |	CSel.PARTIALLY_SORTED -> "partially_sorted_lifo")
  | (CS.STATE(_,_,CS.QUEUE(CS.SELECT(_)),_),Ast.TAIL) ->
      (match dynamic_sort with
	CSel.SORTED -> "sorted_fifo"
      |	CSel.UNSORTED -> "fifo"
      |	CSel.PARTIALLY_SORTED -> "partially_sorted_fifo")
  | (CS.STATE(_,_,CS.QUEUE(CS.SELECT(CS.NODEFAULT)),_),_) ->
      (match dynamic_sort with
	CSel.SORTED -> "sorted"
      |	CSel.UNSORTED -> "queue"
      |	CSel.PARTIALLY_SORTED -> "partially_sorted_lifo") (* lifo is cheaper *)
  | (CS.STATE(_,_,CS.QUEUE(CS.SELECT(CS.FIFODEFAULT)),_),_) ->
      (match dynamic_sort with
	CSel.SORTED -> "sorted_fifo"
      |	CSel.UNSORTED -> "fifo"
      |	CSel.PARTIALLY_SORTED -> "partially_sorted_fifo")
  | (CS.STATE(_,_,CS.QUEUE(CS.SELECT(CS.LIFODEFAULT)),_),_) ->
      (match dynamic_sort with
	CSel.SORTED -> "sorted_lifo"
      |	CSel.UNSORTED -> "lifo"
      |	CSel.PARTIALLY_SORTED -> "partially_sorted_lifo")
  | (CS.STATE(_,_,CS.QUEUE(CS.EMPTY),_),_) -> "nowhere"

let states2str states = Aux.set2c(List.map B.id2c (state_names states))

(* aspects can't refer to local variables of the context and there are no
process-typed global variables, so there is no danger that an aspect will 
modify the source expression *)
let make_aspect_list decls srcexp aspects =
  List.map
    (function (param,stm,fvs) ->
      List.iter
	(function (id,ln) ->
	  if List.mem id decls
	  then error ln
	      (Printf.sprintf
		 "identifier %s captured by a declaration in some enclosing definition" (B.id2c id)))
	fvs;
      (Ast.VALDEF(param,srcexp,Ast.VARIABLE,Ast.get_exp_attr srcexp), stm))
    aspects

let regroup (decl,stm) = Ast.SEQ([decl],[stm],Ast.get_stm_attr stm)

let src2dst (srcinfo,aspects) decls srcexp dstid dfname dstexp auto_allowed
    attr before state_upd fromprocname =
  let (sfname,srcexp,sstates,srcid,srckey) =
    (match srcinfo with
      PROC(CS.STATE(id,_,_,_) as x) ->
	("proc", Ast.VAR(id,attr), [x], Some id, false)
    | KEY(CS.STATE(id,_,_,_) as x) ->
	("queue", srcexp, [x], Some id, true)
    | QUEUE(sstates) ->
	let srcid =
	  match sstates with
	    [CS.STATE(id,_,_,_)] -> Some id
	  | _ -> None in
	("queue", srcexp, sstates, srcid, false)
    | NOWHERE(sstates) -> ("nowhere", srcexp, [], None, false)) in
  let mk_fname str = Ast.VAR(B.mkId str, B.updty attr B.VOID) in
  if not auto_allowed && Some(dstid) = srcid
  then
    match aspects with
      ([],[]) -> None
    | (before,after) ->
	Some (Ast.mkBLOCK attr
		((List.map regroup (make_aspect_list decls srcexp before))@
		 (List.map regroup (make_aspect_list decls srcexp after))))
  else
    let need_resched =
      if !B.schedtype = B.VIRTSCHED
      then []
      else
      match srcinfo with
	PROC(CS.STATE(_,CS.RUNNING,_,_)) ->
	  let int_attr = B.updty attr B.INT in
	  let nd = Ast.FIELD(Ast.VAR(B.mkId "current",B.updty attr B.VOID),
			     B.mkId "need_resched",int_attr) in
	  [Ast.ASSIGN(nd,Ast.INT(1,int_attr),false,attr)]
      | _ -> [] in
    let args = [Ast.add_comment (states2str sstates) srcexp;dstexp] in
    let args =
      if (!B.hls)
      then
	let sched_inst_type = B.SCHEDULER_INST in
	(Ast.VAR(B.mkId "scheduler_instance",B.updty attr sched_inst_type))
	:: args
      else args in
    let stmt1 =
      Ast.PRIMSTMT(mk_fname(Printf.sprintf "move_%s_%s" sfname dfname),
		   args,attr) in
    let stmt2 =
      if srckey
      then
	let src =
	  match srcid with
	    Some(id) -> Ast.VAR(id,attr)
	  | None -> failwith "internal error" in
	[Ast.PRIMSTMT(mk_fname "dec_bitmap",
		      [src;Ast.PRIM(mk_fname "key_index",[srcexp],
				    B.updty attr B.INT)],
		      attr)]
      else [] in
    let (before_aspects,after_aspects) = aspects in
    let before_aspects =
      List.map regroup (make_aspect_list decls srcexp before_aspects) in
    let after_aspects = 
      let after_aspects =
	(* first aspect goes outermost (see call to loop below) so have to
	   reverse *)
	List.rev(make_aspect_list decls srcexp after_aspects) in
      if fromprocname
      then
	let rec loop prev body = function
	    [] -> body
	  | (Ast.VALDEF(param,srcexp,v,a),stmt)::rest ->
	      let srcexp =
		match prev with
		  None -> srcexp
		| Some x -> Ast.VAR(List.hd (Ast.params2ids [x]),
				    Ast.get_exp_attr srcexp) in
	       Ast.SEQ([Ast.VALDEF(param,srcexp,v,a)],
		       [(loop (Some param) body rest);stmt],
		       Ast.get_stm_attr stmt)
	  | _ -> raise (Error.Error "internal error") in
	(function body -> loop None body after_aspects)
      else
	(function body ->
	  Ast.mkBLOCK attr (body::(List.map regroup after_aspects)))
    in
    if before
    then
      Some
	(Ast.mkBLOCK attr
	   (before_aspects@
	    [(after_aspects(Ast.mkBLOCK attr
			     (state_upd::stmt1::stmt2@need_resched)))]))
    else
      Some
	(Ast.mkBLOCK attr
	   (before_aspects@
	    [(after_aspects(Ast.mkBLOCK attr (stmt1::stmt2@need_resched)))]))

let mkifin (CS.STATE(id,_,_,_) as test_state)
    new_src_exp srcinfo1 srcinfo2 dstid dfname dstexp auto_allowed attr
    before state_upd decls =
  let after = if before then [] else [state_upd] in
  let in_test = Ast.IN(new_src_exp,id,[test_state],B.BOTH,false,attr) in
  let thn =
    src2dst srcinfo1 decls new_src_exp dstid dfname dstexp auto_allowed attr
      before state_upd false in
  let els =
    src2dst srcinfo2 decls new_src_exp dstid dfname dstexp auto_allowed attr
      before state_upd false in
  match (thn,els) with
    (None,None) -> []
  | (None,Some els) ->
      [Ast.IF(in_test,Ast.mkBLOCK attr [],Ast.postBLOCK attr els after,attr)]
  | (Some thn,None) ->
      [Ast.IF(in_test,Ast.postBLOCK attr thn after,Ast.mkBLOCK attr [],attr)]
  | (Some thn,Some els) -> Ast.IF(in_test,thn,els,attr) :: after

let srcs2dst srcexp srcs dstexp state auto_allowed attr
    dst override aspects dynamic_sort key decls =
  let dfname = classify_dst dynamic_sort dst override in
  let CS.STATE(dstid,_,state_type,_) = dst in
  let before =
    match state_type with
      CS.QUEUE(CS.EMPTY) -> true
    | _ -> false in
  let auto_allowed =
    match state_type with
      CS.QUEUE(CS.SELECT(_)) -> auto_allowed
    | _ -> false in
  let mk_fname str = Ast.VAR(B.mkId str, B.updty attr B.VOID) in
  let state_upd =
    Ast.PRIMSTMT(mk_fname "update_state",
		 [srcexp;Ast.VAR(make_state_cst dstid, attr);
		  Ast.VAR(make_class_cst dstid,attr)],
		 attr) in
  match classify_sources key srcs aspects dst with
    (* this is the only one that might be a process variable *)
    [srcinfo] ->
      (match src2dst srcinfo decls srcexp dstid dfname dstexp
	  auto_allowed attr true state_upd true with
	None -> []
      |	Some s -> [s])
  | [(PROC(s1),_) as srcinfo1;srcinfo2]
  | [(KEY(s1),_) as srcinfo1;srcinfo2] -> (* use if where possible *)
      mkifin s1 srcexp srcinfo1 srcinfo2 dstid dfname dstexp
	auto_allowed attr before state_upd decls
  | [srcinfo1;(PROC(s2),_) as srcinfo2]
  | [srcinfo1;(KEY(s2),_) as srcinfo2] ->
      mkifin s2 srcexp srcinfo2 srcinfo1 dstid dfname dstexp
	auto_allowed attr before state_upd decls
  | [(QUEUE([s1]),_) as srcinfo1;srcinfo2] ->
      mkifin s1 srcexp srcinfo1 srcinfo2 dstid dfname dstexp
	auto_allowed attr before state_upd decls
  | [srcinfo1; (QUEUE([s2]),_) as srcinfo2] ->
      mkifin s2 srcexp srcinfo2 srcinfo1 dstid dfname dstexp
	auto_allowed attr before state_upd decls
  | srcinfos -> (* more than two possibilities, so use switch *)
      let branches =
	List.map
	  (function (srcinfo,aspects) ->
	    let sstates =
	      (match srcinfo with
		PROC(state) | KEY(state) -> [state]
	      | QUEUE(sstates) -> sstates
	      | NOWHERE(sstates) -> sstates) in
	    Ast.SEQ_CASE(state_names sstates,sstates,
			 (match src2dst (srcinfo,aspects) decls srcexp
			     dstid dfname dstexp
			     auto_allowed attr before state_upd false with
			   None -> Ast.mkBLOCK attr []
			 | Some s -> s),
			 attr))
	  srcinfos in
      [Ast.SWITCH(srcexp,branches @ switch_error srcs attr,None,attr)]@
      (if before then [] else [state_upd])

(* ------------------------- Previous variable ---------------------- *)

let save_previous previd procid attr =
    Ast.ASSIGN(Ast.VAR(previd,attr),Ast.VAR(procid,attr),false,attr)

let rec proc_lookup id = function
    [] -> raise (Error.Error "internal error")
  | Ast.PROCESS(_,pid,_,_,_) as x :: xs ->
      if B.ideq(id,pid) then x else proc_lookup id xs
  | Ast.QUEUE(_,_,qid,_,_) as x :: xs ->
      if B.ideq(id,qid) then x else proc_lookup id xs

let previous_needed states process srcs attr =
  let prevprocs_others =
    List.fold_left
      (function (prevprocs,others) ->
	function
	    CS.STATE(id,_,CS.PROC,_) as st ->
	      (match proc_lookup id states with
		Ast.PROCESS(_,_,Ast.BOSSA_PREV(previd),_,_) ->
		  ((st,previd)::prevprocs,others)
	      |	_ -> (prevprocs,st::others))
	  | st -> (prevprocs,st::others))
      ([],[]) srcs in
  match prevprocs_others with
    ([],_) -> []
  | ([(CS.STATE(procid,_,_,_),previd)],[]) ->
      [save_previous previd procid attr]
  | ([(CS.STATE(procid,_,_,_) as st,previd)],x::xs) ->
      [Ast.IF(Ast.IN(process,procid,[st],B.BOTH,false,attr),
	      save_previous previd procid attr,
	      Ast.SEQ([],[],attr),
	      attr)]
  | (prevprocs,others) ->
      let (procs,_) = List.split prevprocs
      and otherids = List.map (function CS.STATE(id,_,_,_) -> id) others in
      [Ast.SWITCH(process,
		  List.map
		    (function (CS.STATE(procid,_,_,_) as procstate,previd) ->
		      Ast.SEQ_CASE([procid],[procstate],
				   save_previous previd procid attr,
				   attr))
		    prevprocs
		  @ [Ast.SEQ_CASE(otherids,others,Ast.SEQ([],[],attr),attr)]
		  @ switch_error (procs@others) attr,
		  None,
		  attr)]

(* ---------------- if dest is RUNNING update running --------------- *)

let running_dst attr = function
    CS.STATE(id,CS.RUNNING,CS.PROC,_) ->
      B.sched_select
	[Ast.ASSIGN(Ast.VAR(B.mkId "running",
			    B.updty attr (B.STRUCT (B.mkId "bossa_struct"))),
		    Ast.VAR(id,attr),false,
		    attr)]
	[]
  | _ -> []

(* --- convert the destination of move to an array ref, if needed --- *)

let check_for_key dstexp srcexp attr = function
    (true,CS.STATE(_,_,CS.QUEUE(CS.SELECT(_)),_)) ->
      let new_index_id = B.fresh_id() in
      let new_id = B.fresh_id() in
      let ptype = B.STRUCT (B.mkId "list_head") in
      let mk_fname str = Ast.VAR(B.mkId str, B.updty attr B.VOID) in
      let mk_id str = Ast.VAR(str, B.updty attr B.INT) in
      ([Ast.VALDEF(Ast.VARDECL(B.INT,new_index_id,false,attr),
		   Ast.PRIM(mk_fname "key_index", [srcexp], attr),
		   Ast.VARIABLE,attr);
	 Ast.VALDEF(Ast.VARDECL(ptype,new_id,false,attr),
		    Ast.AREF(dstexp,mk_id new_index_id,attr),
		    Ast.VARIABLE,attr)],
       [Ast.PRIMSTMT(mk_fname "inc_bitmap",[dstexp;mk_id new_index_id],attr)],
       Ast.VAR(new_id,attr))
  | _ -> ([],[],dstexp)

(* --------------------------- Processing --------------------------- *)

let rec process_stmt aspects decls dynamic_sort key states = function
    Ast.IF(exp,stm1,stm2,attr) ->
      Ast.IF(exp,
	     process_stmt aspects decls dynamic_sort key states stm1,
	     process_stmt aspects decls dynamic_sort key states stm2,
	     attr)

  | Ast.FOR(id,stid,l,dir,stm,attr) ->
      Ast.FOR(id,stid,l,dir,
	      process_stmt aspects (id::decls) dynamic_sort key states stm,
	      attr)

  | Ast.FOR_WRAPPER(label,stms,attr) ->
      Ast.FOR_WRAPPER(label,
		      List.map
			(process_stmt aspects decls dynamic_sort key states)
			stms,
		      attr)

  | Ast.SWITCH(exp,cases,default,attr) ->
      Ast.SWITCH(exp,
		 List.map
		   (function Ast.SEQ_CASE(pat,l,stm,attr) ->
		     Ast.SEQ_CASE(pat,l,
				  process_stmt aspects decls dynamic_sort key
				    states stm,
				  attr))
		   cases,
		 Aux.app_option
		   (process_stmt aspects decls dynamic_sort key states)
		   default,
		 attr)

  | Ast.SEQ(ldecls,stmts,attr) ->
      let new_ids =
	List.map
	  (function
	      Ast.VALDEF(Ast.VARDECL(ty,id,sys,ln),exp,const,a) -> id
	    | _ -> raise (Error.Error "unexpected local decl"))
	  ldecls in
      Ast.SEQ(ldecls,
	      List.map
		(process_stmt aspects (new_ids @ decls) dynamic_sort
		   key states)
		stmts,
	      attr)

  | Ast.MOVE(exp,state,_,srcs,Some(dst),auto_allowed,state_end,attr) ->
      let (decls1,new_src_exp) = Ast.name_expr exp attr in
      let (decls2,stmt2,new_dst_exp) =
	check_for_key (Ast.VAR(state,attr)) new_src_exp attr (key,dst) in
      let save_old = previous_needed states new_src_exp srcs attr in
      let move =
	srcs2dst new_src_exp srcs new_dst_exp state auto_allowed attr
	  dst state_end aspects dynamic_sort key decls in
      let post_move = running_dst attr dst in
      Ast.mkDBLOCK attr (decls1 @ decls2, stmt2 @ save_old @ move @ post_move)

  | stmt -> stmt
    
(* ----------------------------- Aspects ---------------------------- *)

let preprocess_aspects aspects =
  List.map
    (function (transition,(bef,aft)) ->
      let process l =
	List.map
	  (function 
	      Ast.ASPECT(srcs,src_infos,dsts,dst_infos,param,pos,stm,fvs,a) ->
		(param,stm,fvs))
	  l in
      (transition,(process bef,process aft)))
    aspects

(* --------------------- Handlers and functions --------------------- *)

let process_handlers aspects dynamic_sort key states handlers =
  in_handler := true;
  List.map
    (function (Ast.EVENT(nm,specd,(Ast.VARDECL(_,id,_,_) as param),
			 stmt,attr)) ->
      Ast.EVENT(nm,specd,param,
		process_stmt aspects [id] dynamic_sort key states stmt,
		attr))
    handlers

let process_ifunctions aspects dynamic_sort key states ifunctions =
  in_handler := false;
  List.map
    (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) ->
      let init_decls =
	List.map (function Ast.VARDECL(_,id,_,_) -> id) params in
      Ast.FUNDEF(ret,nm,params,
		 process_stmt aspects init_decls dynamic_sort key states stmt,
		 inl,attr))
    ifunctions

let compile_move
    (dynamic_sort,crit_info,
     (Ast.SCHEDULER(nm,default,cstdefs,enums,procdecls,fundecls,valdefs,states,
		    criteria,admit,trace,handlers,ifunctions,functions,
		      aspects,attr))) =
  let key = Aux.member CSel.KEYCRIT crit_info in
  let aspect_env =
    match aspects with
      Ast.ASPECTINFO(aspects) -> preprocess_aspects aspects
    | _ -> failwith "compile_move: unexpected aspects" in
  let new_ast =
   Ast.SCHEDULER(nm,default,
		 cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		 admit,trace,
		 process_handlers aspect_env dynamic_sort key states handlers,
		 process_ifunctions aspect_env dynamic_sort key states
		 ifunctions,functions,aspects,attr) in
  (crit_info,Braces.braces new_ast)

