(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: compile_ordering.ml,v 1.8 2004/11/11 16:15:00 julia Exp $ *)

(* convert the ordering_criteria to the definition of a "greater" function *)

module B = Objects
module CS = Class_state
module CSel = Compile_select
open Format

(* -------------------- Miscellaneous operations -------------------- *)

let rec id_member x = function
    [] -> false
  | (y::r) -> B.ideq(x,y) || id_member x r

let mkif test tr fl attr =
  Ast.IF(test,Ast.SEQ([],tr,attr),Ast.SEQ([],fl,attr),attr)

(* ------------ Tests derived from the ordering criteria ------------ *)

let sysdecls = ref ([] : B.identifier list)

let collect_sys_fields procdecls =
  List.fold_left
    (function prev ->
      function
	  Ast.VARDECL(_,id,true,_) -> id :: prev
	| _ -> prev)
    [] procdecls

let compile_exp procfields vr exp =
  let rec loop = function
      Ast.VAR(id,attr) ->
	if id_member id (!sysdecls)
	then Ast.FIELD(vr,id,attr)
	else if id_member id procfields
	then Ast.FIELD(vr,id,attr)
	else Ast.VAR(id,attr)

    | Ast.FIELD(exp,fld,attr) -> Ast.FIELD(loop exp,fld,attr)

    | Ast.UNARY(uop,exp,attr) ->
	Ast.UNARY(uop,loop exp,attr)

    | Ast.BINARY(bop,exp1,exp2,attr) ->
	Ast.BINARY(bop, loop exp1, loop exp2, attr)

    | Ast.PBINARY(bop,exp1,states1,exp2,states2,attr) ->
	raise (Error.Error "compile_ordering: unexpected pbinary")

    | Ast.INDR(exp,attr) -> Ast.INDR(loop exp,attr)
	  
    | Ast.PRIM(f,args,attr) ->
	Ast.PRIM(loop f,List.map loop args,attr)

    | exp -> exp (* make no sense in a criteria or ok as is *)
  in loop exp

let order2op = function
    Ast.LOWEST -> Ast.LT
  | Ast.HIGHEST -> Ast.GT

let order2notop = function
    Ast.LOWEST -> Ast.GT
  | Ast.HIGHEST -> Ast.LT

let compile_test tst thn els p pv procfields attr =
  mkif (compile_exp procfields p tst)
    [Ast.ASSIGN(pv,compile_exp procfields p thn,false,attr)]
    [Ast.ASSIGN(pv,compile_exp procfields p els,false,attr)]
    attr

let compare_pvid ty op pv1id pv2id attr =
  Ast.BINARY(op,pv1id,pv2id,attr)

let compile_one p1 p2 pv1 pv2 procfields = function
    Ast.CRIT_ID(key,order,id,attr) ->
      let ty = B.ty attr in
      let pv1id = Ast.VAR(pv1, B.updty attr ty) in
      let pv2id = Ast.VAR(pv2, B.updty attr ty) in
      (ty,
       [Ast.ASSIGN(pv1id,Ast.FIELD(p1,id,attr),false,attr);
       Ast.ASSIGN(pv2id,Ast.FIELD(p2,id,attr),false,attr)],
       compare_pvid ty (order2op order) pv1id pv2id attr,
       compare_pvid ty (order2notop order) pv1id pv2id attr)
  | Ast.CRIT_TEST(key,order,tst,thn,els,attr) ->
      let ty = B.ty (Ast.get_exp_attr thn) in
      let pv1id = Ast.VAR(pv1, B.updty attr ty) in
      let pv2id = Ast.VAR(pv2, B.updty attr ty) in
      (ty,
       [compile_test tst thn els p1 pv1id procfields attr;
	 compile_test tst thn els p2 pv2id procfields attr],
       compare_pvid ty (order2op order) pv1id pv2id attr,
       compare_pvid ty (order2notop order) pv1id pv2id attr)

(* ------- Tests derived from the ordering of the sorted queue ------ *)

let check_sorted lifo_fifo =
  match (lifo_fifo,!CS.selected) with
    (false,_) -> None
  | (true,Some(CS.STATE(id,_,CS.QUEUE(CS.SELECT(default)),_))) ->
      (match default with
	CS.NODEFAULT -> None
      | _ -> Some(id,default))
  | _ -> raise (Error.Error "internal error")

let queue_order_test id default p1 p2 attr =
  let order =
    (match default with
      CS.FIFODEFAULT -> "first_fifo"
    | CS.LIFODEFAULT -> "first_lifo"
    | CS.NODEFAULT -> raise (Error.Error "internal error")) in
  let order = if (!CS.selected_size) > 1 then order ^"_ar" else order in
  let info =
    match !CS.selected with
      Some x -> x
    | None -> raise (Error.Error "internal error") in
  mkif
    (Ast.IN(p1,id,[info],B.BOTH,false,attr))
    [mkif
	(Ast.IN(p2,id,[info],B.BOTH,false,attr))
	[Ast.RETURN(Some(Ast.PRIM(Ast.VAR(B.mkId(order),B.updty attr B.VOID),
				  [Ast.VAR(id,attr);p1;p2],
				  attr)),
		    attr)]
	[Ast.RETURN(Some(Ast.BOOL(true,attr)),attr)]
	attr]
    [Ast.RETURN(Some(Ast.BOOL(false,attr)),attr)]
    attr

(* --------------------- Select an array element -------------------- *)

let key_index procdecls key_crit =
  let attr = B.mkAttr (-1) in
  let attr = B.updty attr B.INT in
  let pid = B.fresh_idx "p" in
  let p = Ast.VAR(pid, attr) in
  let tmpid = B.fresh_idx "tmp" in
  let tmp = Ast.VAR(tmpid, attr) in
  let accid = B.fresh_idx "acc" in
  let acc = Ast.VAR(accid, attr) in
  let ptype = B.sched_select B.PROCESS B.SCHEDULER in
  let procfields =
    List.map (function Ast.VARDECL(ty,id,imported,attr) -> id) procdecls in
  let reorder mx exp attr = function
      Ast.HIGHEST -> Ast.BINARY(Ast.MINUS,Ast.INT(mx,attr),exp,attr)
    | Ast.LOWEST -> exp in
  let addend exp = function
      0 -> exp
    | mn -> Ast.BINARY(Ast.MINUS,exp,Ast.INT(mn,attr),attr) in
  let mult_pattern exp mn = function
      0 -> Ast.ASSIGN(acc,addend exp mn,false,attr)
    | offset ->
	Ast.ASSIGN(acc,Ast.BINARY(Ast.PLUS,
				  Ast.BINARY(Ast.TIMES,acc,
					     Ast.INT(offset,attr),
					     attr),
				  addend exp mn,attr),
		   false,attr) in
  let (_,code) =
    List.fold_left
      (function (offset,code) ->
	function
	    Ast.CRIT_ID(Ast.KEY(mn,mx),order,id,attr) ->
	      (true,
	       (mult_pattern (reorder mx (Ast.FIELD(p,id,attr)) attr order) mn
		  (if offset then (mx-mn+1) else 0))
	       :: code)
	  | Ast.CRIT_TEST(Ast.KEY(mn,mx),order,tst,thn,els,attr) ->
	      (true,
	       [mult_pattern (reorder mx tmp attr order) mn
		   (if offset then (mx-mn+1) else 0);
		 compile_test tst thn els p tmp procfields attr] @
	       code)
	  | _ -> raise (Error.Error "internal error"))
      (false,[]) key_crit in
  Ast.FUNDEF(B.INT,B.mkId("key_index"),
	     [Ast.VARDECL(ptype,pid,false,attr)],
	     Ast.SEQ([Ast.VALDEF(Ast.VARDECL(B.INT,accid,false,attr),
				 Ast.INT(0,attr),
				 Ast.VARIABLE,attr)],
		     List.rev (Ast.RETURN(Some(acc),attr) :: code), attr),
	     Ast.INLINE,attr)

(* ------------------ Making the complete function ------------------ *)

let mkdecl nm ty attr =
  Ast.SYSDEF(Ast.VARDECL(ty,nm,false,attr),Ast.CONST,attr)

let rec make_ifs tr fls fail pv1 pv2 attr = function
    ([],None,_) -> Ast.SEQ([],[Ast.RETURN(Some(fail),attr)],attr)
  | ([],Some(stmt),_) -> Ast.SEQ([],[stmt],attr)
  | ([(ty,stmts,test,ntest)],None,B.BOOL) ->
      Ast.SEQ([mkdecl pv1 ty attr; mkdecl pv2 ty attr],
	      stmts @ [Ast.RETURN(Some(test),attr)],attr)
  | ((ty,stmts,test,ntest)::rest,queue_order_test_stmt,ty1) ->
      let new_stmts =
	stmts @
	[mkif test
	    [Ast.RETURN(Some(tr),attr)]
	    [mkif ntest [Ast.RETURN(Some(fls),attr)] [] attr]
	    attr] in
      (match
	make_ifs tr fls fail pv1 pv2 attr (rest,queue_order_test_stmt,ty1)
      with
	Ast.SEQ([],rest_stmts,rest_attr) ->
	  Ast.SEQ([],
		  Ast.SEQ([mkdecl pv1 ty attr; mkdecl pv2 ty attr],
			  new_stmts,attr)
		    :: rest_stmts,
		  attr)
      |	rest_stmts ->
	  Ast.SEQ([],
		  [Ast.SEQ([mkdecl pv1 ty attr; mkdecl pv2 ty attr],
			   new_stmts,attr);
		    rest_stmts],
		  attr))

type critbody = SORTED of B.identifier option |
                FALL_THROUGH of B.identifier | NORMAL

let compile_crit name procdecls criteria prevname ty tr fls fail lifo_fifo =
  let ptype = B.sched_select B.PROCESS B.SCHEDULER in
  let attr = B.mkAttr (-1) in
  let p1id = B.fresh_idx "p1" in
  let p1 = Ast.VAR(p1id, B.updty attr ptype) in
  let p2id = B.fresh_idx "p2" in
  let p2 = Ast.VAR(p2id, B.updty attr ptype) in
  let pv1 = B.fresh_idx "pv1" in
  let pv2 = B.fresh_idx "pv2" in
  let procfields =
    List.map (function Ast.VARDECL(ty,id,imported,attr) -> id) procdecls in
  let stmts_tests = List.map (compile_one p1 p2 pv1 pv2 procfields) criteria in
  let args =
    [Ast.VARDECL(ptype,p1id,false,B.updty attr ptype);
     Ast.VARDECL(ptype,p2id,false,B.updty attr ptype)] in
  let args =
    if (!B.hls)
    then
      let sched_inst_type = B.SCHEDULER_INST in
      (Ast.VARDECL(sched_inst_type,B.mkId "scheduler_instance",false,
		   B.updty attr sched_inst_type))
      :: args
    else args in
  let mktest queue_order_test_stmt =
    let body =
      make_ifs tr fls fail pv1 pv2 attr
	(stmts_tests, queue_order_test_stmt, ty) in
    Ast.FUNDEF(ty,B.mkId(name),args,body,Ast.INLINE,attr) in
  let si_args =
    let si_arg = Ast.VAR(B.mkId "scheduler_instance",
			 B.updty attr B.SCHEDULER_INST) in
    let args = [Ast.VAR(p1id,attr);Ast.VAR(p2id,attr)] in
    if (!B.hls) then si_arg :: args else args in
  match prevname with
    SORTED info ->
      (match (check_sorted lifo_fifo,info) with
	(None,Some previous_name) ->
	  (* no extra sorting and a previous function has already been defined.
	     just make a call to it, and hope for inlining *)
	  Ast.FUNDEF(B.BOOL,B.mkId(name),args,
		     Ast.SEQ([],
			     [Ast.RETURN(Some(Ast.PRIM(Ast.VAR(previous_name,
							       B.updty attr
								 B.VOID),
						       si_args,attr)),
					 attr)],
			     attr),
		     Ast.INLINE,attr)
      |	(None,None) -> mktest None
      |	(Some(id,default),_) ->
	  mktest (Some (queue_order_test id default p1 p2 attr)))
  | FALL_THROUGH previous_name ->
      mktest (Some (Ast.RETURN(Some(Ast.PRIM(Ast.VAR(previous_name,
						     B.updty attr B.VOID),
					     si_args,attr)),
			      attr)))
  | NORMAL -> mktest None

let dummy_function name =
  let ptype = B.sched_select B.PROCESS B.SCHEDULER in
  let attr = B.mkAttr (-1) in
  let p1id = B.fresh_idx "p1" in
  let p1 = Ast.VAR(p1id, B.updty attr ptype) in
  let p2id = B.fresh_idx "p2" in
  let p2 = Ast.VAR(p2id, B.updty attr ptype) in
  let args =
    [Ast.VARDECL(ptype,p1id,false,B.updty attr ptype);
     Ast.VARDECL(ptype,p2id,false,B.updty attr ptype)] in
  let args =
    if (!B.hls)
    then
      let sched_inst_type = B.SCHEDULER_INST in
      (Ast.VARDECL(sched_inst_type,B.mkId "scheduler_instance",false,
		   B.updty attr sched_inst_type))
      :: args
    else args in
  Ast.FUNDEF(B.BOOL,B.mkId(name),args,
	     Ast.SEQ([],[Ast.RETURN(Some(Ast.BOOL(false,attr)),attr)],attr),
	     Ast.INLINE,attr)

let compile_sd_crit crit_info procdecls criteria attr =
  let rec loop = function
      ([],[]) -> ([],[],[])
    | (CSel.DYNCRIT::_,criteria) -> ([],[],criteria)
    | (CSel.STATCRIT::rest_info,crit::rest_crit) ->
	let (key_crit,stat_crit,dyn_crit) = loop (rest_info,rest_crit) in
	(key_crit,crit::stat_crit,dyn_crit) 
    | (CSel.KEYCRIT::rest_info,crit::rest_crit) ->
	let (key_crit,stat_crit,dyn_crit) = loop (rest_info,rest_crit) in
	(crit::key_crit,stat_crit,dyn_crit) 
    | (_,_) -> raise (Error.Error "internal error") in
  let (key_crit,stat_crit,dyn_crit) = loop (crit_info,criteria) in
  [key_index procdecls key_crit;
    compile_crit "static_compare" procdecls stat_crit NORMAL B.INT
      (Ast.INT(1,attr)) (Ast.INT(-1,attr)) (Ast.INT(0,attr)) false;
    (* should call static_greater is key criteria run out *)
    compile_crit "static_select_compare" procdecls key_crit
      (FALL_THROUGH(B.mkId "static_compare")) B.INT
      (Ast.INT(1,attr)) (Ast.INT(-1,attr)) (Ast.INT(0,attr)) false;
    compile_crit "dynamic_select_greater" procdecls dyn_crit NORMAL B.BOOL
      (Ast.BOOL(true,attr)) (Ast.BOOL(false,attr)) (Ast.BOOL(false,attr))
      false;
    compile_crit "dynamic_greater" procdecls dyn_crit
      (SORTED(Some(B.mkId "dynamic_select_greater"))) B.BOOL
      (Ast.BOOL(true,attr)) (Ast.BOOL(false,attr)) (Ast.BOOL(false,attr))
      false]
  @
  if !Compile_select.used_lifo_fifo
  then
    (* not sure to understand NORMAL, SORTED and fallthrough, and why static
       and dynamic are different in this regard *)
    [compile_crit "static_compare_lifo_fifo" procdecls stat_crit NORMAL B.INT
	(Ast.INT(1,attr)) (Ast.INT(-1,attr)) (Ast.INT(0,attr)) true;
    (* should call static_greater is key criteria run out *)
      compile_crit "static_select_compare_lifo_fifo" procdecls key_crit
	(FALL_THROUGH(B.mkId "static_compare_lifo_fifo")) B.INT
	(Ast.INT(1,attr)) (Ast.INT(-1,attr)) (Ast.INT(0,attr)) true;
      compile_crit "dynamic_select_greater_lifo_fifo" procdecls dyn_crit
	NORMAL B.BOOL
	(Ast.BOOL(true,attr)) (Ast.BOOL(false,attr)) (Ast.BOOL(false,attr))
	true;
      compile_crit "dynamic_greater_lifo_fifo" procdecls dyn_crit
	(SORTED(Some(B.mkId "dynamic_select_greater_lifo_fifo"))) B.BOOL
	(Ast.BOOL(true,attr)) (Ast.BOOL(false,attr)) (Ast.BOOL(false,attr))
	true]
  else List.map dummy_function
      ["static_compare_lifo_fifo";"static_select_compare_lifo_fifo";
	"dynamic_select_greater_lifo_fifo"; "dynamic_greater_lifo_fifo"]

(* ------------------- the size of the key array -------------------- *)

let key_size criteria =
  List.fold_left
    (function sz ->
      function
	  Ast.CRIT_ID(Ast.KEY(mn,mx),order,id,attr) ->
	    sz * (mx - mn + 1)
	| Ast.CRIT_TEST(Ast.KEY(mn,mx),order,tst,thn,els,attr) ->
	    sz * (mx - mn + 1)
	| _ -> sz)
    1 criteria

(* -------------------------- entry point --------------------------- *)

let compile_ordering
    (crit_info,
     Ast.SCHEDULER(nm,default,
		   cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		   admit,trace,handlers,ifunctions,functions,aspects,attr)) =
  CS.selected_size := key_size criteria;
  sysdecls := collect_sys_fields procdecls;
  let greater =
    compile_crit "greater" procdecls criteria (SORTED None) B.BOOL
      (Ast.BOOL(true,attr)) (Ast.BOOL(false,attr)) (Ast.BOOL(false,attr))
      false in
  let greater_lifo_fifo =
    match check_sorted true with
      None -> dummy_function "greater_lifo_fifo"
    | Some (_,_) ->
	compile_crit "greater_lifo_fifo" procdecls criteria (SORTED None)
	  B.BOOL
	  (Ast.BOOL(true,attr)) (Ast.BOOL(false,attr)) (Ast.BOOL(false,attr))
	  true in
  Ast.SCHEDULER(nm,default,cstdefs,enums,procdecls,fundecls,valdefs,states,
		criteria,admit,trace,handlers,ifunctions,
		  greater :: greater_lifo_fifo ::
		  (compile_sd_crit crit_info procdecls criteria attr) @
		  functions,aspects,attr)
