(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: compile_next.ml,v 1.8 2004/11/11 16:15:00 julia Exp $ *)

(*
   - convert next expression to appropriate field access
*)

module B = Objects
module CS = Class_state
module VB = Vobjects

let event_param = ref(B.mkId "")

let rec compile_exp = function

    Ast.FIELD(exp,fld,attr) ->
      let (exp,expfn) = compile_exp exp in
      (Ast.FIELD(exp,fld,attr),expfn)

  | Ast.UNARY(uop,uexp,attr) ->
      let (uexp,expfn) = compile_exp uexp in
      (Ast.UNARY(uop,uexp,attr),expfn)

  | Ast.BINARY(bop,exp1,exp2,attr) ->
      let (exp1,exp1fn) = compile_exp exp1 in
      let (exp2,exp2fn) = compile_exp exp2 in
      (Ast.BINARY(bop,exp1,exp2,attr),function stm -> exp1fn(exp2fn stm))

  | Ast.PBINARY(bop,exp1,states1,exp2,states2,attr) ->
      let (exp1,exp1fn) = compile_exp exp1 in
      let (exp2,exp2fn) = compile_exp exp2 in
      (Ast.PBINARY(bop,exp1,states1,exp2,states2,attr),
       function stm -> exp1fn(exp2fn stm))

  | Ast.INDR(exp,attr) ->
      let (exp,expfn) = compile_exp exp in
      (Ast.INDR(exp,attr),expfn)

  | Ast.PRIM(f,args,attr) ->
      let (f,ffn) = compile_exp f in
      let (args,argsfn) =
	List.fold_left
	  (function (prevargs,prevfn) ->
	    function exp ->
	      let (exp,expfn) = compile_exp exp in
	      (exp::prevargs,function stm -> prevfn(expfn stm)))
	  ([], ffn) args in
      let args = List.rev args in
      (Ast.PRIM(f,args,attr),argsfn)

  | Ast.AREF(exp1,exp2,attr) ->
      let (exp1,exp1fn) = compile_exp exp1 in
      let (exp2,exp2fn) = compile_exp exp2 in
      (Ast.AREF(exp1,exp2,attr),function stm -> exp1fn(exp2fn stm))

  | Ast.IN(exp,id,states,vl,insrc,in_attr) ->
      let (exp,expfn) = compile_exp exp in
      (Ast.IN(exp,id,states,vl,insrc,in_attr),expfn)

  | Ast.MOVEFWDEXP(exp,state_end,attr) ->
      let (exp,expfn) = compile_exp exp in
      (Ast.MOVEFWDEXP(exp,state_end,attr),expfn)

  | Ast.SCHEDCHILD(exp,procs,attr) ->
      let (exp,expfn) = compile_exp exp in
      let mkref str =
	Ast.FIELD(Ast.VAR(B.mkId(str^"_next_list"),attr),B.mkId "sched",attr)
      in
      (match procs with
	[] -> raise (Error.Error "next: internal error []")
      | [VB.SRC] -> (mkref "src",expfn)
      | [VB.TGT] -> (mkref "tgt",expfn)
      | [x;y] as l ->
	  if (x = VB.SRC && y = VB.TGT) || (x = VB.TGT && y = VB.SRC)
	  then
	    let new_id = B.fresh_id () in
	    let new_attr = B.updty attr B.SCHEDULER in
	    let var_ref = Ast.VAR(new_id,new_attr) in
	    let var_def =
	      Ast.SYSDEF(Ast.VARDECL(B.SCHEDULER,new_id,false,new_attr),
			 Ast.VARIABLE,new_attr) in
	    let tgt_exp =
	      Ast.FIELD(Ast.VAR(!event_param,B.updty attr B.EVENT),
			B.mkId "target",B.updty attr B.PROCESS) in
	    let test_exp =
	      Ast.BINARY(Ast.EQ,exp,tgt_exp,B.updty attr B.BOOL) in
	    let if_stmt =
	      Ast.IF(test_exp,
		     Ast.ASSIGN(var_ref,mkref "tgt",false,attr),
		     Ast.ASSIGN(var_ref,mkref "src",false,attr),
		     attr) in
	    (var_ref,
	     function stm ->
	       Ast.mkDBLOCK attr ([var_def],[if_stmt;expfn stm]))
	  else raise (Error.Error (Printf.sprintf "next: internal error %s"
				     (Aux.set2c(List.map VB.vprocname2c l))))
      | l -> raise (Error.Error (Printf.sprintf "next: internal error %s"
				   (Aux.set2c(List.map VB.vprocname2c l)))))
	
  | exp -> (exp,function stm -> stm)
	
let compile_decl = function
    Ast.VALDEF(decl,exp,is_const,attr) ->
      let (exp,expfn) = compile_exp exp in
      (Ast.VALDEF(decl,exp,is_const,attr),expfn)
  | decl -> (decl,function stm -> stm)
	
let rec compile_stmt = function
    Ast.IF(exp,stmt1,stmt2,attr) ->
      let (exp,expfn) = compile_exp exp in
      expfn (Ast.IF(exp,compile_stmt stmt1,compile_stmt stmt2,attr))
	
  | Ast.FOR(id,state_id,states,dir,stmt,attr) ->
      Ast.FOR(id,state_id,states,dir,compile_stmt stmt,attr)
	
  | Ast.FOR_WRAPPER(label,stmts,attr) ->
      Ast.FOR_WRAPPER(label,List.map compile_stmt stmts,attr)
	
  | Ast.SWITCH(exp,cases,default,attr) ->
      let (exp,expfn) = compile_exp exp in
      expfn (Ast.SWITCH(exp,List.map compile_case cases,
			Aux.app_option compile_stmt default,attr))
	
  | Ast.SEQ(decls,stmts,attr) ->
      let (decls,declsfn) =
	List.fold_left
	  (function (prevdecls,prevfn) ->
	    function decl ->
	      let (decl,declfn) = compile_decl decl in
	      (decl::prevdecls,function stm -> prevfn(declfn stm)))
	  ([], function stm -> stm) decls in
      let decls = List.rev decls in
      declsfn(Ast.SEQ(decls,List.map compile_stmt stmts,attr))
	
  | Ast.RETURN(None,attr) as x -> x
	
  | Ast.RETURN(Some(exp),attr) ->
      let (exp,expfn) = compile_exp exp in
      expfn(Ast.RETURN(Some(exp),attr))
	
  | Ast.MOVE(exp,state,x,y,z,a,b,attr) ->
      let (exp,expfn) = compile_exp exp in
      expfn(Ast.MOVE(exp,state,x,y,z,a,b,attr))
	
  | Ast.MOVEFWD(exp,x,state_end,attr) ->
      let (exp,expfn) = compile_exp exp in
      expfn(Ast.MOVEFWD(exp,x,state_end,attr))
	
  | Ast.SAFEMOVEFWD(exp,x,stm,state_end,attr) ->
      raise
	(Error.Error
	   "compile_next: unexpected safe move forward statement\n")
	
  | Ast.ASSIGN(expl,expr,sorted_fld,attr) ->
      let (expl,explfn) = compile_exp expl in
      let (expr,exprfn) = compile_exp expr in
      explfn(exprfn(Ast.ASSIGN(expl,expr,sorted_fld,attr)))
	
  | Ast.PRIMSTMT(f,args,attr) ->
      let (f,ffn) = compile_exp f in
      let (args,argsfn) =
	List.fold_left
	  (function (prevargs,prevfn) ->
	    function exp ->
	      let (exp,expfn) = compile_exp exp in
	      (exp::prevargs,function stm -> prevfn(expfn stm)))
	  ([], ffn) args in
      let args = List.rev args in
      argsfn(Ast.PRIMSTMT(f,args,attr))

  | Ast.ASSERT(exp,attr) ->
      let (exp,expfn) = compile_exp exp in
      expfn(Ast.ASSERT(exp,attr))

  | x -> x

and compile_case (Ast.SEQ_CASE(idlist,l,stmt,attr)) =
  Ast.SEQ_CASE(idlist,l,compile_stmt stmt,attr)


(* --------------------- Handlers and functions --------------------- *)

let process_handlers handlers =
  List.map
    (function
	(Ast.EVENT(nm,specd,(Ast.VARDECL(B.EVENT,id,imported,vattr) as param),
		   stmt,attr)) ->
		     event_param := id;
		     Ast.EVENT(nm,specd,param,compile_stmt stmt,attr)
      |	_ -> raise (Error.Error "internal error"))
    handlers

let compile_schedchild
    (Ast.SCHEDULER(nm,default,cstdefs,enums,procdecls,fundecls,valdefs,states,
		   criteria,admit,trace,handlers,ifunctions,functions,aspects,
		     attr)) =
  (Ast.SCHEDULER(nm,default,
		 cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		 admit,trace,
		 process_handlers handlers,
		 ifunctions,functions,aspects,attr))
