(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

module B = Objects
module VB = Vobjects
module CS = Class_state

type 'a fa_state = EDGE of int * 'a * int | FINAL of int
type entry = CS.classname * VB.vtype * VB.modified

(* -------------------- Miscellaneous operations -------------------- *)

let error str = raise (Error.Error str)

(* -------------------- Convert the NFA to a DFA -------------------- *)

let rec update_edge_targets edge_targets nm tgt =
  match edge_targets with
    [] -> [(nm,[tgt])]
  | ((name,l) as x)::rest ->
      if nm = name
      then (name,Aux.union [tgt] l)::rest
      else x :: (update_edge_targets rest nm tgt)

let collect_all_edges n edges =
  List.filter (function (EDGE(src,_,tgt)) -> src = n | _ -> false) edges

let rec nfa_to_dfa_loop edges all_states final all_final next_index = function
    [] -> ([],all_final)
  | ((cur,index)::rest) ->
      let edge_targets =
	List.fold_left
	  (function edge_targets ->
	    function n ->
	      let n_edges = collect_all_edges n edges in
	      List.fold_left
		(function edge_targets ->
		  function
		      (EDGE(src,nm,tgt)) ->
			update_edge_targets edge_targets nm tgt
		    | _ -> edge_targets)
		edge_targets n_edges)
	  [] cur in
      let edge_targets =
	List.map (function (nm,l) -> (nm,List.sort compare l))
	  edge_targets in
      let (new_edges,new_states,all_states,all_final,next_index) =
	List.fold_left
	  (function (new_edges,new_states,all_states,all_final,next_index) ->
	    function (nm,tgts) ->
	      try
		let dest = List.assoc tgts all_states in
		(Aux.union [EDGE(index,nm,dest)] new_edges,
		 new_states,all_states,all_final,next_index)
	      with Not_found ->
		(Aux.union [EDGE(index,nm,next_index)] new_edges,
		 (tgts,next_index)::new_states,
		 (tgts,next_index)::all_states,
		 (if (List.exists (function n -> Aux.member n final) tgts)
		 then next_index::all_final
		 else all_final),
		 next_index+1))
	  ([],[],all_states,all_final,next_index) edge_targets in
      let (newer_edges,all_final) =
	nfa_to_dfa_loop edges all_states final all_final next_index
	  (new_states @ rest) in
      (new_edges @ newer_edges, all_final)

let rec nfa_to_dfa edges start_states =
  let (real_edges,final_states) =
    List.partition
      (function EDGE(_,_,_) -> true | FINAL(_) -> false)
      edges in
  let final = List.map (function (FINAL(n)) -> n | _ -> -1) final_states in
  let all_final =
    match Aux.intersect start_states final with [] -> [] | _ -> [0] in
  nfa_to_dfa_loop real_edges [(start_states,0)] final all_final
    1 [(start_states,0)]
