(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: genaux.ml,v 1.5 2004/07/11 11:41:49 julia Exp $ *)

module B = Objects
module CS = Class_state

open Format
open Ioctl

let policy_name = ref ""
let external_functions = ref ([] : B.identifier list)

let init_external_functions _ =
  external_functions :=
    List.map B.mkId
      ["create_bossa_timer";"create_static_bossa_timer";
	"create_persistent_bossa_timer";
	"delete_bossa_timer";"delete_static_bossa_timer";
	"start_absolute_timer";"start_relative_timer";"stop_timer";"now"]

(* ----------------------- Environment lookup ----------------------- *)

exception LookupErrException

let rec lookup_id x = function
    [] -> raise LookupErrException
  | ((y,v)::r) -> if B.ideq(x,y) then v else lookup_id x r

(* ----------------------- Instance variables ----------------------- *)
(* For HLS *)
(* should be different from all local variables *)

let instance_variables = ref ([] : B.identifier list)

(* -------------------- Miscellaneous operations -------------------- *)

let print_nl_string str = print_string str; force_newline()

let rec print_between between_fn elem_fn = function
    [] -> ()
  | [x] -> elem_fn x
  | x :: xs -> (elem_fn x; between_fn(); print_between between_fn elem_fn xs)

let is_struct ar id =
  try
    (match lookup_id id (!CS.state_env) with
      CS.STATE(state_id,_,CS.QUEUE(CS.SELECT(_)),_) ->
	ar || ((!CS.selected_size) = 1)
    | CS.STATE(state_id,_,CS.QUEUE(_),_) -> true
    | _ -> false)
  with LookupErrException -> false

let is_state id =
  try let _ = lookup_id id (!CS.state_env) in true
  with LookupErrException -> false

let pp_id id =
  let nm = B.id2c id in
  if ((is_struct false id) && not (!B.csharp)) then print_string "&";
  if is_state id
  then
    if (!B.hls)
    then print_string "scheduler_instance->"
    else print_string "state_info.";
  if (nm="NULL") && (!B.csharp)
  then print_string "null"
  else
    (if List.mem id (!instance_variables)
    then print_string "scheduler_instance->";
     print_string nm)

let pp_id_cs id =
  if ((is_struct false id) && not (!B.csharp)) then
    if is_state id then
      if (!B.hls)
      then ("&" ^ "scheduler_instance->" ^ B.id2c(id))
      else ("&" ^ "state_info." ^ B.id2c(id))
    else ("&" ^ B.id2c(id))
  else
    if is_state id then
      if (!B.hls)
      then ("scheduler_instance->" ^ B.id2c(id))
      else ("state_info." ^ B.id2c(id))
    else B.id2c(id)

let pp_ar_id id =
  if (is_struct true id) then print_string "&";
  if is_state id then    (if (!B.hls)
    then print_string "scheduler_instance->"
    else print_string "state_info.");
  print_string(B.id2c(id))

(* & not wanted, eg for structure fields *)
let pp_struct_id id = print_string(B.id2c(id))
(* & not wanted, eg for field reference *)
let pp_struct_id2 id =
  if is_state id
  then
    (if (!B.hls)
    then print_string "scheduler_instance->"
    else print_string "state_info.");
  print_string(B.id2c(id))

let pp_comma _ = (print_string ","; print_space())
let pp_comma_nl _ = (print_string ","; force_newline())

let pp_semic _ = (print_string ";"; force_newline())

let nl_ifnonempty = function
    x::xs -> force_newline()
  | _ -> ()

let pp_arrayref_int id index =
  pp_ar_id id; print_string "["; print_int index; print_string "]"

let pp_arrayref_id id index =
  pp_ar_id id; print_string "["; pp_id index; print_string "]"

let start_block _ =
  print_string "{"; force_newline(); print_string "  "; open_box 0

let end_block _ =
  close_box(); force_newline (); print_string "}"

let de_structify str =
  if !B.csharp
  then
    let len = String.length str in
    if len > 7 && String.sub str 0 7 = "struct "
    then String.sub str 7 (len-7)
    else str
  else str

let pp_arglist long l =
  print_string "("; open_box 0;
  print_between (if long then pp_comma_nl else pp_comma)
    (function str -> print_string (de_structify str)) l;
  close_box(); print_string ")"

let pp_for_header id ptr maxfn =
  print_string "for ("; open_box 0; pp_id(id); 
  print_string " = "; maxfn(); print_string ".next; ";
  pp_id(id); print_string " != &"; maxfn(); print_string ";"; print_space();
  pp_id(id); print_string " = "; pp_id(ptr);
  print_string ") "; close_box()

let pp_for_header_deref id ptr maxfn =
  print_string "for ("; open_box 0; pp_id(id); 
  print_string " = "; maxfn(); print_string "->next; ";
  pp_id(id); print_string " != "; maxfn(); print_string ";"; print_space();
  pp_id(id); print_string " = "; pp_id(ptr);
  print_string ") "; close_box()

(* convert one name into another *)
let pp_state_cst id =
  print_string (Printf.sprintf "%s_STATE" (String.uppercase (B.id2c id)))

let check_simulate simfn modfn =
  print_string "\n#ifdef __KERNEL__"; force_newline();
  modfn();
  print_string "\n#else"; force_newline();
  simfn();
  print_string "\n#endif"; force_newline()

let process_state pre_array process_array_element process_queue process_var
    process_prev_var = function
    Ast.QUEUE(clsname,CS.EMPTY,id,vis,attr) -> () (* shouldn't be needed *)
  | Ast.QUEUE(clsname,CS.SELECT(_),id,vis,attr) ->
      let sz = !CS.selected_size in
      if sz > 1
      then
	(start_block();
	 let new_id = B.fresh_idx "i" in
	 print_string "int "; pp_id new_id; pp_semic();
	 pre_array id;
	 print_string "for ("; pp_id new_id; print_string "=0; ";
	 pp_id new_id; print_string " != "; print_int sz;
	 print_string "; "; pp_id new_id; print_string "++) ";
	 start_block(); 
	 process_array_element id new_id;
	 end_block();
	 end_block())
      else process_queue id
  | Ast.QUEUE(clsname,queue_typ,id,vis,attr) ->
      process_queue id
  | Ast.PROCESS(clsname,id,previd,vis,attr) ->
      process_var id;
      match previd with
	Ast.NO_PREV -> ()
      | Ast.BOSSA_PREV id -> force_newline(); process_prev_var id
      | Ast.RTS_PREV id -> ()

(* --------------------- Information about types -------------------- *)

let pp_bossa_struct _ =
  if (!B.csharp)
  then print_string "___"
  else print_string "struct ";
  print_string (!policy_name);
  print_string "_bossa_struct"

let pp_bossa_struct_cs _ =
  if (!B.csharp)
  then ("___" ^ (!policy_name) ^ "_bossa_struct")
  else ("struct " ^ (!policy_name) ^ "_bossa_struct")

let pp_scheduler_struct _ =
  if (!B.csharp)
  then print_string "___"
  else print_string "struct ";
  print_string (!policy_name);
  print_string "_sched_inst_struct"

let pp_scheduler_struct_cs _ =
  if (!B.csharp)
  then ("___" ^ (!policy_name) ^ "_sched_inst_struct")
  else ("struct " ^ (!policy_name) ^ "_sched_inst_struct")


let pp_type = function
    B.PROCESS ->
      (match !B.schedtype with
	B.PROCSCHED -> pp_bossa_struct(); print_string " *"
      |	B.VIRTSCHED -> print_string "struct bossa_struct *")
  | B.SRCPROCESS ->
      (match !B.schedtype with
	B.PROCSCHED -> pp_bossa_struct(); print_string " *"
      |	B.VIRTSCHED -> print_string "struct bossa_struct *")
  | B.SCHEDULER ->
      B.check_virt "scheduler type"; pp_bossa_struct(); print_string " *"
  | B.SCHEDULER_INST ->
      pp_scheduler_struct(); print_string " *"
  | B.RANGE(id) -> pp_id id
  | ty -> print_string (B.type2ctype ty)


let pp_type_cs = function
    B.PROCESS ->
      (match !B.schedtype with
	B.PROCSCHED -> (pp_bossa_struct_cs() ^ " *")
      |	B.VIRTSCHED -> "struct bossa_struct *")
  | B.SRCPROCESS ->
      (match !B.schedtype with
	B.PROCSCHED -> (pp_bossa_struct_cs() ^ " *")
      |	B.VIRTSCHED -> "struct bossa_struct *")
  | B.SCHEDULER ->
      (pp_bossa_struct_cs() ^ " *")
  | B.SCHEDULER_INST ->
      (pp_scheduler_struct_cs() ^ " *")
  | B.RANGE(id) -> pp_id_cs id
  | ty -> (B.type2ctype ty)


let pp_etype = function (* for external functions - policy type not known *)
    B.PROCESS ->
      if (!B.csharp) then (pp_bossa_struct(); print_string " *")
      else print_string "PROTOCOL_PROCESS_STRUCT *"
  | B.SRCPROCESS ->
      if (!B.csharp) then (pp_bossa_struct(); print_string " *")
      else print_string "PROTOCOL_PROCESS_STRUCT *"
  | B.SCHEDULER ->
      B.check_virt "scheduler type";
      if (!B.csharp) then (pp_bossa_struct(); print_string " *")
      else print_string "PROTOCOL_PROCESS_STRUCT *"
  | B.SCHEDULER_INST ->
      if (!B.csharp) then (pp_scheduler_struct(); print_string " *")
      else print_string "SCHEDULER_INSTANCE_STRUCT *"
  | B.RANGE(_) -> print_string "int"
  | ty -> print_string (B.type2ctype ty)

(* --------------------- C/Csharp compatibility --------------------- *)

let mk_public str =
  if (!B.csharp)
  then print_string "public ";
  print_string (de_structify str)

let mk_inline _ =
  if not (!B.csharp)
  then print_string "inline "

let mk_static _ =
  if (!B.csharp)
  then print_string "static private unsafe "
  else print_string "static "

let mk_static_private _ =
  if (!B.csharp)
  then print_string "static private "
  else print_string "static "

let mk_empty_prototype _ =
  if (!B.csharp)
  then print_string "()"
  else print_string "(void)"

let fix_slashes s =
  let count = ref 0 in
  String.iter (function c -> if c = '\\' then count := !count + 1) s;
  let res = String.create ((String.length s) + (!count)) in
  count := 0;
  String.iter
    (function c ->
      if c = '\\' then (String.set res (!count) '\\'; count := !count + 1);
      String.set res (!count) c;
      count := !count + 1)
    s;
  res

let mk_printk_string str =
  if (!B.csharp)
  then fix_slashes str
  else str

let rec div_string l s =
  if ((String.length s) > l) then
    let first = String.sub s 0 l in
    let last_char = String.get first (l-1) in
    if last_char = '\\'
    then
      (* surely there is a better way to do the following *)
      let tmp = String.create 1 in
      String.set tmp 0 (String.get s l);
      (first ^ tmp) ::
      (div_string l (String.sub s (l+1) ((String.length s)-(l+1))))
    else
      first ::
      (div_string l (String.sub s l ((String.length s)-l)))
  else [s]

let mk_printk str =
  if (!B.csharp)
  then
    List.iter
      (function str -> print_nl_string ("printk(\""^(fix_slashes str)^"\");"))
      (div_string 50 str)
  else print_nl_string ("printk(\""^str^"\");")

let mk_attr p =
  if (!B.csharp)
  then
    "((task_struct *)((char *)("^p^
    ")-(ulong)((int)(&((task_struct *)0)->bossa)/2)))"
  else "attr("^p^")"
