{
(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: module_bossa_lexer.mll,v 1.5 2004/11/11 16:15:00 julia Exp $ *)
open Module_bossa_parser
open Format
open Lexing

type error =
  | Illegal_character of char
  | Unterminated_comment

(* Error defines error types and positions*)
exception Error of error * int * int

(* To store the position of the beginning of a comment and bitpat *)
let comment_start_pos = ref 0

(* Report Error  *)

let report_error fmt = function
  | Illegal_character c  -> fprintf fmt "Illegal character (%s)" (Char.escaped c)
  | Unterminated_comment -> fprintf fmt "Comment not terminated"

let scan_escape str =
	match str with
	"n" -> "\n"
	| "r" -> "\r"
	| "t" -> "\t"
	| "b" -> "\b"
	| _ -> str
let get_value chr =
	match chr with
	'0'..'9' -> (Char.code chr) - (Char.code '0')
	| 'a'..'z' -> (Char.code chr) - (Char.code 'a') + 10
	| 'A'..'Z' -> (Char.code chr) - (Char.code 'A') + 10
	| _ -> 0
let scan_hex_escape str =
	String.make 1 (Char.chr (
		(get_value (String.get str 0)) * 16
		+ (get_value (String.get str 1))
	))
let scan_oct_escape str =
	String.make 1 (Char.chr (
		(get_value (String.get str 0)) * 64
		+ (get_value (String.get str 1)) * 8
		+ (get_value (String.get str 2))
	))
}

(* define rules *)

let id = ['a'-'z' 'A'-'Z'] ['a'-'z' 'A'-'Z' '0'-'9' '_']*
let decimal_literal = ['0'-'9']+

(* for strings *)
let escape = '\\' _
let octdigit = ['0'-'7']
let hexdigit = ['0'-'9' 'a'-'f' 'A'-'F']
let hex_escape = '\\' ['x' 'X'] hexdigit hexdigit
let oct_escape = '\\' octdigit  octdigit octdigit
let blank = [' ' '\t' '\n']

rule token = parse
	|	'#'		{line lexbuf}
    	|   	"&&"            { AND	        }
    	|   	"||"            { OR            }
   	|   	"=>"            { RARROW        }
    	|   	"="             { EQL   	}
    	|   	">="            { GEQ   	}
    	|   	"<="            { LEQ    	}
    	|   	"=="            { EQUAL_EQUAL   }
    	|   	"!="            { NEQ  		}
	|	"++"		{ PLUS_PLUS	}
	|	"--"		{ MINUS_MINUS	}
	|	"+="		{ PLUS_EQUAL	}
	|	"-="		{ MINUS_EQUAL	}
	|	"*="		{ TIMES_EQUAL	}
	|	"/="		{ DIV_EQUAL	}
	|	"%="		{ MOD_EQUAL	}
	|	"&="		{ BITAND_EQUAL	}
	|	"|="		{ BITOR_EQUAL	}
	|	"<<="		{ LSHIFT_EQUAL	}
	|	">>="		{ RSHIFT_EQUAL	}
	|	">>"		{ RSHIFT	}
	|	"<<"		{ LSHIFT	}
	|	".."		{ DOT_DOT	}
	|	"~"		{ COMPLEMENT	}
	|	"?"		{ QUESTION	}
	|	"%"		{ MOD		}
   	|   	'&'             { BITAND       	}
     	|   	"."             { DOT           }
    	|   	'('             { LPAREN        }
    	|   	')'             { RPAREN        }
    	|   	'['             { LBRACK        }
    	|   	']'             { RBRACK        }
    	|   	'{'             { LCURLY        }
    	|   	'}'             { RCURLY        }
	|	','		{ COMMA		}
    	|   	';'             { SEMICOLON     }
    	|   	':'             { COLON         }
    	|   	'+'             { PLUS          }
    	|   	'-'             { MINUS         }
    	|   	'*'             { TIMES         }
    	|   	'/'             { DIV           }
    	|   	'|'             { BITOR         }
    	|   	'>'             { GT            }
    	|   	'<'             { LT            }
    	|   	'!'             { NOT           }
    	|   eof                 { EOF           }
	|	'"'		{ STRING (str lexbuf) }
    	|   [' ' '\t' ]
        	{ token lexbuf }    (* skip whitespace      *)
    	|   '\n' | '\r' | "\r\n"
		{ 
	  	Module_misc.lineno := !Module_misc.lineno+1;
          	Module_misc.cr_index := (lexeme_start lexbuf) + 1;
	  	Module_misc.incr lexbuf;
          	token lexbuf  (* skip whitespace *)
		}    
    	|   "//" [^'\n' '\r']*       
		{ token lexbuf }    (* skip // comment      *)
    	|   "/*"
        	{ 
          	comment_start_pos := Lexing.lexeme_start lexbuf;
          	comment lexbuf; 
	  	token lexbuf 
		}
    	| decimal_literal
		{ INTEGER(int_of_string(lexeme lexbuf)) }
    	| id                 
		{let str = Lexing.lexeme lexbuf in
		try
                  Hashtbl.find Module_keywords.keyword_table str
                with Not_found ->
		  (match Module_progtypes.scan_ident str with
		    Module_progtypes.IDENT -> IDENT str
		  | Module_progtypes.ENUM_TYPE -> ENUM_TYPE str
		  | Module_progtypes.RANGE_TYPE -> RANGE_TYPE str
		  | Module_progtypes.FN_NAME -> FN_NAME str
		  | Module_progtypes.STATE_NAME -> STATE_NAME str
                  | Module_progtypes.MODULE_NAME -> MODULE_NAME str)}
    	|   _
        { 
	  raise (Error(Illegal_character ((Lexing.lexeme lexbuf).[0]),
		       Lexing.lexeme_start lexbuf,
		       Lexing.lexeme_end lexbuf))
	}

and comment = parse
    | "*/"                  
	{ () } 
    | '\n' | '\r' | "\r\n"
        { 
	  Module_misc.lineno := !Module_misc.lineno+1; 
          Module_misc.cr_index := (lexeme_start lexbuf) + 1;
	  Module_misc.incr lexbuf;
	  comment lexbuf 
	}
    | eof
        { 
          raise (Error (Unterminated_comment, 
			!comment_start_pos,
			!comment_start_pos + 2))
        }
    | _  
        { comment lexbuf }

and str =
  parse	'"' {""}
    | hex_escape
	{let cur = scan_hex_escape (String.sub (Lexing.lexeme lexbuf) 2 2) in
	cur ^ (str lexbuf)}
    | oct_escape
	{let cur = scan_oct_escape (String.sub (Lexing.lexeme lexbuf) 1 3) in
	cur ^ (str lexbuf)}
    | "\\0"
	{(String.make 1 (Char.chr 0)) ^ (str lexbuf)}
    | escape
	{let cur = scan_escape (String.sub (Lexing.lexeme lexbuf) 1 1) in
	cur ^ (str lexbuf)}
    | _	{let cur = Lexing.lexeme lexbuf in cur ^  (str lexbuf)}
(* # <line number> <file name> ... *)
and line =
	parse	'\n'		    {token lexbuf}
	|	blank		    {line lexbuf}
	| decimal_literal           {Module_misc.lineno :=
				       int_of_string (Lexing.lexeme lexbuf);
				      Module_misc.cr_index :=
					(lexeme_start lexbuf) + 1;
				      Module_misc.incr lexbuf;
				     endline lexbuf}
	|	_		    {endline lexbuf}
and endline =
	parse '\n' 		    {token lexbuf}
	|	_		    {endline lexbuf}



{
let init _ = Module_progtypes.init_lexicon ()
}
