(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: objects.ml,v 1.17 2004/10/20 15:55:47 julia Exp $ *)

(* ------------- Process scheduler or virtual scheduler ------------- *)
let csharp = ref false
let hls = ref false
let blast = ref false
let high_res = ref false
let policy_blocked_types = ref false

let init_ioctls = ref ""

let type_return = ref ""

type sched = PROCSCHED | VIRTSCHED
type sched_module = MODULE | PROCSCHED2

let schedtype = ref PROCSCHED
let schedmodule = ref PROCSCHED2

let check_module str =
  match !schedmodule with
    PROCSCHED2 -> ()
  | MODULE -> raise (Error.Error
		       (Printf.sprintf "%s only in a scheduler" str))

let check_virt str =
  match !schedtype with
    PROCSCHED -> raise (Error.Error
			(Printf.sprintf "%s only in a virtual scheduler" str))
  | VIRTSCHED -> ()

let check_proc str =
  match !schedtype with
    VIRTSCHED -> raise (Error.Error
			(Printf.sprintf "%s only in a process scheduler" str))
  | PROCSCHED -> ()

let sched_select proc virt =
  match !schedtype with
    VIRTSCHED -> virt
  | PROCSCHED -> proc

let check_blocked str =
  if !policy_blocked_types
  then check_virt str

let check_policy_blocked str =
  if not (!policy_blocked_types)
  then
    raise
      (Error.Error (Printf.sprintf "check_policy_blocked: unexpected %s" str))
  else check_proc str

(* --------------- Choice of blocked or kernel_blocked -------------- *)

type ps_using_blocked = B_UNKNOWN | IS_USING_BLOCKED | IS_USING_KERNEL_BLOCKED
let ps_using_blocked = ref B_UNKNOWN

let is_using_blocked _ =
  match !schedtype with
    PROCSCHED ->
      if !ps_using_blocked = IS_USING_KERNEL_BLOCKED
      then raise (Error.Error "cannot mix BLOCKED and other variants in a VS")
      else ps_using_blocked := IS_USING_BLOCKED
  | VIRTSCHED -> ()

let is_using_kernel_blocked _ =
  match (!policy_blocked_types,!schedtype) with
    (true,PROCSCHED) ->
      if !ps_using_blocked = IS_USING_BLOCKED
      then raise (Error.Error "cannot mix BLOCKED and other variants in a VS")
      else ps_using_blocked := IS_USING_KERNEL_BLOCKED
  | (false,PROCSCHED) -> raise (Error.Error "only BLOCKED supported")
  | (_,VIRTSCHED) -> ()

(* --------------------------- Identifiers -------------------------- *)

(* string for printing, int for comparison *)
type identifier = ID of string * int

let symtab = ref (Hashtbl.create(500) : (string, identifier) Hashtbl.t)
let index = ref 0
let fresh_ctr = ref 0

let init_id () =
  symtab := Hashtbl.create(500);
  index := 0;
  fresh_ctr := 0

let mkId str =
  try Hashtbl.find (!symtab) str
  with Not_found ->
    (let n = !index in
    let entry = ID(str,n) in
    (index := n + 1;
     Hashtbl.add (!symtab) str entry;
     entry))

(*let ideq (ID(_,m),ID(_,n)) = m = n*)
let ideq (ID(_,m),ID(_,n)) = m = n
let strideq (m,ID(n,_)) = m = n

let id2c (ID(str,n)) = str
let id2n (ID(_,n)) = n

let nopid = mkId("")

let rec fresh_id _ =
  let cur = !fresh_ctr in
  (fresh_ctr := cur + 1;
   let str = Printf.sprintf "_fresh_%d" cur in
   try (let _ = Hashtbl.find (!symtab) str in fresh_id())
   with Not_found -> mkId str) (* a bit inefficient... *)

let rec fresh_idx xstr =
  let cur = !fresh_ctr in
  (fresh_ctr := cur + 1;
   let str = Printf.sprintf "%s_%d" xstr cur in
   try (let _ = Hashtbl.find (!symtab) str in fresh_idx xstr)
   with Not_found -> mkId str) (* a bit inefficient... *)

(* ------------------------------ Types ----------------------------- *)

(* SRCPROCESS is the type of the source variable.  Its fields can be
   accessed, but it cannot be use in next, moved, etc *)

type typ =
  ENUM of identifier | RANGE of identifier | SYSTEM of typ |
  INT | BOOL | TIME | CYCLES | PROCESS | SRCPROCESS | SCHEDULER | VOID |
  EVENT | INDR of typ | STRUCT of identifier | NULL | EMPTY | PORT |
  SCHEDULER_INST | TIMER of timer_persistence

and timer_persistence = NOT_PERSISTENT | PERSISTENT of identifier option

let rec type2c = function
    ENUM(id) -> if (!csharp) then (id2c id) else "enum "^(id2c id)
  | RANGE(id) -> "range "^(id2c id)
  | SYSTEM(typ) -> "system "^(type2c typ)
  | INT -> "int"
  | BOOL -> "bool"
  | TIME -> "time"
  | CYCLES -> "cycles"
  | PROCESS -> "process"
  | SRCPROCESS -> "process"
  | SCHEDULER -> "scheduler"
  | SCHEDULER_INST -> "scheduler_instance"
  | VOID -> "void"
  | EVENT -> "event"
  | INDR(t) -> (type2c t)^"*"
  | STRUCT(id) ->
      if (!csharp) then (id2c id)^" *" else "struct "^(id2c id)^" *"
  | NULL -> "null"
  | EMPTY -> "" (* hack for #defines *)
  | TIMER(_) -> "struct timer_list *"
  | PORT -> "port"

let rec type2ctype = function
  SYSTEM(typ) -> type2c typ
  | CYCLES -> "int"
  | TIME ->
      if !high_res
      then "struct timespec"
      else
	if (!csharp)
	then "int"
	else "unsigned long"
  | INDR(t) -> (type2ctype t)^"*"
  | typ -> type2c typ

(* ---------------------------- Attributes -------------------------- *)

type location = MODULE_NAME of identifier | SCHEDULER_NAME of identifier

let module_name = ref (None : location ref option)

type attribute = {modl : location ref option; ln : int; test_index : int;
		   ty : typ; com : string option}
let mkAttr n =
  {modl = !module_name; ln = n; test_index = 0; ty = VOID; com = None}
let mkTestAttr (n,ind) =
  {modl = !module_name; ln = n; test_index = ind; ty = VOID; com = None}
let modl attr = attr.modl
let line attr = attr.ln
let test_index attr = attr.test_index
let ty attr = attr.ty
let com attr = attr.com
let updty attr newty = {attr with ty = newty}
let updtst attr newtst = {attr with test_index = newtst}
let add_comment attr newcom = {attr with com = Some(newcom)}

let loc2c a =
  match a.modl with
    Some name ->
      (match !name with
	MODULE_NAME(m) ->
	  Printf.sprintf "module %s, line %d" (id2c m) a.ln
      | SCHEDULER_NAME(s) ->
	  Printf.sprintf "scheduler %s, line %d" (id2c s) a.ln)
  | None -> Printf.sprintf "line %d" a.ln

(* ----------------- Optimizing empty and in tests ------------------ *)

type testval = TRUE | FALSE | BOTH | UNK

let lubtestval = function
    (UNK,x) -> x
  | (x,UNK) -> x
  | (TRUE,TRUE) -> TRUE
  | (FALSE,FALSE) -> FALSE
  | _ -> BOTH

let testval2c = function
    TRUE -> "true"
  | FALSE -> "false"
  | BOTH -> "both"
  | UNK -> "unknown"

(* -------------------------- For events ---------------------------- *)

type intinfo = INTIBLE | UNINTIBLE

let mutate = ref false      (* generate mutations *)

type isquasi = QUASI | NOT_QUASI   (* detect quasi event types *)

(* -------------------------- Primitives ---------------------------- *)

let prim_names _ =
  (if !high_res
  then ["make_time";"make_cycle_time";"make_cycles";"time_to_seconds";
	 "time_to_nanoseconds";"time_to_jiffies";"time_to_subjiffies"]
  else []) @
  (if !schedtype = VIRTSCHED then ["next_exists"] else []) @
  ["print_trace_info";"kfree";"select";"empty";
   "now";"start_absolute_timer";"start_relative_timer";"stop_timer";
   "time_to_ticks";"ticks_to_time";"increment_ref_count"]

(* ------------------------ Free variables ------------------------- *)

let capture_check fvs decls error =
  List.iter
    (function (id,ln) ->
      if List.mem id decls
      then
	error
	  (Printf.sprintf
	     "identifier %s used at %s captured by a declaration in\nsome enclosing definition" (id2c id) (loc2c ln)))
    fvs
