(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* We just inline everything.  It would be nice to only inline things that
make state changes, but there is also the need to detect assignments that
can affect attributes relevant to the ordering criteria of processes that
are on the ready queue.  Since we assume that any function is likely to
affect process attributes, and it is not evident at the time the inlining
takes place what is the state of any process-typed arguments, we just
assume that everything must be inlined.  Recursion is not allowed.  This
phase happens after type checking and before verification, so that we can
check that the return type of a function corresponds to each of its return
values. *)

module B = Objects
module CS = Class_state

let error ln str =
  Error.update_error();
  Printf.printf "%s: %s \n" (B.loc2c ln) str

type fenv_type =
    (B.identifier *
       (Objects.typ * Ast.vardecl list * Ast.stmt *
	  (B.identifier * B.attribute) list (* free vars *))) list

(* ------------------ Functions returning a value ------------------- *)

(* Replace returns by assignments of return value to the provided variable.
This is only called when the function has a non-void type, and thus there
should always be a return value.  Note that pre_type.ml has ensured that
return appears only in tail position. *)

let rec fix_returns var_ref = function
    Ast.IF(tst,thn,els,a) ->
      Ast.IF(tst,fix_returns var_ref thn,fix_returns var_ref els,a)
  | Ast.SWITCH(exp,cases,dflt,a) ->
      let cases =
	List.map
	  (function Ast.SEQ_CASE(a,b,stm,c) ->
	    Ast.SEQ_CASE(a,b,fix_returns var_ref stm,c))
	  cases in
      let dflt =
	match dflt with
	  None -> None
	| Some stm -> Some (fix_returns var_ref stm) in
      Ast.SWITCH(exp,cases,dflt,a)
  | Ast.SEQ(decls,stms,a) ->
      let stms = List.map (fix_returns var_ref) stms in
      Ast.SEQ(decls,stms,a)
  | Ast.RETURN(Some exp,a) -> Ast.ASSIGN(var_ref,exp,false,a)
  | Ast.RETURN(None,a) ->
      raise
	(Error.Error (Printf.sprintf "%s: unexpected void return" (B.loc2c a)))
  | Ast.SAFEMOVEFWD(exp,x,stm,b,a) ->
      Ast.SAFEMOVEFWD(exp,x,fix_returns var_ref stm,b,a)
  | stm -> stm (* note: no return possible in for loop *)

and process_fn_with_value (fenv : fenv_type) decls inlined
    (ret_type,params,body,fvs) id args new_stms a =
  B.capture_check fvs decls (error a);
  let name = List.length new_stms in
  let retid = B.mkId(Printf.sprintf "%s_return_%d" (B.id2c id) name) in
  let var_ref = Ast.VAR(retid,a) in
  let ret_decl = Ast.UNINITDEF(Ast.VARDECL(ret_type,retid,false,a),a) in
  let processed_body =
    process_stm fenv decls (id::inlined) (fix_returns var_ref body) in
  let body =
    Ast.SEQ(List.map2
	      (function (Ast.VARDECL(ty,nm,imp,a)) as decl ->
		function arg -> Ast.VALDEF(decl,arg,Ast.VARIABLE,a))
	      params args,
	    [processed_body], B.updty a B.VOID)
  in (var_ref,(ret_decl,body)::new_stms)

(* ------------------ Functions returning no value ------------------ *)

and drop_returns = function
    Ast.IF(tst,thn,els,a) ->
      Some(Ast.IF(tst,drop_return_stmt thn,drop_return_stmt els,a))
  | Ast.SWITCH(exp,cases,dflt,a) ->
      let cases =
	List.map
	  (function Ast.SEQ_CASE(a,b,stm,c) ->
	    Ast.SEQ_CASE(a,b,drop_return_stmt stm,c))
	  cases in
      let dflt =
	match dflt with
	  None -> None
	| Some stm -> drop_returns stm in
      Some(Ast.SWITCH(exp,cases,dflt,a))
  | Ast.SEQ(decls,stms,a) as stm ->
      (match List.rev stms with
	[] -> Some stm
      |	(last::revrest) ->
	  Some
	    (match drop_returns last with
	      None -> Ast.SEQ(decls,List.rev revrest,a)
	    | Some s -> Ast.SEQ(decls,(List.rev revrest)@[s],a)))
  | Ast.RETURN(Some exp,a) ->
      raise
	(Error.Error (Printf.sprintf "%s: unexpected return value"
			(B.loc2c a)))
  | Ast.RETURN(None,a) -> None
  | stm -> Some stm (* note: no return possible in for loop *)

and drop_return_stmt stm =
  match drop_returns stm with
    None -> Ast.SEQ([],[],Ast.get_stm_attr stm)
  | Some s -> s

and process_fn_with_no_value (fenv : fenv_type) decls inlined
    (ret_type,params,body,fvs) id args new_stms a =
  B.capture_check fvs decls (error a);
  let processed_body =
    process_stm fenv (id::inlined) decls (drop_return_stmt body) in
  let body =
    Ast.SEQ(List.map2
	      (function (Ast.VARDECL(ty,nm,imp,a)) as decl ->
		function arg -> Ast.VALDEF(decl,arg,Ast.VARIABLE,a))
	      params args,
	    [processed_body], B.updty a B.VOID)
  in (body,new_stms)

(* -------------------------- Expressions --------------------------- *)

and process_exp (fenv : fenv_type) decls inlined new_stms = function
    Ast.FIELD(exp,fld,a) ->
      let (exp,new_stms) = process_exp fenv decls inlined new_stms exp in
      (Ast.FIELD(exp,fld,a),new_stms)
  | Ast.UNARY(op,exp,a) ->
      let (exp,new_stms) = process_exp fenv decls inlined new_stms exp in
      (Ast.UNARY(op,exp,a),new_stms)
  | Ast.BINARY(op,exp1,exp2,a) ->
      let (exp1,new_stms) = process_exp fenv decls inlined new_stms exp1 in
      let (exp2,new_stms) = process_exp fenv decls inlined new_stms exp2 in
      (Ast.BINARY(op,exp1,exp2,a),new_stms)
  | Ast.PBINARY(op,exp1,states1,exp2,states2,a) ->
      let (exp1,new_stms) = process_exp fenv decls inlined new_stms exp1 in
      let (exp2,new_stms) = process_exp fenv decls inlined new_stms exp2 in
      (Ast.PBINARY(op,exp1,states1,exp2,states2,a),new_stms)
  | Ast.INDR(exp,a) ->
      let (exp,new_stms) = process_exp fenv decls inlined new_stms exp in
      (Ast.INDR(exp,a),new_stms)
  | Ast.PRIM(Ast.VAR(id,va),args,a) ->
      begin
	if List.mem id inlined
	then
	  raise
	    (Error.Error
	       (Printf.sprintf "%s: recursion of %s detected"
		  (B.loc2c va) (B.id2c id)));
	let (args,new_stms) =
	  process_exp_list fenv decls inlined new_stms args in
	try process_fn_with_value fenv decls inlined (List.assoc id fenv)
	    id args new_stms a
	with Not_found -> (Ast.PRIM(Ast.VAR(id,va),args,a),new_stms)
      end
  | Ast.PRIM(fn,args,a) ->
      let (fn,new_stms) = process_exp fenv decls inlined new_stms fn in
      let (args,new_stms) =
	process_exp_list fenv decls inlined new_stms args in
      (Ast.PRIM(fn,args,a),new_stms)
  | Ast.SCHEDCHILD(exp,x,a) ->
      let (exp,new_stms) = process_exp fenv decls inlined new_stms exp in
      (Ast.SCHEDCHILD(exp,x,a),new_stms)
  | Ast.IN(exp,x,y,z,m,a) ->
      let (exp,new_stms) = process_exp fenv decls inlined new_stms exp in
      (Ast.IN(exp,x,y,z,m,a),new_stms)
  | Ast.MOVEFWDEXP(exp,a,b) ->
      let (exp,new_stms) = process_exp fenv decls inlined new_stms exp in
      (Ast.MOVEFWDEXP(exp,a,b),new_stms)
  | exp -> (exp,new_stms)

and process_exp_list (fenv : fenv_type) decls inlined new_stms exps =
  let (exps,new_stms) =
    List.fold_left
      (function (exps,new_stms) ->
	function exp ->
	  let (exp,new_stms) = process_exp fenv decls inlined new_stms exp in
	  (exp::exps,new_stms))
      ([],new_stms) exps in
  (List.rev exps,new_stms)

(* -------------------------- Declarations -------------------------- *)

and process_decl (fenv : fenv_type) decls inlined = function
    Ast.VALDEF(decl,exp,isconst,attr) ->
      let (exp,new_times) = process_exp fenv decls inlined [] exp in
      (Ast.VALDEF(decl,exp,isconst,attr),new_times)
  | decl -> (decl,[])

(* --------------------------- Statements --------------------------- *)

and process_new_stms new_stms a stm =
  match new_stms with
    [] -> stm
  | _ ->
      let (new_decls,new_stms) = List.split (List.rev new_stms) in
      Ast.SEQ(new_decls,new_stms@[stm],a)

and process_stm (fenv : fenv_type) decls inlined = function
    Ast.IF(tst,thn,els,a) ->
      let (tst,new_stms) = process_exp fenv decls inlined [] tst in
      let thn = process_stm fenv decls inlined thn in
      let els = process_stm fenv decls inlined els in
      process_new_stms new_stms a (Ast.IF(tst,thn,els,a))
  | Ast.FOR(x,y,z,d,stm,a) ->
      let stm = process_stm fenv decls inlined stm in
      Ast.FOR(x,y,z,d,stm,a)
  | Ast.FOR_WRAPPER(x,stms,a) ->
      Ast.FOR_WRAPPER(x,List.map (process_stm fenv decls inlined) stms,a)
  | Ast.SWITCH(exp,cases,dflt,a) ->
      let (exp,new_stms) = process_exp fenv decls inlined [] exp in
      let cases =
	List.map
	  (function Ast.SEQ_CASE(a,b,stm,c) ->
	    let stm = process_stm fenv decls inlined stm in
	    Ast.SEQ_CASE(a,b,stm,c))
	  cases in
      let dflt =
	match dflt with
	  None -> None
	| Some stm -> Some (process_stm fenv decls inlined stm) in
      process_new_stms new_stms a (Ast.SWITCH(exp,cases,dflt,a))
  | Ast.SEQ(ldecls,stms,a) ->
      let decls_newtimes = List.map (process_decl fenv decls inlined) ldecls in
      let new_decls = Ast.valdefs2ids ldecls in
      let stms =
	List.map
	  (function stm -> process_stm fenv (new_decls@decls) inlined stm)
	  stms in
      let rec decl_loop acc_decls = function
	  [] -> Ast.SEQ(List.rev acc_decls,stms,a)
	| ((decl,[])::rest) -> decl_loop (decl::acc_decls) rest
	| ((decl,new_stms)::rest) ->
	    let rest = decl_loop [] rest in
	    process_new_stms new_stms a
	      (match rest with
		Ast.SEQ(d,s,a1) -> Ast.SEQ((List.rev (decl::acc_decls))@d,s,a1)
	      |	_ -> (Ast.SEQ (List.rev (decl::acc_decls), [rest], a))) in
      decl_loop [] decls_newtimes
  | Ast.RETURN(Some exp,a) ->
      let (exp,new_stms) = process_exp fenv decls inlined [] exp in
      process_new_stms new_stms a (Ast.RETURN(Some exp,a))
  | Ast.MOVE(exp,x,y,z,r,s,b,a) ->
      let (exp,new_stms) = process_exp fenv decls inlined [] exp in
      process_new_stms new_stms a (Ast.MOVE(exp,x,y,z,r,s,b,a))
  | Ast.MOVEFWD(exp,x,b,a) ->
      let (exp,new_stms) = process_exp fenv decls inlined [] exp in
      process_new_stms new_stms a (Ast.MOVEFWD(exp,x,b,a))
  | Ast.SAFEMOVEFWD(exp,x,s,b,a) ->
      let (exp,new_stms) = process_exp fenv decls inlined [] exp in
      let s = process_stm fenv decls inlined s in
      process_new_stms new_stms a (Ast.SAFEMOVEFWD(exp,x,s,b,a))
  | Ast.ASSIGN(lexp,rexp,x,a) ->
      let (lexp,new_stms) = process_exp fenv decls inlined [] lexp in
      let (rexp,new_stms) = process_exp fenv decls inlined new_stms rexp in
      process_new_stms new_stms a (Ast.ASSIGN(lexp,rexp,x,a))
  | Ast.PRIMSTMT((Ast.VAR(id,va)) as fn,args,a) ->
      begin
	if List.mem id inlined
	then
	  raise
	    (Error.Error
	       (Printf.sprintf "%s: recursion of %s detected"
		  (B.loc2c va) (B.id2c id)));
	let (args,new_stms) = process_exp_list fenv decls inlined [] args in
	try
	  let (ret_type,params,body,fvs) as x = List.assoc id fenv in
	  match ret_type with
	    B.VOID ->
	      let (body,new_stms) =
		process_fn_with_no_value fenv decls inlined x id args
		  new_stms a in
	      process_new_stms new_stms a body
	  | _ ->
	      (match process_fn_with_value fenv decls inlined x id args
		  new_stms a
	      with
		(var_ref,(ret_decl,body)::new_stms) ->
		  Ast.SEQ([ret_decl],
			  [process_new_stms new_stms a body],
			  a)
	      | _ -> raise (Error.Error "inline: internal error"))
	with Not_found -> process_new_stms new_stms a (Ast.PRIMSTMT(fn,args,a))
      end
  | Ast.PRIMSTMT(fn,args,a) ->
      let (fn,new_stms) = process_exp fenv decls inlined [] fn in
      let (args,new_stms) =
	process_exp_list fenv decls inlined new_stms args in
      process_new_stms new_stms a (Ast.PRIMSTMT(fn,args,a))
  | stm -> stm

(* ------------------------------ Admit ----------------------------- *)

let process_admit fenv = function
    None -> None
  | Some(Ast.ADMIT(locals,crit_params,crits,attach,detach,attr)) ->
      (* locals are fresh so no capture problem *)
      let new_crits =
	List.map
	  (function crit ->
	    process_stm fenv (Ast.params2ids crit_params) [] crit)
	  crits in
      let new_attach_detach update =
	match update with
	  None -> None
	| Some (param,stm) ->
	    Some(param,process_stm fenv (Ast.params2ids [param]) [] stm) in
      Some(Ast.ADMIT(locals,crit_params,new_crits,
		     new_attach_detach attach,new_attach_detach detach,attr))

(* --------------------------- Entry point -------------------------- *)

let create_fn_env fns : fenv_type =
  List.map
    (function Ast.FUNDEF(ret_ty,nm,params,body,inl,a) ->
      let clsenv = Ast.make_class_env (!CS.state_env) in
      let init_env = Ast.params2ids params in
      let free_vars = Ast.collect_stm_vars init_env body in
      (nm,(ret_ty,params,body,free_vars)))
    fns

let inline
    (Ast.SCHEDULER(nm,default,
		   cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		   admit,trace,handlers,ifunctions,functions,aspects,attr)) =
  let fenv = create_fn_env functions in
  let handlers =
    List.map
      (function
	  Ast.EVENT(nm,specd,(Ast.VARDECL(_,id,_,_) as param),stm,attr) ->
	Ast.EVENT(nm,specd,param,process_stm fenv [id] [] stm,attr))
      handlers in
  let ifunctions =
    List.map
      (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) ->
	let init_decls = Ast.params2ids params in
	Ast.FUNDEF(ret,nm,params,process_stm fenv init_decls [] stmt,inl,attr))
      ifunctions in
  let aspects =
    match aspects with
      Ast.ASPECTINFO(aspects) ->
	let process l =
	  List.map
	  (function
	      Ast.ASPECT(srcs,src_infos,dsts,dst_infos,param,pos,stm,fvs,a) ->
		let init_decls = Ast.params2ids [param] in
		Ast.ASPECT(srcs,src_infos,dsts,dst_infos,param,pos,
			   process_stm fenv init_decls [] stm,fvs,a))
	  l in
	Ast.ASPECTINFO(
	List.map
	  (function (transition,(bef,aft)) ->
	    (transition,(process bef,process aft)))
	  aspects)
    | _ -> failwith "inline: unexpected aspects" in
  let admit = process_admit fenv admit in
  Ast.SCHEDULER(nm,default,
		cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		admit,trace,handlers,ifunctions,[],aspects,attr)
