(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: pp.ml,v 1.14 2004/11/11 16:15:01 julia Exp $ *)

module B = Objects
module CS = Class_state
open Format

(* -------------------- Miscellaneous operations -------------------- *)

let print_comments = ref false (* print with analysis comments? *)

let rec print_between between_fn elem_fn = function
    [] -> ()
  | [x] -> elem_fn x
  | x :: xs -> (elem_fn x; between_fn(); print_between between_fn elem_fn xs)

let pp_id id = print_string(B.id2c(id))

let pp_comma _ = (print_string ","; print_space())

let nl_ifnonempty = function
    x::xs -> force_newline()
  | _ -> ()

(* ------------------ Information about structures ------------------ *)

let pp_decl (Ast.VARDECL(ty,id,imported,attr)) =
  (print_string (B.type2c(ty)); print_string " "; pp_id(id))

let pp_procdecls procdecls =
  (print_string "process = {"; force_newline(); print_string "  "; open_box 0;
   print_between force_newline (function x -> pp_decl x; print_string ";")
     procdecls; close_box(); force_newline();
   print_string "}"; force_newline(); force_newline())

(* -------------------- Information about states -------------------- *)

let pp_vis = function
    CS.PUBLIC -> ()
  | CS.PRIVATE -> print_string "private "

let pp_states states =
  (print_string "states = {"; force_newline(); print_string "  "; open_box 0;
   print_between force_newline
     (function
	 Ast.QUEUE(CS.NOWHERE,_,id,_,attr) -> ()(* don't print nowhere state *)
       | Ast.QUEUE(clsname,CS.EMPTY,id,vis,attr) ->
	   (print_string (CS.safe_class2c(clsname)); print_space(); pp_id(id);
	    print_string ";")
       | Ast.QUEUE(clsname,queue_typ,id,vis,attr) ->
	   (print_string (CS.safe_class2c(clsname)); print_space(); pp_id(id);
	    print_space(); print_string ":"; print_space();
	    pp_vis vis;
	    (match queue_typ with
	    | CS.FIFO -> print_string "fifo "
	    | CS.SELECT(CS.NODEFAULT) -> print_string "select "
	    | CS.SELECT(CS.LIFODEFAULT) -> print_string "select lifo "
	    | CS.SELECT(CS.FIFODEFAULT) -> print_string "select fifo "
	    | _ -> raise (Error.Error("printing internal error")));
	    print_string "queue";
	    print_string ";")
       | Ast.PROCESS(clsname,id,previd,vis,attr) ->
	   (print_string (CS.safe_class2c(clsname)); print_space(); pp_id(id);
	    print_space(); print_string ":"; print_space();
	    pp_vis vis;
	    print_string "process";
	    (match previd with
	      Ast.NO_PREV -> ()
	    | Ast.BOSSA_PREV x ->
		pp_comma(); print_string "previous"; print_space(); pp_id(x)
	    | Ast.RTS_PREV x ->
		pp_comma(); print_string "previous"; print_space(); pp_id(x));
	    print_string ";"))
     states; close_box(); force_newline();
   print_string "}"; force_newline(); force_newline())

(* --------------------------- Expressions -------------------------- *)

let pp_id id = print_string (B.id2c(id))

let rec pp_ucexp = function
    Ast.INT(n,attr) -> print_int n

    | Ast.VAR(id,attr) -> pp_id id

    | Ast.FIELD(exp,fld,attr) as x -> (pp_exp exp; print_string "."; pp_id fld)

    | Ast.BOOL(true,attr) -> print_string "true"
    | Ast.BOOL(false,attr) -> print_string "false"

    | Ast.UNARY(Ast.NOT,exp,attr) ->
	(print_string "!"; print_paren exp)

    | Ast.UNARY(Ast.COMPLEMENT,exp,attr) ->
	(print_string "~"; print_paren exp)

    | Ast.BINARY(bop,exp1,exp2,attr)
    | Ast.PBINARY(bop,exp1,_,exp2,_,attr) ->
	(print_paren exp1; pp_bop bop; print_paren exp2)

    | Ast.INDR(exp,attr) ->
	(print_string "*"; print_paren exp)
	  
    | Ast.SELECT(_,attr) ->
	print_string "select()"

    | Ast.EMPTY(id,_,_,insrc,attr) when not(!print_comments) ->
	print_string "empty("; pp_id id; print_string ")"

    | Ast.EMPTY(id,_,B.TRUE,insrc,attr) ->
	(print_string "empty_true("; pp_id id; print_string ")";
	 print_string (Printf.sprintf "/* %s: insrc: %b */"
			 (B.loc2c attr) insrc))

    | Ast.EMPTY(id,_,B.FALSE,insrc,attr) ->
	(print_string "empty_false("; pp_id id; print_string ")";
	 print_string (Printf.sprintf "/* %s: insrc: %b */"
			 (B.loc2c attr) insrc))

    | Ast.EMPTY(id,_,B.BOTH,_,attr) ->
	(print_string "empty_both("; pp_id id; print_string ")")

    | Ast.EMPTY(id,_,B.UNK,insrc,attr) ->
	(print_string "empty_unknown("; pp_id id; print_string ")";
	 print_string (Printf.sprintf "/* %s: insrc: %b */"
			 (B.loc2c attr) insrc))

    | Ast.PRIM(f,args,attr) ->
	(print_paren f; print_string "(";
	 open_box 0; print_between pp_comma pp_exp args;
	 close_box(); print_string ")")

    | Ast.SCHEDCHILD(arg,procs,attr) ->
	(print_string "scheduler("; open_box 0; pp_exp arg; close_box();
	 print_string ")")

    | Ast.SRCONSCHED(attr) -> print_string "srcOnSched()"

    | Ast.ALIVE(_,_,attr) when not(!print_comments) ->
	print_string "tgtAlive()"

    | Ast.ALIVE(B.TRUE,_,attr) -> print_string "tgtAlive_true()"

    | Ast.ALIVE(B.FALSE,_,attr) -> print_string "tgtAlive_false()"

    | Ast.ALIVE(B.BOTH,_,attr) -> print_string "tgtAlive_both()"

    | Ast.ALIVE(B.UNK,_,attr) -> print_string "tgtAlive_unknown()"

    | Ast.AREF(exp1,exp2,_) ->
	(pp_exp exp1; print_string "["; pp_exp exp2; print_string "]")

    | Ast.IN(exp,id,_,_,_,attr) when not(!print_comments) ->
	(print_paren exp; print_string " in "; pp_id id)

    | Ast.IN(exp,id,_,B.TRUE,_,attr) ->
	(print_paren exp; print_string " in_true "; pp_id id)

    | Ast.IN(exp,id,_,B.FALSE,_,attr) ->
	(print_paren exp; print_string " in_false "; pp_id id)

    | Ast.IN(exp,id,_,B.BOTH,_,attr) ->
	(print_paren exp; print_string " in_both "; pp_id id)

    | Ast.IN(exp,id,_,B.UNK,_,attr) ->
	(print_paren exp; print_string " in_unk "; pp_id id)

    | Ast.MOVEFWDEXP(exp,state_end,attr) ->
	(open_box 2; pp_exp exp; print_string " => forwardImmediate()";
	 close_box())


and pp_exp exp =
  pp_ucexp exp;
  match B.com (Ast.get_exp_attr exp) with
    None -> ()
  | Some(str) -> print_string " /* "; print_string str; print_string " */"

(* eventually, compare the precedence of the expression with the precedence
of the context, as done for devil *)
and print_paren exp =
  match exp with
    Ast.INT(n,attr) -> pp_exp exp

    | Ast.VAR(id,attr) -> pp_exp exp

    | Ast.FIELD(exp1,fld,attr) -> pp_exp exp

    | Ast.BOOL(v,attr) -> pp_exp exp

    | Ast.UNARY(uop,uexp,attr) -> pp_exp exp

    | Ast.BINARY(bop,exp1,exp2,attr)
    | Ast.PBINARY(bop,exp1,_,exp2,_,attr) ->
	(print_string "("; pp_exp exp; print_string ")")

    | Ast.INDR(exp,attr) -> pp_exp exp
	  
    | Ast.SELECT(_,attr) -> pp_exp exp

    | Ast.EMPTY(id,_,_,_,attr) -> pp_exp exp

    | Ast.PRIM(f,args,attr) -> pp_exp exp

    | Ast.SCHEDCHILD(arg,procs,attr) -> pp_exp exp

    | Ast.SRCONSCHED(attr) -> pp_exp exp

    | Ast.ALIVE(_,_,attr) -> pp_exp exp

    | Ast.AREF(exp1,exp2,attr) -> pp_exp exp

    | Ast.IN(exp1,id,_,_,_,attr) ->
	(print_string "("; pp_exp exp; print_string ")")

    | Ast.MOVEFWDEXP(exp1,state_end,attr) -> pp_exp exp

and pp_bop op = print_string (" "^(Ast.bop2c op)^" ")

(* ----------------------- User-declared types ---------------------- *)

let rec pp_typedefs typedefs =
  List.iter
    (function
	Ast.ENUMDEF(id,idlist,attr) ->
	  (print_string "enum"; print_space(); pp_id(id);
	   print_space(); print_string "=";
	   print_space(); print_string "{";
	     print_between pp_comma pp_id idlist;
	   print_string "}"; force_newline();)
	| Ast.ENUMVALDEF(id,idlist,attr) ->
	  (print_string "enum"; print_space(); pp_id(id);
	   print_space(); print_string "=";
	   print_space(); print_string "{";
	   print_between pp_comma
	     (function (id,exp) -> pp_id id; print_string " = "; pp_exp exp)
	     idlist;
	   print_string "}"; force_newline();)
	| Ast.RANGEDEF(id,exp1,exp2,attr) ->
	    (print_string "enum"; print_space(); pp_id(id);
	     print_space(); print_string "=";
	     print_space(); print_string "[";
	     pp_exp exp1; print_space(); print_string ".."; print_space();
	     pp_exp exp2; print_string "]"; print_string ";";
	     force_newline())
	| Ast.TIMERDEF(id,decls,attr) ->
	    pp_id id; print_string " timer {";
	    force_newline(); print_string "  "; open_box 0;
	    print_between force_newline
	      (function x -> pp_decl x; print_string ";")
	      decls;
	    close_box(); force_newline();
	    print_string "}"; force_newline())
    typedefs; nl_ifnonempty(typedefs)

(* ----------------------- External functions ---------------------- *)

let pp_fundecls fundecls =
  List.iter
    (function
	Ast.FUNDECL(ret,nm,params,attr) ->
	  print_string (B.type2c(ret)); print_string " ";
	  print_string (B.id2c(nm)); print_string "(";
	  print_between pp_comma (function x -> print_string(B.type2c(x)))
	    params;
	  print_string ");"; force_newline())
    fundecls; nl_ifnonempty(fundecls)

(* -------------------------- Declarations ------------------------- *)

let rec pp_def = function
    Ast.VALDEF(decl,exp,isconst,attr) ->
      (open_box 2;
       if isconst=Ast.CONST then (print_string "const"; print_space());
       pp_decl decl; print_string " = ";
       pp_exp exp; print_string ";"; close_box(); force_newline())
  | Ast.UNINITDEF(decl,attr) ->
      (open_box 2;
       pp_decl decl; print_string ";"; close_box(); force_newline())
  | Ast.SYSDEF(decl,isconst,attr) ->
      (open_box 2;
       pp_decl decl; print_string ";"; close_box(); force_newline())
  | Ast.DUMMYDEF(id,exp,attr) ->
      (open_box 2; print_string "???"; print_space(); pp_id(id);
       print_string " = "; pp_exp exp; print_string ";";
       close_box(); force_newline())

let pp_defs defs = (List.iter pp_def defs; nl_ifnonempty(defs))

(* --------------------------- Statements -------------------------- *)

let rec pp_stmt = function
    Ast.IF(exp,stmt1,stmt2,attr) ->
      (print_string "if ("; pp_exp exp;
       print_string ") "; pp_seq stmt1 attr;
       match stmt2 with
	 Ast.SEQ([],[],_) -> ()
       | _ -> (force_newline();
	       print_string "else"; print_space(); pp_seq stmt2 attr))
	
    | Ast.FOR(id,state_id,_,dir,stmt,attr) ->
	(print_string "foreach ("; pp_id(id);
	 (match state_id with
	   None -> ()
	 | Some(state_ids) ->
	     (print_string " in ";
	      print_between pp_comma pp_id state_ids));
	 print_string ") "; pp_seq stmt attr)

    | Ast.FOR_WRAPPER(label,stms,attr) ->
	print_string "/* an exploded for loop */"; force_newline();
	print_between force_newline pp_stmt stms; force_newline();
	print_string "/* end of an exploded for loop */"
	  
    | Ast.SWITCH(exp,cases,default,attr) ->
	(print_string "switch "; pp_exp exp;
	 print_string " in "; pp_cases cases default)
	  
    | (Ast.SEQ(decls,stmts,attr) as x) -> pp_seq x attr
	  
    | Ast.RETURN(None,attr) ->
	print_string "return;"
	  
    | Ast.RETURN(Some(exp),attr) ->
	(open_box 2; print_string "return "; pp_exp exp;
	 print_string ";"; close_box())
	  
    | Ast.MOVE(exp,state,_,_,_,_,_,attr) ->
	(open_box 2; pp_exp exp; print_string " => ";
	 pp_id state; print_string ";"; close_box())
	       
    | Ast.MOVEFWD(exp,src_dests,_,attr) ->
	(open_box 2; pp_exp exp; print_string " => forwardImmediate();";
	 if !print_comments
	 then
	   begin
	     print_string "/* ";
	     print_between pp_comma
	       (function (src,dst) ->
		 print_string (CS.state2c src);
		 print_string "->";
		 print_string (CS.state2c dst))
	       src_dests;
	     print_string " */"
	   end;
	   close_box())
	       
    | Ast.SAFEMOVEFWD(exp,src_dsts,stm,_,attr) ->
	(open_box 2;
	 print_string "if(";
	 pp_exp exp;
	 print_string ") ";
	 pp_exp exp;
	 print_string " => forwardImmediate();";
	 if !print_comments
	 then
	   begin
	     print_string "/* ";
	     print_between pp_comma
	       (function (src,dst) ->
		 print_string (CS.state2c src);
		 print_string "->";
		 print_string (CS.state2c dst))
	       src_dsts;
	     print_string " */"
	   end;
	 force_newline();
	 print_string "else ";
	 pp_seq stm attr;
	 close_box())
	       
    | Ast.ASSIGN(expl,expr,sorted_fld,attr) ->
	(open_box 2;
	 pp_exp expl; print_space(); print_string "="; print_space();
	 pp_exp expr; print_string ";";
	 if !print_comments
	 then
	   begin
	     print_string "/* ";
	     if sorted_fld then print_string "true" else print_string "false";
	     print_string " */"
	   end;
	 close_box())

    | Ast.DEFER(attr) -> print_string "deferEvent();"

    | Ast.BREAK(attr) -> print_string "break;"

    | Ast.PRIMSTMT(f,args,attr) ->
	(print_paren f; print_string "(";
	 open_box 0;
	 print_between pp_comma pp_exp args; close_box(); print_string ");")

    | Ast.ERROR(str,attr) ->
	print_string "error(\""; print_string str; print_string "\");"

    | Ast.ASSERT(exp,_) ->
	(open_box 2; print_string "assert("; pp_exp exp; print_string ");";
	 close_box())

and pp_seq s attr =
  let (decls,stmts) =
    match s with
      Ast.SEQ(d,s,_) -> (d,s)
    | _ -> ([],[s]) in
  (print_string "{";
   (* print_string "/* "; print_int (B.line attr); print_string " */"; *)
   force_newline(); print_string "  "; open_box 0;
   (match decls with
     [] -> ()
   | x::xs -> pp_defs decls);
   print_between force_newline pp_stmt stmts;
   close_box(); force_newline(); print_string "}")

and pp_cases cases default =
  (print_string "{"; force_newline(); print_string "  "; open_box 0;
   print_between force_newline
     (function Ast.SEQ_CASE(idlist,_,stmt,attr) ->
       (print_string "case ";
	print_between pp_comma pp_id idlist;
	print_string ": "; pp_seq stmt attr;))
     cases;
   close_box(); force_newline();
   (match default with
     None -> ()
   | Some x -> pp_stmt x);
   print_string "}")

(* ------------------------ Ordering criteria ----------------------- *)

let rec pp_criteria criteria =
  (print_string "ordering_criteria"; print_space(); print_string "=";
   print_space(); print_string "{";
   print_between pp_comma pp_crit criteria;
   print_string "}"; force_newline(); force_newline())

and pp_crit = function
    Ast.CRIT_ID(key,order,id,attr) ->
      (pp_key key; pp_crit_order order; pp_id(id))
  | Ast.CRIT_TEST(key,order,tst,thn,els,attr) ->
      (pp_key key; pp_crit_order order;
       print_string "("; print_paren tst; print_space(); print_string "?";
       print_space(); print_paren thn; print_space(); print_string ":";
       print_space(); print_paren els; print_string ")")

and pp_key = function
    Ast.KEY(_,_) -> print_string "key "
  | Ast.NOKEY    -> ()

and pp_crit_order = function
    Ast.LOWEST ->  print_string "lowest "
  | Ast.HIGHEST -> print_string "highest "

(* ----------------------- Admission criteria ----------------------- *)

let pp_admit = function
    None -> ()
  | Some(Ast.ADMIT(locals,crit_params,crits,attach,detach,attr)) ->
      let pp_params params =
	print_string "(";
	print_between pp_comma pp_decl params;
	print_string ")" in
      print_string "admission = {"; force_newline();
      print_string "  "; open_box 0;
      print_between (function x -> ()) pp_def locals;
      force_newline();
      print_string "admission_criteria "; pp_params crit_params;
      print_string " = {";
      force_newline(); print_string "  "; open_box 0;
      print_between
	(function _ -> print_space(); print_string "&&";
	  force_newline())
	(function
	    Ast.RETURN(Some exp,a) -> pp_exp exp
	  | s -> pp_stmt s; print_string "/*error*/")
	crits;
      close_box(); force_newline(); print_string "}";
      force_newline(); force_newline();
      (match attach with
	Some(attach_params,attach) ->
	  print_string "admission_attach "; pp_params [attach_params];
	  print_string " ";
	  pp_stmt attach; force_newline(); force_newline()
      |	_ -> ());
      (match detach with
	Some(detach_params,detach) ->
	  print_string "admission_detach "; pp_params [detach_params];
	  print_string " ";
	  pp_stmt detach; close_box(); force_newline()
      |	_ -> close_box(); force_newline());
      print_string "}"; force_newline(); force_newline()

(* --------------------- Handlers and functions --------------------- *)

(* parser should check that event names are ok *)
let rec pp_handlers handlers =
  let param = match handlers with
    (Ast.EVENT(nm,specd,param,stmt,attr) :: _) -> param
  | _ -> raise (Error.Error "internal error: handler list must not be empty")
  in print_string "handler (";
  pp_decl param;
  print_string ") {"; force_newline(); print_string "  "; open_box 0;
  print_between (function _ -> force_newline(); force_newline())
    (function Ast.EVENT(nm,specd,param,stmt,attr) ->
      (print_string "On ";
       print_string (String.concat "." (List.map Ast.event_name2c specd));
       if not (specd = [])
       then print_string ".";
       print_string (Ast.event_name2c(nm));
       print_string " "; pp_seq stmt attr))
    handlers;
  close_box(); force_newline();
  print_string "}"; force_newline(); force_newline()

let pp_ifunctions functions =
  print_string "interface = {";
  force_newline(); print_string "  "; open_box 0;
  print_between (function _ -> force_newline(); force_newline())
    (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) ->
      (print_string (B.type2c(ret)); print_space (); pp_id(nm);
       print_string "(";
       print_between pp_comma pp_decl params;
       print_string ")"; print_space(); pp_seq stmt attr))
    functions;
  close_box(); force_newline();
  print_string "}"; force_newline(); force_newline()

let pp_functions functions =
  print_string "functions = {";
  force_newline(); print_string "  "; open_box 0;
  print_between (function _ -> force_newline(); force_newline())
    (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) ->
      (print_string (B.type2c(ret)); print_space (); pp_id(nm);
       print_string "(";
       print_between pp_comma pp_decl params;
       print_string ")"; print_space(); pp_seq stmt attr))
    functions;
  close_box(); force_newline();
  print_string "}"; force_newline(); force_newline()

let print_pos = function
    Ast.ON -> print_string "On "
  | Ast.BEFORE -> print_string "Before "
  | Ast.AFTER -> print_string "After "

let pp_aspects = function
    Ast.INITASPECTS(aspects) ->
      print_string "transition = {";
      force_newline(); print_string "  "; open_box 0;
      print_between (function _ -> force_newline(); force_newline())
	(function Ast.ASPECT(srcs,_,dsts,_,param,pos,stmt,_,a) ->
	  (print_pos pos;
	   print_between pp_comma pp_id srcs;
	   print_string " => ";
	   print_between pp_comma pp_id dsts;
	   print_space();
	   print_string "("; pp_decl param; print_string ") ";
	   pp_seq stmt a))
	aspects;
      close_box(); force_newline();
      print_string "}"
  | Ast.ASPECTINFO(aspects) ->
      print_string "transition = {";
      force_newline(); print_string "  "; open_box 0;
      let process (src,dst) l =
	print_between (function _ -> force_newline())
	  (function Ast.ASPECT(srcs,_,dsts,_,param,pos,stmt,_,a) ->
	    (print_pos pos;
	     print_string (CS.state2c src);
	     print_string " => ";
	     print_string (CS.state2c dst);
	     print_space();
	     print_string "("; pp_decl param; print_string ") ";
	     pp_seq stmt a))
	  l in
      List.iter
	(function (transition,(bef,aft)) ->
	  process transition bef;
	  force_newline();
	  process transition aft;
	  force_newline();
	  force_newline())
	aspects;
      close_box(); force_newline();
      print_string "}"
  | _ -> failwith "pp: unexpected aspects"


(* -------------------------- entry point --------------------------- *)

let pretty_print comments out
    (Ast.SCHEDULER(nm,default,
		   cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		   admit,trace,handlers,ifunctions,functions,aspects,attr)) =
  print_comments := comments;
  set_formatter_out_channel (open_out out);
  set_margin 80;
  print_string (Objects.sched_select "scheduler " "virtual_scheduler ");
  print_string nm; print_string " = {"; force_newline(); print_string "  ";
  open_box 0;
  pp_defs cstdefs;
  pp_typedefs enums;
  pp_procdecls procdecls;
  pp_fundecls fundecls;
  pp_defs valdefs;
  pp_states states;
  pp_criteria criteria;
  pp_admit admit;
  pp_handlers handlers;
  pp_ifunctions ifunctions;
  pp_functions functions;
  pp_aspects aspects;
  close_box(); force_newline();
  print_string "}";
  print_newline()
