(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: pre_type.ml,v 1.12 2004/10/20 15:55:47 julia Exp $ *)

module B = Objects
module CS = Class_state

let error (line,str) =
  (Error.update_error();
   if (line > 0)
   then Printf.printf "line %d: %s\n" line str
   else Printf.printf "%s\n" str)

(* Add code to compute scheduler state to the end of each handler *)

let make_sched_state_code attr mkRET =
  let line = B.line attr in
  let ss_var = Ast.mkVAR("Scheduler_state",line) in
  let empty_running =
    Ast.EMPTY(B.mkId("RUNNING"),[],B.UNK,false,attr)
  and empty_ready =
    Ast.EMPTY(B.mkId("READY"),[],B.UNK,false,attr) in
  Ast.mkIF(empty_running,
	   Ast.mkIF(empty_ready,mkRET "BLOCKED",mkRET "READY",line),
	   mkRET "RUNNING",line)

let sched_state_code attr =
  let line = B.line attr in
  make_sched_state_code attr
    (function str -> Ast.mkRETURN(Ast.mkVAR(str,line),line))

(* ------------------ Code to end persistent timers ----------------- *)

let end_persistent_timer nm persistent_timers stm attr =
  try
    let snm = Ast.event_name2c nm in
    let (timer_id,ty) =
      List.find
	(function (pt,ty) -> snm = "unblock.timer."^(B.id2c pt))
	persistent_timers in
    let start =
      match ty with
	None -> None
      |	Some ty ->
	  Some(Ast.VALDEF(Ast.VARDECL(B.STRUCT(ty),timer_id,false,attr),
			  Ast.VAR(B.mkId "local_data",attr),Ast.CONST,attr)) in
    let finish =
      Ast.IF (Ast.UNARY(Ast.NOT,Ast.ALIVE(B.UNK,false,attr),attr),
	      Ast.SEQ([],
		      [Ast.PRIMSTMT(Ast.VAR(B.mkId("delete_bossa_timer"),attr),
				    [Ast.VAR(B.mkId "self",attr)],
				    attr);
			Ast.PRIMSTMT(Ast.VAR(B.mkId("decrement_ref_count"),
					     attr),
				     [],attr)],
		      attr),
	      Ast.SEQ([],[],attr),
	      attr) in
    Some (start,finish)
  with Not_found -> None

(* --------------------- Add to errors positions -------------------- *)

let rec add_stm f stm =
  match f stm with
    Some s -> s
  | None ->
      (match stm with
	Ast.IF(exp,stm1,stm2,attr) ->
	  Ast.IF(exp,add_stm f stm1,add_stm f stm2,attr)
      | Ast.SWITCH(exp,cases,default,attr) ->
	  Ast.SWITCH(exp,
		     List.map
		       (function Ast.SEQ_CASE(pat,states,stm1,attr) ->
			 Ast.SEQ_CASE(pat,states,add_stm f stm1,attr))
		       cases,
		     Aux.app_option (add_stm f) default,
		     attr)
      | Ast.SEQ(decls,stms,attr) ->
	  Ast.SEQ(decls,List.map (add_stm f) stms,attr)
      | Ast.FOR(index,state,state_info,dir,stm1,attr) ->
	  Ast.FOR(index,state,state_info,dir,add_stm f stm1,attr)
      | Ast.FOR_WRAPPER(label,stm1,attr) ->
	  Ast.FOR_WRAPPER(label,List.map (add_stm f) stm1,attr)
      | stm -> stm)
	
let add_error stm curstm =
  add_stm
    (function
	Ast.ERROR(str,attr) as stm1 ->
	  Some(Ast.SEQ([],[stm1;stm],attr))
      |	_ -> None)
    curstm

let add_tail stm curstm =
  add_stm
    (function
	Ast.RETURN(None,attr) -> Some(stm)
      |	_ -> None)
    curstm

let add_before_return stm curstm =
  add_stm
    (function
	Ast.RETURN(_,attr) as stm1 -> Some(Ast.SEQ([],[stm;stm1],attr))
      |	_ -> None)
    curstm

(* ----- Ensure there every control-flow path ends with a return ---- *)

let rec drop_past_return tail return_ok = function
    Ast.IF(exp,stm1,stm2,attr) ->
      Ast.IF(exp,
	     drop_past_return tail true stm1,
	     drop_past_return tail true stm2,
	     attr)
  | Ast.SWITCH(exp,cases,default,attr) ->
      Ast.SWITCH(exp,
		 List.map
		   (function Ast.SEQ_CASE(pat,states,stm1,attr) ->
		     Ast.SEQ_CASE(pat,states,drop_past_return tail true stm1,
				  attr))
		   cases,
		 Aux.app_option (drop_past_return tail true) default,
		 attr)
  | Ast.SEQ(decls,stms,attr) ->
      (match List.rev stms with
	[] ->
	  Ast.SEQ(decls,(if tail then [Ast.RETURN(None,attr)] else []),attr)
      | (last::rest) ->
	  Ast.SEQ(decls,
		  List.rev
		    ((drop_past_return tail return_ok last) ::
		     (List.map (drop_past_return false false) rest)),
		  attr))
  | Ast.FOR(index,state,state_info,dir,stm1,attr) ->
      let stm =
	Ast.FOR(index,state,state_info,dir,drop_past_return false true stm1,
		attr) in
      if tail then Ast.SEQ([],[stm;Ast.RETURN(None,attr)],attr) else stm
  | Ast.FOR_WRAPPER(label,stms,attr) ->
      let stm =
	Ast.FOR_WRAPPER(label,List.map (drop_past_return false true) stms,
		attr) in
      if tail then Ast.SEQ([],[stm;Ast.RETURN(None,attr)],attr) else stm
  | Ast.RETURN(_,attr) as x ->
      if not return_ok
      then
	raise (Error.Error (Printf.sprintf "%s: dead code following return"
			      (B.loc2c attr)));
      x
  | Ast.ERROR(_,_) as x -> x
  | stm ->
      let attr = Ast.get_stm_attr stm in
      if tail then Ast.SEQ([],[stm;Ast.RETURN(None,attr)],attr) else stm

(* ----------------------------- States ----------------------------- *)

(* Add a "nowhere" state, so that being not yet in the system is made
explicit.  This state is discarded during postprocessing, in compile_misc.ml *)

let add_nowhere states =
  let nowhere = B.fresh_idx "nowhere" in
  (Ast.QUEUE(CS.NOWHERE,CS.FIFO,nowhere,CS.PRIVATE,B.mkAttr(-1))) :: states

let add_terminated states =
  match !B.schedtype with
    B.PROCSCHED -> states
  | B.VIRTSCHED -> 
      if List.exists
	  (function
	      Ast.QUEUE(CS.TERMINATED,_,_,_,_) -> true
	    | Ast.PROCESS(CS.TERMINATED,_,_,_,_) -> true
	    | _ -> false)
	  states
      then
	(error(-1,"virtual scheduler should not define a terminated state");
	 states)
      else
	let terminated = B.fresh_idx "terminated" in
	(Ast.QUEUE(CS.TERMINATED,CS.EMPTY,terminated,CS.PRIVATE,B.mkAttr(-1)))
	:: states

(* --------------------- Handlers and functions --------------------- *)

let preprocess_handlers handlers persistent_timers =
  let handlers =
    (match handlers with
      Ast.EVENT(nm,specd,(Ast.VARDECL(ty,id,import,vattr) as param),stm,attr)
      :: _ ->
	if !B.schedtype = B.VIRTSCHED
	then
	  let target = Ast.FIELD(Ast.VAR(id,vattr),B.mkId "target",attr) in
	  let forward_dest = Ast.SCHEDCHILD(target,[],attr) in
	  let forward_target = Ast.MOVEFWD(forward_dest,[],Ast.EITHER,attr) in
	  let safe_forward_target =
	    Ast.SAFEMOVEFWD(forward_dest,[],Ast.SEQ([],[],attr),Ast.EITHER,
			    attr) in
	  Ast.EVENT(Ast.EVENT_NAME([B.mkId "compute_state"],false,attr),
		    [],param,safe_forward_target,attr) ::
	  Ast.EVENT(Ast.EVENT_NAME([B.mkId "attach"],false,attr),
		    [],param,forward_target,attr) ::
	  Ast.EVENT(Ast.EVENT_NAME([B.mkId "detach"],false,attr),
		    [],param,forward_target,attr) ::
	  handlers
	    
	else
	  Ast.EVENT(Ast.EVENT_NAME([B.mkId "compute_state"],false,attr),
		    [],param,Ast.SEQ([],[],attr),attr) :: handlers
    | _ -> raise (Error.Error "internal error")) in
  List.map
    (function Ast.EVENT(nm,specd,param,stm,attr) ->
      let stm = if !B.mutate then stm else drop_past_return true true stm in
      match Ast.event_name2c nm with
	"preempt" -> Ast.EVENT(nm,[],param,stm,attr)
      |	_ ->
	  if !B.mutate
	  then
	    (* not bothering with killing timers here... *)
	    Ast.EVENT(nm,[],param,
		      Ast.postBLOCK attr stm [sched_state_code attr],
		      attr)
	  else
	    let stm =
	      match end_persistent_timer nm persistent_timers stm attr with
		None -> stm
	      |	Some (dec,post) ->
		  let body = add_before_return post stm in
		  (match dec with
		    Some dec -> Ast.SEQ([dec],[body],attr)
		  | None -> body) in
	    Ast.EVENT(nm,[],param,
		      add_tail (sched_state_code attr) stm,
		      attr))
    handlers

let call s attr = Ast.PRIMSTMT(Ast.VAR(B.mkId(s),attr),[],attr)

let preprocess_ifunctions ifunctions terminated =
  List.map
    (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) as x ->
      let stmt = drop_past_return true true stmt in
      match (B.id2c nm,!B.schedtype) with
	("attach",B.PROCSCHED) ->
	  let code = sched_state_code attr in
	  let inc = call "increment_ref_count" attr in
	  Ast.FUNDEF(ret,nm,params,
		     add_tail code
		       (add_error code (add_before_return inc stmt)),
		     inl,attr)
      |	("attach",B.VIRTSCHED) ->
	  let inc = call "increment_ref_count" attr in
	  Ast.FUNDEF(ret,nm,params,add_before_return inc stmt,inl,attr)
      |	("detach",_) ->
	  (match (params,terminated) with
	    ([Ast.VARDECL(_,id,imported,attr)],[termid]) ->
	      let code = sched_state_code attr in
	      let move =
		Ast.MOVE(Ast.VAR(id,attr),termid,None,[],None,false,Ast.EITHER,
			 attr) in
	      let (new_code,new_params) =
		match !B.schedtype with
		  B.PROCSCHED -> (Ast.SEQ([],[move;code],attr),params)
		| B.VIRTSCHED ->
		    let free = Ast.PRIMSTMT(Ast.VAR(B.mkId "kfree",attr),
					    [Ast.VAR(id,attr)],
					    attr) in
		    let ef =
		      Ast.IF(Ast.IN(Ast.VAR(id,attr),B.mkId "BLOCKED",[],
				    B.UNK,false,attr),
			     Ast.SEQ([],[move;free],attr),
			     Ast.SEQ([],[Ast.ERROR("bad state",attr)],attr),
			     attr) in
		    let outer_if =
		      Ast.IF(Ast.VAR(B.mkId "discard_child",attr),
			     Ast.SEQ([],[ef],attr),
			     Ast.SEQ([],[move;free],attr),
			     attr) in
		    let retn = Ast.RETURN(None,attr) in
		    let cvar = Ast.mkVAR("children",B.line attr) in
		    let chld = Ast.ASSIGN(cvar,
					  Ast.BINARY(Ast.MINUS,cvar,
						     Ast.INT(1,attr),
						    attr),
					  false,
					  attr) in
		    (Ast.SEQ([],[outer_if;chld;retn],attr),
		     (params@
		      [Ast.VARDECL(B.BOOL,B.mkId "discard_child",false,attr)]))
	      in
	      Ast.FUNDEF(ret,nm,new_params,
			 add_tail new_code (add_error code stmt),
			 inl,attr)
	  | _ -> x) (* no point to change - type checking will give an error *)
      |	_ -> Ast.FUNDEF(ret,nm,params,stmt,inl,attr))
  ifunctions

(* --------------------------- Functions ---------------------------- *)

let preprocess_functions functions =
  List.map
    (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) as x ->
      let stmt = drop_past_return true true stmt in
      Ast.FUNDEF(ret,nm,params,stmt,inl,attr))
  functions

(* ----------------------------- Timers ----------------------------- *)

let add_timer_enum procdecls valdefs enums =
  let mk tmp_timer_names str =
    let (_,timer_names) =
      List.fold_left
	(function (ctr,timer_names) ->
	  function id ->
	    (ctr+1,
	     (B.mkId("EVENT_UNBLOCK_TIMER_"^str^"_"
		     ^String.uppercase(B.id2c id)),
	      Ast.BINARY(Ast.LSHIFT,
			 Ast.INT(ctr,B.mkAttr 0),
			 Ast.INT(8,B.mkAttr 0),
			 B.mkAttr 0)) :: timer_names))
	(0,[]) tmp_timer_names in
    timer_names in
  let collect_timer_names decls f =
    Aux.option_filter
      (function
	  Ast.VARDECL(B.TIMER(ty),id,_,_) when f ty -> Some id
	| _ -> None)
      decls in
  let target_names =
     collect_timer_names procdecls
      (function B.PERSISTENT(_) -> false | _ -> true) in
  let persistent_names =
     collect_timer_names procdecls
      (function B.PERSISTENT(_) -> true | _ -> false) in
  let notarget_names =
    Aux.option_filter
      (function
	  Ast.VALDEF(Ast.VARDECL(B.TIMER(_),id,_,_),_,_,_)
	| Ast.SYSDEF(Ast.VARDECL(B.TIMER(_),id,_,_),_,_) -> Some id
	| _ -> None)
      valdefs in
  let timer_names =
    List.rev ((mk target_names "TARGET") @
	      (mk notarget_names "NOTARGET") @
	      (mk persistent_names "PERSISTENT")) in
  match timer_names with
    [] -> enums
  | _ ->
      Ast.ENUMVALDEF(B.mkId "bossa_local_timers", timer_names, B.mkAttr 0)
      :: enums

(* --------------------------- Entry point -------------------------- *)

let preprocess
    (Ast.SCHEDULER(nm,default,
		   cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		   admit,trace,handlers,ifunctions,functions,aspects,attr)) =
  let null_attr = B.mkAttr 0 in
  let persistent_timers =
    Aux.option_filter
      (function
	  Ast.VARDECL(B.TIMER(B.PERSISTENT(ty)),id,_,_) -> Some (id,ty)
	| _ -> None)
      procdecls in
  let ifunctions =
    (match trace with
      Ast.NOTRACE -> ifunctions
    | _ ->
	Ast.FUNDEF(B.VOID,B.mkId "print_trace",[],
		   Ast.PRIMSTMT(Ast.VAR(B.mkId "print_trace_info",null_attr),
				[],null_attr),
		   Ast.NO_INLINE,null_attr)
	:: ifunctions) in
  let new_states = add_terminated(add_nowhere states) in
  let new_valdefs =
    match !B.schedtype with
      B.PROCSCHED -> valdefs
    | B.VIRTSCHED ->
	Ast.VALDEF(Ast.mkVARDECL(B.INT,"children",-1),
		   Ast.INT(0,null_attr),Ast.VARIABLE,null_attr) ::
	valdefs in
  Preprocess_ifs.preprocess
    (function
      | Ast.SRCONSCHED(attr) -> true
      | Ast.ALIVE(tv,insrc,attr) -> true
      | _ -> false)
  (Ast.SCHEDULER(nm,default,cstdefs,
		 add_timer_enum procdecls valdefs enums,
		 procdecls,fundecls,
		 Ast.SYSDEF(Ast.mkSYSVARDECL(B.ENUM(B.mkId("SchedState")),
					     "root_s_state",0),
			    Ast.VARIABLE,null_attr) ::
		 Ast.SYSDEF(Ast.mkSYSVARDECL(B.STRUCT(B.mkId("task_struct")),
					     "current",0),
			    Ast.VARIABLE,null_attr) ::
		 new_valdefs,
		 new_states,
		 criteria, admit, trace,
		 preprocess_handlers handlers persistent_timers,
		 preprocess_ifunctions ifunctions
		     (Aux.option_filter
			(function
			    Ast.QUEUE(CS.TERMINATED,_,id,_,_) -> Some id
			  | Ast.PROCESS(CS.TERMINATED,id,_,_,_) -> Some id
			  | _ -> None)
			new_states),
		   preprocess_functions functions,aspects,
		   attr))
