(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* Check that global variables and fields are written before they are read.
For attach, not doing this is an error.  Elsewhere, there is a warning
if the read and write are in different modules.

Approximate control flow, ie nothing done about control flow from the bottom
of a loop to the top.

Assume all processes are the same, so ignore the process in a field
reference.  This could give spurious warning messages.  *)

module B = Objects
module CS = Class_state

let error ln str =
  Error.update_error();
  Printf.printf "%s: %s \n" (B.loc2c ln) str

let warning ln str =
  Printf.printf "warning: %s: %s \n" (B.loc2c ln) str

let pr = Printf.sprintf

let in_attach = ref false

let class_ids = List.map CS.class2c (CS.all_classes false)

let is_state id =
  try let _ = List.assoc id (!CS.state_env) in true
  with Not_found -> List.mem (B.id2c id) class_ids

let idunion l1 l2 =
  let names2 = List.map (function (id,_) -> id) l2 in
  let rec loop = function
      [] -> l2
    | ((id,ln) as cur)::rest ->
	if List.mem id names2 then loop rest else cur::(loop rest) in
  loop l1

(* checks variables and fields read by an expression *)
let rec check_exp write_field env field_env = function
    Ast.VAR(id,ln) ->
      if List.mem id env || is_state id
      then ([],[])
      else ([(id,ln)],[])
  | Ast.FIELD(exp,id,ln) ->
      let (read_var,read_field) =
	check_exp write_field env field_env exp in
      if !in_attach && not (List.mem id field_env) (* timers *)
      then
	(try let _ = List.assoc id write_field in ()
	with Not_found ->
	  error ln (pr "field %s read before it is written in attach"
		      (B.id2c id)));
      (read_var, idunion [(id,ln)] read_field)
  | Ast.UNARY(uop,exp,ln) -> check_exp write_field env field_env exp
  | Ast.BINARY(bop,exp1,exp2,ln) ->
      let (read_var1,read_field1) =
	check_exp write_field env field_env exp1 in
      let (read_var2,read_field2) =
	check_exp write_field env field_env exp2 in
      (idunion read_var1 read_var2,idunion read_field1 read_field2)
  | Ast.PBINARY(bop,exp1,states1,exp2,states2,ln) ->
      let (read_var1,read_field1) =
	check_exp write_field env field_env exp1 in
      let (read_var2,read_field2) =
	check_exp write_field env field_env exp2 in
      (idunion read_var1 read_var2,idunion read_field1 read_field2)
  | Ast.INDR(exp,ln) -> check_exp write_field env field_env exp
  | Ast.PRIM(fn,args,ln) ->
      List.fold_left
	(function (read_var,read_field) ->
	  function x ->
	    let (read_var1,read_field1) =
	      check_exp write_field env field_env x in
	    (idunion read_var1 read_var,idunion read_field1 read_field))
	(check_exp write_field env field_env fn) args
  | Ast.SCHEDCHILD(exp,procs,ln) ->
      check_exp write_field env field_env exp
  | Ast.IN(exp,st,sti,tv,insrc,ln) ->
      let (read_var,read_field) =
	check_exp write_field env field_env exp in
      (read_var,read_field)
  | Ast.EMPTY(st,sti,tv,insrc,ln) ->
      ([],[])
  | _ -> ([],[])

(* checks variables and fields read or written by a statement *)
let rec check_stm read_var read_field write_var write_field
    (env : B.identifier list) (field_env : B.identifier list) = function
    Ast.IF(exp,st1,st2,ln) ->
      let (read_var1,read_field1) = check_exp write_field env field_env exp in
      let read_var = Aux.union read_var1 read_var in
      let read_field = Aux.union read_field1 read_field in
      let (read_var2,read_field2,write_var2,write_field2) =
	check_stm read_var read_field write_var write_field env field_env st1 in
      let (read_var3,read_field3,write_var3,write_field3) =
	check_stm read_var read_field write_var write_field env field_env st2 in
      (idunion read_var2 read_var3,
       idunion read_field2 read_field3,
       idunion write_var2 write_var3,
       idunion write_field2 write_field3)
  | Ast.FOR(id,_,sts,_,stmt,ln) ->
      check_stm read_var read_field write_var write_field (id::env) field_env
	stmt
  | Ast.FOR_WRAPPER(_,stmts,ln) ->
      List.fold_left
	(function (read_var,read_field,write_var,write_field) ->
	  function stmt ->
	    check_stm read_var read_field write_var write_field
	      env field_env stmt)
	([],[],[],[]) stmts
  | Ast.SWITCH(exp,lsc,st2,ln) ->
      let (read_var0,read_field00) = check_exp write_field env field_env exp in
      let read_var = idunion read_var0 read_var in
      let read_field = idunion read_field00 read_field in
      List.fold_left
	(function (read_var2,read_field2,write_var2,write_field2) ->
	  function Ast.SEQ_CASE(_,_,st1,_) ->
	    let (read_var1,read_field1,write_var1,write_field1) =
	      check_stm read_var read_field write_var write_field
		env field_env st1 in
	    (idunion read_var1 read_var2,
	     idunion read_field1 read_field2,
	     idunion write_var1 write_var2,
	     idunion write_field1 write_field2))
	(match st2 with
	  None -> ([],[],[],[])
	| Some st ->
	    check_stm read_var read_field write_var write_field
	      env field_env st)
	lsc
  | Ast.SEQ(decls,lstmt,ln) ->
      let (read_var1,read_field1,env) =
	List.fold_left
	  (function (read_var,read_field,env) ->
	    function
		Ast.VALDEF(Ast.VARDECL(ty,id,sys,ln),exp,const,a) ->
		  let (read_var1,read_field1) =
		    check_exp write_field env field_env exp in
		  (idunion read_var1 read_var, idunion read_field1 read_field,
		   Aux.union [id] env)
	      |	_ ->
		  raise (Error.Error (Printf.sprintf
					"%s: unexpected local decl"
					(B.loc2c ln))))
	  ([],[],env) decls in
      let read_var = idunion read_var1 read_var in
      let read_field = idunion read_field1 read_field in
      List.fold_left
	(function (read_var,read_field,write_var,write_field) ->
	  function x ->
	    check_stm read_var read_field write_var write_field env field_env x)
	(read_var,read_field,write_var,write_field) lstmt
  | Ast.RETURN(Some exp,ln) ->
      let (read_var1,read_field1) = check_exp write_field env field_env exp in
      let read_var = idunion read_var1 read_var in
      let read_field = idunion read_field1 read_field in
      (read_var,read_field,write_var,write_field)
  | Ast.MOVE(proc,dst,srcinfo,vsrcinfo,dstinfo,auto,state_end,_) ->
      let (read_var1,read_field1) = check_exp write_field env field_env proc in
      let read_var = idunion read_var1 read_var in
      let read_field = idunion read_field1 read_field in
      (read_var,read_field,write_var,write_field)
  | Ast.MOVEFWD(proc,srcdst,state_end,_) ->
      let (read_var1,read_field1) = check_exp write_field env field_env proc in
      let read_var = idunion read_var1 read_var in
      let read_field = idunion read_field1 read_field in
      (read_var,read_field,write_var,write_field)
  | Ast.ASSIGN(lexp,rexp,affectsrtd,ln) ->
      let check str id a read write =
	(try let rl = List.assoc id read in (* prevly read *)
	try let _ = List.assoc id write in ()(*prevly written too*)
	with Not_found ->
	  if B.modl rl = B.modl a
	  then ()
	  else
	    warning a (pr "written %s %s previously read in %s"
			 str (B.id2c id) (B.loc2c rl))
	with Not_found -> ()) in	
      let (read_var1,read_field1) = check_exp write_field env field_env rexp in
      let read_var = idunion read_var1 read_var in
      let read_field = idunion read_field1 read_field in
      let (read_var2,read_field2,write_var2,write_field2) =
	match lexp with
	  Ast.VAR(id,a) ->
	    if List.mem id env
	    then ([],[],[],[])
	    else (* only check for non-local variables *)
	      begin
		check "variable" id a read_var write_var;
		([],[],[(id,a)],[])
	      end
	| Ast.FIELD(exp,id,a) ->
	    let (read_var3,read_field3) =
	      check_exp write_field env field_env exp in
	    check "field" id a read_field write_field;
	    (read_var3,read_field3,[],[(id,a)])
	| _ -> raise (Error.Error "unexpected left-hand side expression") in
      let read_var = idunion read_var2 read_var in
      let read_field = idunion read_field2 read_field in
      let write_var = idunion write_var2 write_var in
      let write_field = idunion write_field2 write_field in
      (read_var,write_var,write_var,write_field)
  | Ast.PRIMSTMT(fn,args,ln) ->
      let (read_var1,read_field1) =
	List.fold_left
	  (function (read_var,read_field) ->
	    function x ->
	      let (read_var1,read_field1) =
		check_exp write_field env field_env x in
	      (idunion read_var1 read_var,idunion read_field1 read_field))
	  (check_exp write_field env field_env fn) args in
      let read_var = idunion read_var1 read_var in
      let read_field = idunion read_field1 read_field in
      (read_var,read_field,write_var,write_field)
  | Ast.ASSERT(exp,ln) ->
      let (read_var1,read_field1) = check_exp write_field env field_env exp in
      let read_var = idunion read_var1 read_var in
      let read_field = idunion read_field1 read_field in
      (read_var,read_field,write_var,write_field)
  (* the following can't occur or no expression subterms *)
  | _ -> (read_var,read_field,write_var,write_field)

(* --------------------- Handlers and functions --------------------- *)

let process_handlers timers_with_targets timers_without_targets handlers =
  List.iter
    (function (Ast.EVENT(nm,specd,param,stmt,attr)) ->
      in_attach := false;
      let locals = Aux.union timers_without_targets (Ast.params2ids [param]) in
      let _ = check_stm [] [] [] [] locals timers_with_targets stmt in ())
    handlers

let process_functions timers_with_targets timers_without_targets functions =
  List.iter
    (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) ->
      in_attach := (B.id2c nm = "attach");
      let locals = Aux.union timers_without_targets (Ast.params2ids params) in
      let _ = check_stm [] [] [] [] locals timers_with_targets stmt in ())
    functions

(* --------------------------- Entry point -------------------------- *)

let readwrite
    (Ast.SCHEDULER(nm,default,
		   cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		   admit,trace,handlers,ifunctions,functions,aspects,attr)) =
  (* don't both with functions because these aren't yet supported in the
     module language *)
  let timers_with_targets =
    Aux.option_filter
      (function
	  Ast.VARDECL(B.TIMER(_),id,_,_) -> Some id
	| _ -> None)
      procdecls in
  let timers_without_targets =
    Aux.option_filter
      (function
	  Ast.SYSDEF(Ast.VARDECL(B.TIMER(_),id,_,_),_,_) -> Some id
	| _ -> None)
      valdefs in
  process_handlers timers_with_targets timers_without_targets handlers;
  process_functions timers_with_targets timers_without_targets ifunctions
