(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: rename.ml,v 1.11 2004/11/11 16:15:01 julia Exp $ *)


module B = Objects
module CS = Class_state

(* The purpose of this phase is to rename identifiers.  Several kinds of
renaming are done:

1.  Local process-typed variables are renamed to be different from each
    other and to be different from global process-typed variables.

2.  Local variables named "old_running" and all occurrences of "running"
    are renamed.

3.  The previous variable of the RUNNING state, if any, is designated as
    RTS_PREV rather than BOSSA_PREV to indicate that it is managed by the RTS.

4.  We check that no field of the process structure has the name "running"
    or "old_running".

Global variables are not renamed.  If there is a global variable of any
sort named running or old_running, this will be detected later on. *)

exception LookupErrException

let error (loc,str) =
  (Error.update_error();
   match loc with
     Some lstr -> Printf.printf "%s: %s\n" lstr str
   | None -> Printf.printf "%s\n" str)

let rec lookup_id x = function
    [] -> raise LookupErrException
  | ((y,v)::r) -> if B.ideq(x,y) then v else lookup_id x r

let check env id =
  try lookup_id id env with LookupErrException -> id

let extend_env nm vl env = (nm,vl)::env

let reserved = (ref [] : B.identifier list ref)

(* ------------------------ Process structure ----------------------- *)

let check_procdecls procdecls =
  List.iter
    (function Ast.VARDECL(ty,id,imported,attr) ->
      if Aux.member id (!reserved)
      then error(Some(B.loc2c attr),"process field cannot be a reserved word"))
    procdecls

(* --------------------------- Expressions -------------------------- *)

let rec rename env = function
      Ast.VAR(id,attr) as exp -> Ast.VAR(check env id, attr)

    | Ast.FIELD(exp,fld,attr) -> Ast.FIELD(rename env exp,fld,attr)

    | Ast.UNARY(uop,exp,attr) -> Ast.UNARY(uop,rename env exp,attr)

    | Ast.BINARY(bop,exp1,exp2,attr) ->
	Ast.BINARY(bop,rename env exp1,rename env exp2,attr)

    | Ast.PBINARY(bop,exp1,states1,exp2,states2,attr) ->
	Ast.PBINARY(bop,rename env exp1,states1,rename env exp2,states2,attr)

    | Ast.INDR(exp,attr) -> Ast.INDR(rename env exp,attr)

    | Ast.PRIM(f,exps,attr) ->
	Ast.PRIM(rename env f,List.map (rename env) exps,attr)

    | Ast.IN(exp,id,x,y,z,attr) ->
	let id = check env id in
	Ast.IN(rename env exp,id,x,y,z,attr)

    | Ast.EMPTY(id,l,x,y,attr) ->
	let id = check env id in
	Ast.EMPTY(id,l,x,y,attr)

    | exp -> exp

(* -------------------------- Declarations -------------------------- *)

(* A global is only renamed if its name is reserved and if the type is not
   a system type *)
let rec rename_global_decls decls env =
  List.map
    (function
	Ast.VALDEF(Ast.VARDECL(ty,id,imported,attr) as x,exp,cst,attr1) ->
	  Ast.VALDEF(Ast.VARDECL(ty,check env id,imported,attr),
		     rename env exp,cst,attr1)
      | Ast.UNINITDEF(Ast.VARDECL(ty,id,imported,attr) as x,attr1) ->
	  Ast.UNINITDEF(Ast.VARDECL(ty,check env id,imported,attr),attr1)
      | Ast.SYSDEF(Ast.VARDECL(ty,id,imported,attr),isconst,attr1) as x ->
	  if Aux.member id (!reserved)
	  then error(Some(B.loc2c attr),
		     "system global cannot be a reserved word");
	  x
      | Ast.DUMMYDEF(id,exp,attr) -> (* never of type process *)
	  Ast.DUMMYDEF(id,rename env exp,attr))
    decls

(* A local variable is renamed if it is of type process or if its name is
   reserved *)
let rec rename_decls decls env =
  List.fold_left
    (function (env,prev_decls) ->
      function
	  Ast.VALDEF(Ast.VARDECL(B.PROCESS,id,imported,attr),exp,cst,attr1) ->
	    let newid = pick_new id env in
	    (extend_env id newid env,
	     Ast.VALDEF(Ast.VARDECL(B.PROCESS,newid,imported,attr),
			rename env exp,cst,attr1) :: prev_decls)
	| Ast.VALDEF(Ast.VARDECL(ty,id,imported,attr),exp,cst,attr1) ->
	    (env,
	     Ast.VALDEF(Ast.VARDECL(ty,check env id,imported,attr),
			rename env exp,cst,attr1) :: prev_decls)
	| Ast.UNINITDEF(Ast.VARDECL(B.PROCESS,id,imported,attr),attr1) ->
	    let newid = pick_new id env in
	    (extend_env id newid env,
	     Ast.UNINITDEF(Ast.VARDECL(B.PROCESS,newid,imported,attr),
			   attr1) :: prev_decls)
	| Ast.UNINITDEF(Ast.VARDECL(ty,id,imported,attr),attr1) ->
	    (env,
	     Ast.UNINITDEF(Ast.VARDECL(ty,check env id,imported,attr),
			   attr1) :: prev_decls)
	| Ast.SYSDEF(Ast.VARDECL(B.PROCESS,id,imported,attr),isconst,attr1) ->
	    let newid = pick_new id env in
	    (extend_env id newid env,
	     Ast.SYSDEF(Ast.VARDECL(B.PROCESS,newid,imported,attr),isconst,
			attr1)
	     :: prev_decls)
	| Ast.SYSDEF(Ast.VARDECL(ty,id,imported,attr),isconst,attr1) as x ->
	    if Aux.member id (!reserved)
	    then
	      let newid = lookup_id id env in
	      (env,Ast.SYSDEF(Ast.VARDECL(ty,newid,imported,attr),isconst,
			      attr1)
	       :: prev_decls)
	    else (env,x::prev_decls)
	| Ast.DUMMYDEF(id,exp,attr) -> (* never of type process *)
	    (env,Ast.DUMMYDEF(id,rename env exp,attr) :: prev_decls))
    (env,[])
    decls

and pick_new x env =
  try let _ = lookup_id x env in B.fresh_idx (B.id2c x)
  with LookupErrException -> x

(* ------------------------------ States ---------------------------- *)

let rename_states states env =
  let (env,states) =
    List.fold_left
      (function (newenv,states) ->
	function
	    Ast.PROCESS(CS.RUNNING,id,Ast.BOSSA_PREV previd,vis,attr) ->
	      let id = check newenv id
	      and old_running = B.mkId "old_running" in
	      (extend_env previd old_running newenv,
	       Ast.PROCESS(CS.RUNNING,id,Ast.RTS_PREV old_running,vis,attr)::
	       states)
	  | Ast.PROCESS(cls,id,prev,vis,attr) ->
	      let id = check newenv id in
	      (newenv,Ast.PROCESS(cls,id,prev,vis,attr)::states)
	  | Ast.QUEUE(cls,order,id,vis,attr) ->
	      let id = check newenv id in
	      (newenv,Ast.QUEUE(cls,order,id,vis,attr)::states))
      (env,[])
      states in
  (env,List.rev states)

(* --------------------------- Statements --------------------------- *)

let rec preprocess_stmt env = function
    Ast.IF(exp,stm1,stm2,attr) ->
      Ast.IF(rename env exp,
	     preprocess_stmt env stm1,preprocess_stmt env stm2,attr)

  | Ast.FOR(id,stid,l,dir,stm,attr) ->
      let new_id = pick_new id env in
      let new_stid =
	match stid with
	  None -> None
	| Some x -> Some (List.map (check env) x) in
      Ast.FOR(new_id,new_stid,l,dir,
	      preprocess_stmt (extend_env id new_id env) stm,
	      attr)

  | Ast.FOR_WRAPPER(label,stms,attr) -> raise (Error.Error "not generated yet")

  | Ast.SWITCH(exp,cases,default,attr) ->
      Ast.SWITCH(rename env exp,
		 List.map
		   (function Ast.SEQ_CASE(pat,l,stm,attr) ->
		     let pat = List.map (check env) pat in
		     Ast.SEQ_CASE(pat,l,preprocess_stmt env stm,attr))
		   cases,
		 Aux.app_option (preprocess_stmt env) default,
		 attr)

  | Ast.SEQ(decls,stmts,attr) ->
      let (env,new_decls) = rename_decls decls env in
      Ast.SEQ(List.rev new_decls,List.map (preprocess_stmt env) stmts,attr)

  | Ast.RETURN(Some(e),attr) -> Ast.RETURN(Some(rename env e),attr)

  | Ast.MOVE(exp,id,src,srcs,dst,auto_allowed,state_end,attr) ->
      let id = check env id in
      Ast.MOVE(rename env exp,id,src,srcs,dst,auto_allowed,state_end,attr)

  | Ast.MOVEFWD(exp,srcs_dsts,state_end,attr) ->
      Ast.MOVEFWD(rename env exp,srcs_dsts,state_end,attr)

  | Ast.SAFEMOVEFWD(exp,srcs_dsts,stm,state_end,attr) ->
      Ast.SAFEMOVEFWD(rename env exp,srcs_dsts,preprocess_stmt env stm,
		      state_end,attr)

  | Ast.ASSIGN(exp1,exp2,sorted_fld,attr) ->
      Ast.ASSIGN(rename env exp1,rename env exp2,sorted_fld,attr)

  | Ast.PRIMSTMT(f,args,attr) ->
      Ast.PRIMSTMT(rename env f,List.map (rename env) args,attr)

  | stmt -> stmt

(* ------------------------ Ordering criteria ----------------------- *)

let preprocess_criteria criteria env =
  List.map
    (function
	Ast.CRIT_ID(key,order,id,attr) as crit ->
	  Ast.CRIT_ID(key,order,check env id,attr)
      |	Ast.CRIT_TEST(key,order,e1,e2,e3,attr) ->
	  Ast.CRIT_TEST(key,order,rename env e1,rename env e2,rename env e3,
			attr))
    criteria

(* ----------------------- Admission criteria ----------------------- *)

let preprocess_admit attach_env env = function
    None -> None
  | Some(Ast.ADMIT(locals,crit_params,crits,attach,detach,attr)) ->
    (* ensure that locals are disjoint and different from other variables *)
    Ast.no_dup_valdefs locals;
    let (env,new_locals) =
      List.fold_left
	(function (env,new_locals) ->
	  function
	      Ast.VALDEF(Ast.VARDECL(ty,id,imp,a),exp,iscst,a1) ->
		let new_exp = rename env exp in
		let newid = pick_new id env in
		let new_decl =
		  Ast.VALDEF(Ast.VARDECL(ty,newid,imp,a),new_exp,iscst,a1) in
		(extend_env id newid env, new_decl::new_locals)
	    | Ast.UNINITDEF(Ast.VARDECL(a,id,b,c),d) ->
		let newid = pick_new id env in
		let new_decl = Ast.UNINITDEF(Ast.VARDECL(a,newid,b,c),d) in
		(extend_env id newid env, new_decl::new_locals)
	    | Ast.SYSDEF(Ast.VARDECL(a,id,b,c),d,e) ->
		let newid = pick_new id env in
		let new_decl = Ast.SYSDEF(Ast.VARDECL(a,newid,b,c),d,e) in
		(extend_env id newid env, new_decl::new_locals)
	    | Ast.DUMMYDEF(id,_,_) ->
		raise (Error.Error "internal error"))
	(env,[]) locals in
    let new_locals = List.rev new_locals in
    (* prefer attach_env to param_env so that criteria parameters will
       be renamed in the same way as in the attach function.  But include
       param_env so that we don't get an error derived from misnamed
       criteria parameters from here, but rather later from the type
       checking phase. *)
    let crit_rename_env =
      let param_env =
	List.map (function id -> (id,id)) (Ast.params2ids crit_params) in
      attach_env@param_env@env in
    let new_crit_params =
      List.map
	(function Ast.VARDECL(ty,id,x,a) ->
	  Ast.VARDECL(ty,check crit_rename_env id,x,a))
	crit_params in
    let new_crits = List.map (preprocess_stmt crit_rename_env) crits in
    let new_attach =
      match attach with
	None -> None
      |	Some(attach_param,attach) ->
	  let param_env =
	    List.map (function id -> (id,id))
	      (Ast.params2ids [attach_param]) in
	  let new_attach = preprocess_stmt (param_env@env) attach in
	  Some(attach_param,new_attach) in
    let new_detach =
      match detach with
	None -> None
      |	Some(detach_param,detach) ->
	  let param_env =
	    List.map (function id -> (id,id))
	      (Ast.params2ids [detach_param]) in
	  let new_detach = preprocess_stmt (param_env@env) detach in
	  Some(detach_param,new_detach) in
    Some(Ast.ADMIT(new_locals,new_crit_params,new_crits,new_attach,new_detach,
		   attr))

(* ------------------------------ Trace ----------------------------- *)

let preprocess_trace env = function
    Ast.PRETRACE(count,trace_info,attr) ->
      Ast.PRETRACE(count,
		   List.map
		     (function
			 Ast.TRACEEVENTS(events) -> Ast.TRACEEVENTS(events)
		       | Ast.TRACEEXPS(exps) ->
			   let new_exps =
			     List.map
			       (function
				   Ast.TRACEVAR(fldid,id,attr) ->
				     Ast.TRACEVAR(fldid,check env id, attr)
				 | _ -> raise (Error.Error "internal error"))
			       exps in
			   Ast.TRACEEXPS(new_exps)
		       | Ast.TRACETEST(exp) -> Ast.TRACETEST(rename env exp))
		     trace_info,
		   attr)
  | Ast.NOTRACE -> Ast.NOTRACE
  | _ -> raise (Error.Error "internal error")

(* --------------------- Handlers and functions --------------------- *)

let preprocess_handlers handlers env =
  List.map
    (function
	(Ast.EVENT(nm,specd,(Ast.VARDECL(_,id,_,_) as param),stmt,attr)) ->
	  let env = (id,id) :: env
	  in Ast.EVENT(nm,specd,param,preprocess_stmt env stmt,attr))
    handlers
    
let preprocess_functions functions env =
  let (attach_env,new_functions) =
    List.fold_left
      (function (attach_env,new_functions) ->
	function (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) ->
	  let snm = B.id2c nm in
	  let (new_vars,new_params) =
	    List.split
	      (if snm = "attach"
 	       (* parameters of attach will become static globals, so they
		  need to be disjoint from other variables *)
	      then List.map
		  (function Ast.VARDECL(a,id,b,c) as x ->
		    let new_id = B.fresh_idx(B.id2c id) in
		    ((id,new_id),Ast.VARDECL(a,new_id,b,c)))
		  params
	      else
		List.map (function Ast.VARDECL(_,id,_,_) as x -> ((id,id),x))
		  params) in
	  let env = new_vars @ env in
	  ((if snm = "attach" then Some new_vars else attach_env),
	   (Ast.FUNDEF(ret,nm,new_params,preprocess_stmt env stmt,inl,attr))::
	   new_functions))
      (None,[]) functions in
  (attach_env,List.rev new_functions)
    
let preprocess_aspects env = function
    Ast.INITASPECTS(aspects) ->
      Ast.INITASPECTS(
      List.map
	(function Ast.ASPECT(srcs,src_infos,dsts,dst_infos,
			     (Ast.VARDECL(_,id,_,_) as param),pos,stmt,fvs,a)->
	  let srcs = List.map (check env) srcs in
	  let dsts = List.map (check env) dsts in
	  let env = (id,id) :: env in
	  Ast.ASPECT(srcs,src_infos,dsts,dst_infos,param,pos,
		     preprocess_stmt env stmt,fvs,a))
	aspects)
  | _ -> failwith "rename: unexpected aspects"
    
let preprocess
    (Ast.SCHEDULER(nm,default,
		   cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		   admit,trace,handlers,ifunctions,functions,aspects,attr)) =
  reserved := List.map B.mkId ["running";"root_s_state";"_checksum";"self";
				"local_data";
				"current";"children";"fwdres"(* for unblock *);
				"trace_buffer";"trace_ctr";"trace_wrapped"];
  let env0 = List.map (function x -> (x,B.fresh_idx(B.id2c x))) (!reserved) in
  check_procdecls procdecls;
  let cstdefs = rename_global_decls cstdefs env0 in
  let valdefs = rename_global_decls valdefs env0 in
  let (env,states) = rename_states states env0 in
  let (_,functions) = preprocess_functions functions env in
  match preprocess_functions ifunctions env with
    (None,_) -> raise (Error.Error "rename: missing attach")
  | ((Some attach_env),ifunctions) ->
      (Ast.SCHEDULER(nm,default,cstdefs,enums,procdecls,fundecls,
		     valdefs,states,
		     preprocess_criteria criteria env,
		     preprocess_admit attach_env env admit,
		     preprocess_trace env trace,
		     preprocess_handlers handlers env,
		     ifunctions,functions,preprocess_aspects env aspects,
		       attr))
