(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: type.ml,v 1.26 2004/11/11 16:15:01 julia Exp $ *)

(* The semantics of switch has changed.  Now if there are three BLOCKED states
and only one is mentioned specifically, then a subsequent case for BLOCKED
refers to the other two.  Referring to either of the remaining BLOCKED states
after the reference to BLOCKED, however, is an error. *)
  
module B = Objects
module CS = Class_state
    
(* -------------------- Miscellaneous operations -------------------- *)
    
exception LookupErrException of string option
exception StmtErrException
exception ExpErrException of B.typ
    
let in_handler = ref false
let in_persistent_timer_handler = ref false
let move_allowed = ref false (* not allowed in VS attach *)
    
let pr = Printf.sprintf
    
let nl_error str =
  (Error.update_error();
   Printf.printf "%s\n" str)
    
let error (attr,str) =
  (Error.update_error();
   Printf.printf "%s: %s\n" (B.loc2c attr) str)
    
let stmtErr (attr,str) =
  (Error.update_error();
   Printf.printf "%s: %s\n" (B.loc2c attr) str;
   raise StmtErrException)
    
let expErr (res,attr,str) =
  (Error.update_error();
   Printf.printf "%s: %s\n" (B.loc2c attr) str;
   raise (ExpErrException res))
    
let expected_err attr str = function
    Some(t) -> expErr(t,attr,str)
  | None -> stmtErr(attr,str)
	
let rec lookup x = function
    [] -> raise (LookupErrException None)
  | ((y,v)::r) -> if x = y then v else lookup x r
      
let rec lookup_id x = function
    [] -> raise (LookupErrException (Some(B.id2c x)))
  | ((y,v)::r) -> if B.ideq(x,y) then v else lookup_id x r
      
let rec lookup_str x = function
    [] -> raise (LookupErrException (Some x))
  | ((y,v)::r) -> if B.strideq(x,y) then v else lookup_str x r
      
let extend_env nm vl env = (nm,vl)::env
				      
let rec id_member x = function
    [] -> false
  | (y::r) -> B.ideq(x,y) || id_member x r
      
(* ----------------------- User-declared types ---------------------- *)
	
(* boolean indicates whether the program can modify the fields of the structure
   (true implies that the program cannot) *)
	
let struct_env =
  ref ([] : (B.identifier * (bool * (B.identifier * B.typ) list)) list)

let init_struct_env _ =
  struct_env :=
    [(B.mkId("event"),
      (true,[(B.mkId("source"),B.SRCPROCESS);(B.mkId("target"),B.PROCESS)]));
      (B.mkId("task_struct"),
       (false,[(B.mkId("need_resched"),B.INT)]))]
      
let extend_struct_env info = (struct_env := info :: (!struct_env))

(* returns a new environment, reflecting the enum declarations, and an
   environment containing the maximum and minimum value for each type *)
let rec typedef_env typedefs env constenv =
  let rr = (B.ENUM(B.mkId("SchedState")),true (* is const *)) in
  let env =
    extend_env (B.mkId("RUNNING")) rr
      (extend_env (B.mkId("READY")) rr
	 (extend_env (B.mkId("BLOCKED")) rr env)) in
  let rr = (B.ENUM(B.mkId("SchedState")),false) in
  let env = extend_env (B.mkId("Scheduler_state")) rr env in
  let rr = (B.NULL,true) in
  let env = extend_env (B.mkId("NULL")) rr env in
  let (env,erenv) =
    List.fold_left
      (function (env,erenv) ->
	function
	    Ast.ENUMDEF(id,idlist,attr) ->
	      ((List.map (function x -> (x,(B.ENUM(id),true))) idlist) @ env,
	       (id, (0,(List.length idlist)-1 (* max value *))) :: erenv)
	  | Ast.ENUMVALDEF(id,idlist,attr) ->
	      ((List.map
		  (function (x,exp) ->
		    let _ = all_int_constant constenv attr exp in
		    (x,(B.ENUM(id),true)))
		  idlist)
	       @ env,
	       (id, (0,(List.length idlist)-1 (* max value *))) :: erenv)
	  | Ast.RANGEDEF(id,exp1,exp2,attr) ->
	      let v1 = all_int_constant constenv attr exp1
	      and v2 = all_int_constant constenv attr exp2 in
	      if v2 < v1
	      then error(attr,pr "range %s is empty" (B.id2c id));
	      (env, (id,(v1,v2)) :: erenv)
	  | Ast.TIMERDEF(id,decls,attr) ->
	      let idtys =
		List.map (function Ast.VARDECL(ty,id,imported,attr) -> (id,ty))
		  decls in
	      extend_struct_env (id,(false,idtys));
	      (env,erenv))
      (env,[]) typedefs in
  (* combine enum-range env with constenv to allow key ordering criteria to
     refer to constant variables with the value 0 or 1 *)
  (env, erenv @ (List.map (function (id,vl) -> (id,(vl,vl))) constenv))
    
and all_int_constant env attr = function
    Ast.INT(n,attr) -> n
  | Ast.VAR(id,attr) ->
      (try lookup_id id env
      with LookupErrException _ ->
	(error(attr,
	       pr "unknown or non-constant identifier %s" (B.id2c id));
	 0))
  | Ast.UNARY(Ast.COMPLEMENT,exp,attr) ->
      (* not sure if the following is right *)
      lnot (all_int_constant env attr exp)
  | Ast.BINARY(bop,exp1,exp2,attr) ->
      (let v1 = all_int_constant env attr exp1
      and v2 = all_int_constant env attr exp2 in
      match bop with
	Ast.PLUS   -> v1 + v2
      | Ast.MINUS  -> v1 - v2
      | Ast.TIMES  -> v1 * v2
      | Ast.DIV    -> v1 / v2
      | Ast.MOD    -> v1 mod v2
      | Ast.BITAND -> v1 land v2
      | Ast.BITOR  -> v1 lor v2
      | Ast.LSHIFT -> v1 lsl v2
      | Ast.RSHIFT -> v1 lsr v2
      | _ ->
	  error(attr, "invalid operation in enum def or range bound");
	  0)
  | _ -> (error(attr,
		"invalid expression in enum def or range bound");
	  0)
	
(* ------------------ Information about structures ------------------ *)
    
let add_process_defn procdecls =
  Ast.no_dup_params procdecls "attribute";
  let idtys =
    List.map (function Ast.VARDECL(ty,id,imported,attr) -> (id,ty))
      procdecls in
  match !B.schedtype with
    B.PROCSCHED -> extend_struct_env (B.mkId("process"), (false, idtys))
  | B.VIRTSCHED -> extend_struct_env (B.mkId("scheduler"), (false, idtys))
	
let type_structure_field struct_type field attr
    (src_valid : bool) (tgt_valid : B.testval) expected =
  let (isconst,fldenv) =
    (try
      match struct_type with
	B.EVENT -> lookup_str "event" (!struct_env)
      | B.PROCESS ->
	  B.check_proc "process type";
	  if !in_persistent_timer_handler && not (tgt_valid = B.TRUE)
	  then expected_err attr "access to unchecked target" expected;
	  lookup_str "process" (!struct_env)
      | B.SRCPROCESS ->
	  if src_valid
	  then
	    (B.check_proc "process type"; lookup_str "process" (!struct_env))
	  else expected_err attr "access to field of unchecked source" expected
      | B.SCHEDULER ->
	  B.check_virt "scheduler type"; lookup_str "scheduler" (!struct_env)
      | B.STRUCT(nm) -> lookup_id nm (!struct_env)
      | B.TIMER(B.PERSISTENT(Some(nm))) -> lookup_id nm (!struct_env)
      | _ -> expected_err attr "field of non-structure type" expected
    with LookupErrException _ ->
      expected_err attr "unknown structure" expected) in
  (try
    let fldty = lookup_id field fldenv in
    (fldty,isconst)
  with LookupErrException _ ->
    expected_err attr ("unknown field "^(B.id2c(field))) expected)
    
(* ------------------ Information about functions ------------------- *)
    
let fun_env =
  (ref [] : (B.identifier * (B.typ * B.typ list)) list ref)
    
let init_fun_env _ =
  fun_env :=
    (if !B.high_res
    then [(B.mkId "make_time",(B.TIME,[B.INT;B.INT]));
	   (B.mkId "make_cycle_time",(B.TIME,[B.INT;B.CYCLES]));
	   (B.mkId "make_cycles",(B.CYCLES,[B.INT]));
	   (B.mkId "time_to_seconds",(B.INT,[B.TIME]));
	   (B.mkId "time_to_nanoseconds",(B.INT,[B.TIME]));
	   (B.mkId "time_to_jiffies",(B.INT,[B.TIME]));
	   (B.mkId "time_to_subjiffies",(B.CYCLES,[B.TIME]))]
    else []) @
    (if !B.schedtype = B.VIRTSCHED
    then [B.mkId "next_exists",(B.BOOL,[B.PROCESS])]
    else []) @
    ((B.mkId "print_trace_info",(B.VOID,[])) ::
     (B.mkId "now",((if (!B.high_res) then B.INDR(B.TIME) else B.TIME),[])) ::
     (B.mkId "time_to_ticks",(B.INT,[B.TIME])) ::
     (B.mkId "ticks_to_time",(B.TIME,[B.INT])) ::
     (B.mkId "start_absolute_timer",
      (B.VOID,[B.TIMER(B.NOT_PERSISTENT);B.TIME])) ::
     (B.mkId "start_relative_timer",
      (B.VOID,[B.TIMER(B.NOT_PERSISTENT);B.TIME])) ::
     (B.mkId "delete_bossa_timer",(*not in source language, added in pre_type*)
      (B.VOID,[B.TIMER(B.NOT_PERSISTENT)])) ::
     (B.mkId "stop_timer",(B.VOID,[B.TIMER(B.NOT_PERSISTENT)])) ::
     (B.mkId "get_user_int",(B.INT,[B.PORT])) ::
     (B.mkId "increment_ref_count",(B.VOID,[])) ::
     (B.mkId "decrement_ref_count",(B.VOID,[])) ::
     (B.sched_select [] [B.mkId "kfree",(B.VOID,[B.SCHEDULER])]))
      
(* return type cannot contain process or event *)
let rec add_fundecls fundecls =
  let names =
    List.map
      (function
	  Ast.FUNDECL(ret,nm,params,attr) ->
	    if Aux.member B.VOID params
	    then error(attr,"void not allowed as a parameter type");
	    if (not(valid_return_type ret))
	    then error(attr,
		       pr "invalid return type %s from function %s"
			 (B.type2c(ret)) (B.id2c(nm)));
	    nm)
      fundecls in
  if Aux.is_set ((List.map B.mkId (B.prim_names())) @ names)
  then fun_env :=
    (List.map
       (function Ast.FUNDECL(ret,nm,params,attr) -> (nm, (ret,params)))
       fundecls) @ (!fun_env)
  else nl_error "duplicate function name declarations"
      
and valid_return_type = function
    B.PROCESS -> false
  | B.SCHEDULER -> false
  | B.EVENT -> false
  | B.INDR(t) -> valid_return_type t
  | _ -> true
	
(* -------------------- Information about states -------------------- *)
	
let state_names = (ref [] : B.identifier list ref)
let prev_names = (ref [] : B.identifier list ref)
    
let type_states states env =
  
  (* In each class there must be exactly one public state.  This only
     applies to virtual schedulers. *)
  let check_public_private _ =
    let all_classes = CS.all_classes false in
    let needed_classes =
      List.fold_left
	(function needed_classes ->
	  function
	      Ast.QUEUE(clsname,queue_typ,id,CS.PRIVATE,attr) ->
		needed_classes
	    | Ast.PROCESS(clsname,id,previd,CS.PRIVATE,attr) ->
		needed_classes
	    | Ast.QUEUE(clsname,queue_typ,id,CS.PUBLIC,attr) ->
		if Aux.member clsname all_classes &&
		  Aux.member clsname needed_classes
		then Aux.remove clsname needed_classes
		else
		  (error(attr,
			 "at most one state can be declared public per class");
		   needed_classes)
	    | Ast.PROCESS(clsname,id,previd,CS.PUBLIC,attr) ->
		if Aux.member clsname all_classes &&
		  Aux.member clsname needed_classes
		then Aux.remove clsname needed_classes
		else
		  (error(attr,
			 "at most one state can be declared public per class");
		   needed_classes))
	all_classes states in
    match needed_classes with
      [] -> ()
    | _ ->
	nl_error "one state must be declared public for each class" in
  
  let new_state_names =
    List.map
      (function
	  Ast.QUEUE(clsname,queue_typ,id,vis,attr) -> id
        | Ast.PROCESS(clsname,id,previd,vis,attr) -> id)
      states in
  
  let ready_count =
    List.length
      (List.filter
	 (function
	     Ast.QUEUE(CS.READY,queue_typ,id,vis,attr) -> true
	   | Ast.PROCESS(CS.READY,id,previd,vis,attr) -> true
	   | _ -> false)
	 states) in
  
  (* to ignore NOWHERE.  This information is only used
     for checking the exhaustiveness of case labels *)
  let non_empty_state_names =
    List.fold_left
      (function prev ->
	function
	    Ast.QUEUE(CS.NOWHERE,_,id,vis,attr) -> prev
	  | Ast.QUEUE(CS.TERMINATED,CS.EMPTY,id,vis,attr) ->
	      B.sched_select (id::prev) prev
	  | Ast.QUEUE(clsname,CS.EMPTY,id,vis,attr) -> id::prev
	  | Ast.QUEUE(clsname,_,id,vis,attr) -> id::prev
          | Ast.PROCESS(clsname,id,previd,vis,attr) -> id::prev)
      [] states in
  
  let new_prev_names =
    List.fold_left
      (function acc ->
	function
            Ast.PROCESS(clsname,id,Ast.BOSSA_PREV previd,vis,attr) ->
	      previd::acc
          | Ast.PROCESS(clsname,id,Ast.RTS_PREV previd,vis,attr) ->
	      previd::acc
	  | _ -> acc)
      []
      states in
  
  let new_classes =
    List.map (function Ast.QUEUE(clsname,queue_typ,id,vis,attr) -> clsname
      | Ast.PROCESS(clsname,id,previd,vis,attr) -> clsname)
      states in
  
  let count_sorted _ =
    match
      (List.filter
	 (function
	     Ast.QUEUE(clsname,CS.SELECT(_),id,vis,attr) -> true
	   | _ -> false)
	 states)
    with
      [] ->
	(match
	  (List.filter
	     (function
		 Ast.QUEUE(CS.READY,queue_typ,id,vis,attr) -> true
	       | _ -> false)
	     states)
	with
	  [Ast.QUEUE(CS.READY,queue_typ,id,vis,attr)] ->
	    (let sort_order =
	      match queue_typ with
		CS.FIFO -> CS.FIFODEFAULT
	      |	_ -> CS.NODEFAULT in
	    CS.selected :=
	      Some(CS.STATE(id,CS.READY,CS.QUEUE(CS.SELECT(sort_order)),vis)))
	| _ -> raise (Error.Error("no SELECT queue and more than one "^
				  "READY queue")))
    | [Ast.QUEUE(CS.READY,CS.SELECT(x),id,vis,attr)] ->
	CS.selected := Some(CS.STATE(id,CS.READY,CS.QUEUE(CS.SELECT(x)),vis))
    | _ -> raise (Error.Error("only one SELECT queue allowed, "
			      ^ " and sorted queue must be READY"))
	  
  (* is is important that a state not be created more than once because
     of the interrupt_dest field *)
  and type_state = function
      Ast.QUEUE(clsname,CS.SELECT(x),id,vis,attr) ->
	let cur =
	  (match !CS.selected with
	    Some queue -> (id,queue)
	  | None -> raise (Error.Error "internal error")) in
	(CS.state_env := cur :: (!CS.state_env); [])
    | Ast.QUEUE(clsname,queue_typ,id,vis,attr) ->
	let cur =
	  if ready_count = 1 && clsname = CS.READY
	  then
	    (match !CS.selected with
	      Some queue -> (id,queue)
	    | None -> raise (Error.Error "internal error"))
	  else (id,CS.STATE(id,clsname,CS.QUEUE(queue_typ),vis)) in
	(CS.state_env := cur :: (!CS.state_env); [])
    | Ast.PROCESS(clsname,id,previd,vis,attr) ->
	(CS.state_env := (id,CS.STATE(id,clsname,CS.PROC,vis))
	  :: (!CS.state_env);
	 let ty = B.sched_select B.PROCESS B.SCHEDULER in
	 match previd with
	   Ast.BOSSA_PREV prev -> [(id,(ty,true));(prev,(ty,true))]
	 | Ast.RTS_PREV prev -> [(id,(ty,true));(prev,(ty,true))]
	 | Ast.NO_PREV -> [(id,(ty,true))])
	  
  in
  (count_sorted();
   if not(Aux.subset
	    (Aux.subtract (CS.all_classes false)
	       [CS.POLICY_BLOCKED;CS.KERNEL_POLICY_BLOCKED])
	    new_classes)
   then nl_error("at least one state must be declared for each class");
   if List.mem CS.KERNEL_POLICY_BLOCKED new_classes &&
     not (List.mem CS.POLICY_BLOCKED new_classes)
   then nl_error("if there is a state in the KERNEL_POLICY_BLOCKED state "^
		 "class, there must also be one in the POLICY_BLOCKED state "^
		 "class");
   if !B.schedtype = B.VIRTSCHED then check_public_private();
   if Aux.is_set new_state_names
   then (state_names := non_empty_state_names;
	 prev_names := new_prev_names;
	 List.fold_left
	   (function env -> function st -> (type_state st) @ env)
	   env states)
   else (nl_error("state names must be disjoint"); env))
    
(* ---------------------------- Coercion ---------------------------- *)
    
let coerce msg actual desired attr =
  match (actual,desired) with
    (B.INT,B.RANGE(_)) -> ()
  | (B.RANGE(_),B.INT) -> ()
  | (B.INDR(B.TIME),B.TIME) -> ()
  | (B.ENUM(x),B.ENUM(y)) ->
      if not(B.ideq(x,y))
      then expErr(desired,attr,
		  "cannot coerce "^B.type2c(actual)^" to "^B.type2c(desired)^
		  " in "^msg)
  | (B.NULL,B.SYSTEM(_)) -> ()
  | (B.NULL,B.PORT) -> ()
  | (B.NULL,B.STRUCT(_)) -> ()
  | (B.TIMER(B.PERSISTENT(_)),B.TIMER(B.NOT_PERSISTENT)) -> ()
  | (B.TIMER(B.PERSISTENT(Some(_))),B.TIMER(B.PERSISTENT(None))) -> ()
  | _ ->
      if not(actual = desired)
      then expErr(desired,attr,
		  "cannot coerce "^B.type2c(actual)^" to "^B.type2c(desired)^
		  " in "^msg)
	  
(* --------------------------- Expressions -------------------------- *)
	  
let rec type_exp
    (exp : Ast.expr)
    (env : (B.identifier * (B.typ * bool)) list)
    (expected : B.typ option)
    (src_valid : bool)
    (tgt_valid : B.testval) :
    B.typ * Ast.expr =
  try
    (match exp with
      Ast.INT(n,attr) ->
	(check_expected src_valid tgt_valid expected B.INT attr;
	 (B.INT, Ast.INT(n, B.updty attr B.INT)))
	  
    | Ast.VAR(id,attr) ->
	(try
	  let (ty,_) = lookup_id id env in
	  (check_expected src_valid tgt_valid expected ty attr;
	   (ty, Ast.VAR(id,B.updty attr ty)))
	with LookupErrException _ ->
	  expected_err attr ("unknown identifier "^(B.id2c(id))) expected)
	  
    | Ast.FIELD(exp,fld,attr) ->
	(* isconst is defined at the level of the structure *)
	let (ty,exp1) = type_exp exp env None src_valid tgt_valid in
	let (fldty,_) = type_structure_field ty fld attr src_valid tgt_valid
	    expected in
	(check_expected src_valid tgt_valid expected fldty attr;
	 (fldty,Ast.FIELD(exp1,fld,B.updty attr fldty)))
	  
    | Ast.BOOL(v,attr) ->
	(check_expected src_valid tgt_valid expected B.BOOL attr;
	 (B.BOOL, Ast.BOOL(v,B.updty attr B.BOOL)))
	  
    | Ast.UNARY(Ast.NOT,exp,attr) ->
	let (_,exp1) = type_exp exp env (Some(B.BOOL)) src_valid tgt_valid in
	(check_expected src_valid tgt_valid expected B.BOOL attr;
	 (B.BOOL, Ast.UNARY(Ast.NOT,exp1,B.updty attr B.BOOL)))
	  
    | Ast.UNARY(Ast.COMPLEMENT,exp,attr) ->
	let (_,exp1) = type_exp exp env (Some(B.INT)) src_valid tgt_valid in
	(check_expected src_valid tgt_valid expected B.INT attr;
	 (B.INT, Ast.UNARY(Ast.COMPLEMENT,exp1,B.updty attr B.INT)))
	  
    | Ast.BINARY(bop,exp1,exp2,attr) ->
	let (t1,exp1a) = type_exp exp1 env None src_valid tgt_valid
	and (t2,exp2a) = type_exp exp2 env None src_valid tgt_valid in
	let t1 = unrange (unsys (drop_indr_time attr (t1,!B.high_res)))
	and t2 = unrange (unsys (drop_indr_time attr (t2,!B.high_res))) in
	let res =
	  (match bop with
	    Ast.PLUS -> bin_check (t1,t2) attr [B.INT;B.TIME;B.CYCLES] "+"
	  | Ast.MINUS -> bin_check (t1,t2) attr [B.INT;B.TIME;B.CYCLES] "-"
	  | Ast.TIMES -> mul_time_int_op attr "*" (t1,t2)
	  | Ast.DIV -> div_time_int_op attr "/" (t1,t2)
	  | Ast.MOD -> div_time_int_op attr "%" (t1,t2)
	  | Ast.AND -> bin_check (t1,t2) attr [B.BOOL] "&&"
	  | Ast.OR -> bin_check (t1,t2) attr [B.BOOL] "||"
	  | Ast.BITAND -> bin_check (t1,t2) attr [B.INT] "&"
	  | Ast.BITOR -> bin_check (t1,t2) attr [B.INT] "|"
	  | Ast.EQ -> eq_check (t1,t2) attr "=="
	  | Ast.NEQ -> eq_check (t1,t2) attr "!="
	  | Ast.LT -> rel_check (t1,t2) attr "<"
	  | Ast.GT -> rel_check (t1,t2) attr ">"
	  | Ast.LEQ -> rel_check (t1,t2) attr "<="
	  | Ast.GEQ -> rel_check (t1,t2) attr ">="
	  | Ast.LSHIFT ->
	      bin_check_int (t1,t2) attr [B.CYCLES;B.TIME;B.INT] "<<"
	  | Ast.RSHIFT ->
	      bin_check_int (t1,t2) attr [B.CYCLES;B.TIME;B.INT] ">>")
	in (check_expected src_valid tgt_valid expected res attr;
	    (res, Ast.BINARY(bop,exp1a,exp2a,B.updty attr res)))

    | Ast.PBINARY(bop,exp1,states1,exp2,states2,attr) ->
	raise (Error.Error "internal error: pbinary unexpected")
	  
    | Ast.INDR(exp,attr) ->
	(match type_exp exp env None src_valid tgt_valid with
	  (B.INDR(t),exp1) ->
	    (check_expected src_valid tgt_valid expected t attr;
	     (t, Ast.INDR(exp1,B.updty attr t)))
	| _ -> expected_err attr "invalid indirection expression" expected)
	  
    | Ast.SELECT(_,attr) ->
	let ty = B.sched_select B.PROCESS B.SCHEDULER in
	(check_expected src_valid tgt_valid expected ty attr;
	 (ty, Ast.SELECT(!CS.selected, B.updty attr ty)))
	  
    | Ast.EMPTY(id,_,vl,insrc,attr) ->
	let concretized = CS.concretize id in
	(check_expected src_valid tgt_valid expected B.BOOL attr;
	 (B.BOOL, Ast.EMPTY(id,concretized,vl,insrc,B.updty attr B.BOOL)))
	  
    | Ast.PRIM(Ast.VAR(id,fn_attr),exps,attr) ->
	(try
	  let (ret,param_types) = lookup_id id (!fun_env) in
	  let exps1 =
	    List.map2
	      (function exp ->
		function ptype ->
		  let (ty,exp1) = type_exp exp env None src_valid tgt_valid in
		  coerce (pr "call to %s" (B.id2c id)) ty ptype attr;
		  exp1)
	      exps param_types in
	  let snm = B.id2c id in
	  if (snm = "start_absolute_timer" || snm = "start_relative_timer")
	      && !in_persistent_timer_handler && not (tgt_valid = B.TRUE)
	  then error(attr,
		     ("restarting timer is only valid when there is a "^
		      "target process"));
	  (ret,Ast.PRIM(Ast.VAR(id,fn_attr),exps1,B.updty attr ret))
	with
	  LookupErrException _ ->
	    expected_err attr (pr "unknown function %s" (B.id2c(id))) expected
	| Invalid_argument str ->
	    expected_err attr
	      (pr "%s\nwrong number of arguments to %s" str (B.id2c(id)))
	      expected)
	  
    | Ast.PRIM(_,exps,attr) ->
	expected_err attr (pr "unknown non-id function") expected
	  
    | Ast.SCHEDCHILD(exp,procs,attr) ->
	B.check_virt "next";
	if !in_handler
	then
	  let (ty,exp1) = type_exp exp env (Some(B.PROCESS)) src_valid tgt_valid in
	  (B.SCHEDULER,Ast.SCHEDCHILD(exp1,procs,B.updty attr B.SCHEDULER))
	else expErr(B.SCHEDULER,attr,"next only allowed in handlers")
	    
    | Ast.SRCONSCHED(attr) ->
	B.check_proc "srcOnSched";
	if !in_handler
	then (B.BOOL, Ast.SRCONSCHED(B.updty attr B.BOOL))
	else expErr(B.BOOL,attr,"srcOnSched only allowed in handlers")
	    
    | Ast.ALIVE(tv,insrc,attr) ->
	if !in_persistent_timer_handler
	then (B.BOOL,Ast.ALIVE(tgt_valid,insrc,B.updty attr B.BOOL))
	else expErr(B.SCHEDULER,attr,
		    "alive only allowed in persistent timer handlers")
	    
    | Ast.AREF(_,_,attr) -> raise (Error.Error "type: internal error")
	  
    | Ast.IN(exp,id,_,vl,insrc,attr) ->
	let (t1,exp1) = type_exp exp env None src_valid tgt_valid in
	(check_expected src_valid tgt_valid expected B.BOOL attr;
	 if t1 = (B.sched_select B.PROCESS B.SCHEDULER)
	     or t1 = (B.sched_select B.SRCPROCESS B.SCHEDULER)
	 then
	   (B.BOOL,
	    Ast.IN(exp1,id,CS.concretize id,vl,insrc,B.updty attr B.BOOL))
	 else expErr(B.BOOL,attr,"invalid argument to in"))
	  
    | Ast.MOVEFWDEXP(_,_,_) -> raise (Error.Error "type: internal error"))
  with ExpErrException x -> (x,exp)
      
and err res attr op expected a1 a2 =
  expErr(res,attr,
	 pr "invalid arguments to %s:\n  expected %s, found (%s,%s)"
	   op expected (B.type2c(a1)) (B.type2c(a2)))
    
and mul_time_int_op attr op = function
    (B.INT,B.TIME) -> B.TIME
  | x -> bin_check_int x attr [B.TIME;B.INT] op
	
and div_time_int_op attr op = function
    (B.TIME,B.TIME) -> B.INT
  | (B.CYCLES,B.CYCLES) -> B.INT
  | x -> bin_check_int x attr [B.CYCLES;B.TIME;B.INT] op
	
and drop_indr_time attr = function
    (B.INDR(B.TIME),true) -> B.TIME
  | (B.INDR(B.TIME),false) -> expErr(B.TIME,attr,"invalid type time*")
  | (t,_) -> t
	
and bin_check (a1,a2) attr ts op =
  match List.filter (function t -> a1 = t && a2 = t) ts with
    [res] -> res
  | _ ->
      let choose_res =
	try List.find (function t -> a1 = t) ts with Not_found ->
	  try List.find (function t -> a2 = t) ts with Not_found ->
	    List.hd ts in
      let type_strings =
	List.map (function t -> pr "(%s,%s)" (B.type2c(t)) (B.type2c(t))) ts in
      expErr(choose_res,attr,
	     pr "invalid arguments to %s:\n  expected %s, found (%s,%s)"
	       op (Aux.set2c type_strings) (B.type2c(a1)) (B.type2c(a2)))
	
and bin_check_int (a1,a2) attr ts op =
  match List.filter (function t -> a1 = t && a2 = B.INT) ts with
    [res] -> res
  | _ ->
      let type_strings =
	List.map (function t -> pr "(%s,int)" (B.type2c(t))) ts in
      expErr(a1,attr,
	     pr "invalid arguments to %s:\n  expected %s, found (%s,%s)"
	       op (Aux.set2c type_strings) (B.type2c(a1)) (B.type2c(a2)))
	
and eq_check args attr op =
  match args with
    (B.ENUM(x),B.ENUM(y)) ->
      if (B.ideq(x,y))
      then B.BOOL
      else bop_error B.BOOL attr op args
  | (B.STRUCT(x),B.STRUCT(y)) ->
      if (B.ideq(x,y))
      then B.BOOL
      else bop_error B.BOOL attr op args
  | (B.STRUCT(_),B.NULL) | (B.NULL,B.STRUCT(_))
  | (B.PORT,B.NULL) | (B.NULL,B.PORT)| (B.INT,B.INT)
  | (B.BOOL,B.BOOL) | (B.TIME,B.TIME) | (B.PROCESS,B.PROCESS)
  | (B.SCHEDULER,B.SCHEDULER) -> B.BOOL
  | (_,_) -> bop_error B.BOOL attr op args
	
and rel_check args attr op =
  match args with
    (B.INT,B.INT) -> B.BOOL
  | (B.TIME,B.TIME) -> B.BOOL
  | (B.CYCLES,B.CYCLES) -> B.BOOL
  | (B.PROCESS,B.PROCESS) -> B.BOOL
  | (B.SRCPROCESS,B.PROCESS) -> B.BOOL
  | (B.PROCESS,B.SRCPROCESS) -> B.BOOL
  | (B.SRCPROCESS,B.SRCPROCESS) -> B.BOOL
  | (B.SCHEDULER,B.SCHEDULER) -> B.BOOL
  | (B.ENUM(x),B.ENUM(y)) ->
      if (B.ideq(x,y))
      then B.BOOL
      else bop_error B.BOOL attr op args
  | (_,_) -> bop_error B.BOOL attr op args
	
and bop_error res attr op (arg1,arg2) =
  expErr(res,attr,
	 pr "invalid arguments to %s: (%s,%s)"
	   op (B.type2c(arg1)) (B.type2c(arg2)))
    
and check_expected src_valid tgt_valid expected res attr =
  match expected with
    None -> ()
  | Some(t) ->
      match(t,unrange (unsys res)) with
	(B.ENUM(x),B.ENUM(y)) ->
	  if not(B.ideq(x,y))
	  then expErr(t,attr,"expected "^B.type2c(t)^" found "^B.type2c(res))
      |	(B.PROCESS,B.SRCPROCESS) ->
	  if not src_valid
	  then expErr(t,attr,"source process not valid")
      | (B.NULL,B.SYSTEM(_)) -> ()
      |	(a,b) ->
	  if not(a = b)
	  then expErr(t,attr,"expected "^B.type2c(t)^" found "^B.type2c(res))
	      
and type_state_exp exp =
  match exp with
    Ast.VAR(nm,attr) ->
      (try Some(lookup_id nm (!CS.state_env))
      with LookupErrException _ -> None)
  | _ -> None
	
and unrange = function
    B.RANGE(_) -> B.INT
  | t -> t
	
and unsys = function
    B.SYSTEM(t) -> t
  | t -> t
	
(* -------------------------- Declarations ------------------------- *)
	
let type_decl (Ast.VARDECL(ty,id,imported,attr)) = (id,(ty,false))

let rec type_def d env src_valid tgt_valid =
  match d with
    Ast.VALDEF(decl,exp,isconst,attr) ->
      let (id,(ty,_)) = type_decl decl in
      let (t1,exp1) = type_exp exp env None src_valid tgt_valid in
      (match (ty,isconst) with
	(B.TIMER(_),Ast.CONST) -> ()
      |	(B.TIMER(_),_) ->
	  error(attr, "timer typed declaration must be constant")
      |	_ -> ());
      (coerce (pr "initialization of %s" (B.id2c id)) t1 ty attr;
       ((id,(ty,isconst=Ast.CONST)),
	Ast.VALDEF(decl,exp1,isconst,B.updty attr ty)))
  | Ast.UNINITDEF(decl,attr) ->
      let (id,(ty,_)) = type_decl decl in
      ((id,(ty,false)),Ast.UNINITDEF(decl,B.updty attr ty))
  | Ast.SYSDEF(decl,isconst,attr) ->
      let (id,(ty,_)) = type_decl decl in
      (match (ty,isconst) with
	(B.TIMER(_),Ast.CONST) -> ()
      |	(B.TIMER(_),_) ->
	  error(attr, "timer typed declaration must be constant")
      |	_ -> ());
      ((id,(ty,isconst=Ast.CONST)),Ast.SYSDEF(decl,isconst,B.updty attr ty))
  | Ast.DUMMYDEF(id,exp,attr) ->
      let (t1,exp1) = type_exp exp env None src_valid tgt_valid in
      ((id,(t1,true)),
       Ast.VALDEF(Ast.VARDECL(t1,id,false,attr),exp1,
		  Ast.CONST,B.updty attr t1))
	
let type_defs defs env src_valid tgt_valid =
  let (new_env,new_defs) =
    List.fold_left
      (function (new_env,new_defs) ->
	function x ->
	  let (env_elem, def_elem) = type_def x new_env src_valid tgt_valid in
	  (env_elem :: new_env, def_elem :: new_defs))
      (env,[]) defs in
  (List.rev new_env,List.rev new_defs)
    
(* the following assumes that the environment is always extended with
   extend_env, which just conses new things onto the front, making an
   environment usable as a record of all the names that have been declared *)
let rec check_globals = function
    [] -> ()
  | (x,_)::rest ->
      (try let _ = List.assoc x rest in
      nl_error (pr "variable %s appears more than once among global variables, enumerated types, and state names" (B.id2c x))
      with Not_found -> ());
      check_globals rest
	
(* ----------- Extra treatment for constant declarations ----------- *)
	
let constant_values defs =
  List.fold_left
    (function const_env ->
      function
	  Ast.VALDEF(Ast.VARDECL(B.INT,id,imported,attr),exp,Ast.CONST,_) ->
	    (id,all_int_constant const_env attr exp) :: const_env
	| Ast.VALDEF(Ast.VARDECL(_,id,imported,attr),exp,Ast.CONST,_) ->
	    const_env
	| _ -> nl_error("unexpected non-constant declaration"); const_env)
    [] defs
    
(* --------------------------- Statements -------------------------- *)
    
let rec type_stmt
    (stmt : Ast.stmt)
    (env : (B.identifier * (B.typ * bool)) list)
    (ret : B.typ option)
    (src_valid : bool)
    (tgt_valid : B.testval)
    (break_valid : bool)
    : Ast.stmt * bool (* contains break *) =
  try
    (match stmt with
      Ast.IF(exp,stmt1,stmt2,attr) ->
	let (_,exp1) = type_exp exp env (Some(B.BOOL)) src_valid tgt_valid in
	let new_src_valid =
	  (match exp with Ast.SRCONSCHED(attr) -> true | _ -> false) in
	let (true_tgt_valid,false_tgt_valid) =
	  (match exp with
	    Ast.ALIVE(_,_,_) -> (B.TRUE,B.FALSE)
	  | _ -> (tgt_valid,tgt_valid)) in
	let (stmt1a,cb1) =
	  type_stmt stmt1 env ret (src_valid || new_src_valid)
	    true_tgt_valid break_valid in
	let (stmt2a,cb2) =
	  type_stmt stmt2 env ret (not new_src_valid && src_valid)
	    false_tgt_valid break_valid in
	(Ast.IF(exp1,stmt1a,stmt2a,attr),cb1||cb2)
	  
    | Ast.FOR_WRAPPER(label,stmts,attr) -> raise (Error.Error "cannot occur")
	  
    | Ast.FOR(id,state_ids,_,dir,stmt,attr) ->
	(* state_id can be either a state or a class or missing, and can
	   refer to process variables or queues *)
	let state_info =
	  match state_ids with
	    Some(stids) -> Aux.flat_map CS.concretize stids
	  | None -> CS.all_states() in
	let new_env =
	  extend_env id (B.sched_select B.PROCESS B.SCHEDULER,false) env in
	let (stmt1,cb) = type_stmt stmt new_env ret src_valid tgt_valid true in
	(* duplicate code for each state *)
	let loops =
	  List.map
	    (function
		CS.STATE(state_id,_,CS.PROC,_) as info ->
		  let pattr =
		    B.updty attr (B.sched_select B.PROCESS B.SCHEDULER) in
		  let expr = Ast.VAR(state_id,pattr) in
		  let vardecl =
		    Ast.VARDECL(B.sched_select B.PROCESS B.SCHEDULER,
				id,false,pattr) in
		  Ast.IF(Ast.EMPTY(state_id,[info],B.UNK,true,
				   B.updtst (B.updty attr B.BOOL)
				     (Ast.mk_test_index())),
			 Ast.SEQ([],[],attr),
			 Ast.SEQ([Ast.VALDEF(vardecl,expr,Ast.VARIABLE,attr)],
				 [stmt1],attr),
			 attr)
	      | CS.STATE(state_id,_,CS.QUEUE(_),_) as info ->
		  Ast.FOR(id,Some [state_id],[info],dir,stmt1,attr))
	    state_info in
	if cb
	then (Ast.FOR_WRAPPER(B.fresh_idx "break",loops,attr), false)
	else (Ast.SEQ([],loops,attr),false)
	    
    | Ast.SWITCH(exp,cases,default,attr) ->
	let (_,exp1) =
	  type_exp exp env (Some(B.sched_select B.PROCESS B.SCHEDULER))
	    src_valid tgt_valid in
	let cases_info = check_cases cases attr in
	let (new_cases,cbs) =
	  List.split
	    (List.map2
	       (function Ast.SEQ_CASE(pat,_,stmt,attr) ->
		 function case_info ->
		   let (new_stmt,cb) =
		     type_stmt stmt env ret src_valid tgt_valid break_valid in
		   (Ast.SEQ_CASE(pat,case_info,new_stmt,attr), cb))
	       cases cases_info) in
	let (new_default,cb) =
	  (match default with
	    None -> (None,false)
	  | Some x ->
	      let (new_default,cb) =
		type_stmt x env ret src_valid tgt_valid break_valid in
	      ((Some new_default),cb)) in
	(Ast.SWITCH(exp1,new_cases,new_default,attr),
	 List.exists (function x -> x) (cb::cbs))
	  
    | Ast.SEQ(decls,stmts,attr) ->
	let (new_env,new_decls) = type_defs decls env src_valid tgt_valid in
	let (new_stmts,cb) =
	  List.fold_left
	    (function (new_stmts,cbs) ->
	      function stmt ->
		let (new_stmt,cb) =
		  type_stmt stmt new_env ret src_valid tgt_valid break_valid in
		(new_stmt :: new_stmts, cb||cbs))
	    ([],false) stmts in
	(Ast.SEQ(new_decls,List.rev new_stmts,attr), cb)
	  
    | Ast.RETURN(exp,attr) ->
	(match (ret,exp) with
	  (Some(B.VOID), Some(e)) ->
	    stmtErr(attr,
		    "no return expression allowed in void-typed function")
	| (Some(t), Some(e)) ->
	    let (ty,e1) = type_exp e env ret src_valid tgt_valid in
	    (coerce "return" ty t attr;
	     (Ast.RETURN(Some(e1),attr),false))
	| (None,Some(e)) ->
	    let (ty,e1) =
	      type_exp e env (Some(B.ENUM(B.mkId("SchedState"))))
		src_valid tgt_valid in
	    (Ast.RETURN(Some(e1),attr),false)
	| (Some(B.VOID),None) -> (Ast.RETURN(exp,attr),false)
	| (Some(t),None) ->
	    stmtErr(attr,pr "missing return expression %s" (B.type2c t))
	| (None,None) -> (Ast.RETURN(exp,attr),false))
	  
    | Ast.MOVE(exp,state,_,_,_,auto_allowed,state_end,attr) ->
	(try
	  let (CS.STATE(_,cldst,state_type,_) as dst) =
	    lookup_id state (!CS.state_env) in
	  (match state_end with
	    Ast.HEAD
	  | Ast.TAIL ->
	      (match state_type with
		CS.QUEUE(CS.SELECT(CS.LIFODEFAULT))
	      |	CS.QUEUE(CS.SELECT(CS.FIFODEFAULT)) -> ()
	      |	_ -> stmtErr(attr,"invalid head or tail reference"))
	  | Ast.EITHER -> ());
	  if !move_allowed
	  then
	    let (_,exp1) =
	      type_exp exp env (Some(B.sched_select B.PROCESS B.SCHEDULER))
		src_valid tgt_valid in
	    (match type_state_exp exp with
	      Some((CS.STATE(srcid,clsrc,_,_) as src)) ->
		if srcid = state
		then stmtErr(attr,
			     pr "invalid move: %s => %s"
			       (B.id2c srcid) (B.id2c state))
		else (Ast.MOVE(exp1,state,Some src,[src],Some dst,
			       auto_allowed,state_end,attr),
		      false)
	    | None -> (Ast.MOVE(exp1,state,None,[],Some dst,
				auto_allowed,state_end,attr),
		       false))
	  else stmtErr(attr,"move not allowed")
	with LookupErrException _ ->
	  stmtErr(attr,pr "destination %s must be a state" (B.id2c(state))))
	  
    | Ast.MOVEFWD(exp,_,state_end,attr) ->
	if !in_handler
	then
	  let (_,exp1) =
	    type_exp exp env (Some(B.sched_select B.PROCESS B.SCHEDULER))
	      src_valid tgt_valid in
	  (Ast.MOVEFWD(exp1,[],state_end,attr),false)
	else
	  stmtErr(attr,"forwardImmediate not allowed in interface functions")
	    
    | Ast.SAFEMOVEFWD(exp,_,stm,state_end,attr) ->
	if !in_handler
	then
	  let (_,exp1) =
	    type_exp exp env (Some(B.sched_select B.PROCESS B.SCHEDULER))
	      src_valid tgt_valid in
	  let (stmta,cb) =
	    type_stmt stm env ret src_valid tgt_valid break_valid in
	  (Ast.SAFEMOVEFWD(exp1,[],stmta,state_end,attr),cb)
	else
	  stmtErr(attr,"forwardImmediate not allowed in interface functions")
	    
    | Ast.ASSIGN(expl,expr,sorted_fld,attr) ->
	let ((ty,isconst),expl1) = type_loc env src_valid tgt_valid expl
	and (actual,expr1) = type_exp expr env None src_valid tgt_valid in
	if isconst
	then stmtErr(attr,"constant left hand side of assignment")
	else (coerce "assignment" actual ty attr;
	      (Ast.ASSIGN(expl1,expr1,sorted_fld,attr),false))
	    
    | Ast.DEFER(attr) -> (stmt,false)
	  
    | Ast.BREAK(attr) ->
	if break_valid
	then (stmt,true)
	else stmtErr(attr,"unexpected break")
	    
    | Ast.PRIMSTMT(Ast.VAR(id,fn_attr),exps,attr) ->
	(try
	  let (ret,param_types) = lookup_id id (!fun_env) in
	  let exps1 =
	    List.map2
	      (function exp ->
		function ptype ->
		  let ptype =
		    (* not clear why this is done, but check_expected
		       drops the SYSTEM from a type before comparing it *)
		    match ptype with
		      B.SYSTEM(t) -> t
		    | t -> t in
		  let (ty,exp1) = type_exp exp env None src_valid tgt_valid in
		  coerce (pr "call to %s" (B.id2c id)) ty ptype attr;
		  exp1)
	      exps param_types in
	  (Ast.PRIMSTMT(Ast.VAR(id,fn_attr),exps1,attr),false)
	with
	  LookupErrException _ ->
	    stmtErr(attr,(pr "unknown function %s" (B.id2c id)))
	| Invalid_argument str ->
	    stmtErr(attr,
		    pr "%s\nwrong number of arguments to %s" str (B.id2c(id))))
	  
    | Ast.PRIMSTMT(_,exps,attr) -> stmtErr(attr,"unknown function")
	  
    | Ast.ERROR(str,attr) -> (stmt,false)
	  
    | Ast.ASSERT(exp,_) -> raise (Error.Error "type: internal error"))
  with
    StmtErrException -> (stmt,false)
  | ExpErrException _ -> (stmt,false)
	
and check_cases cases attr =
  let (remaining_states,line_labels) =
    List.fold_left
      (function (remaining_states,line_labels) ->
	function (Ast.SEQ_CASE(lbls,_,_,_)) ->
	  let (remaining_states,local_line_labels) =
	    List.fold_left
	      (function (remaining_states,line_labels) ->
		function label ->
		  let conc_states = CS.concretize label in
		  let inter = Aux.intersect conc_states remaining_states in
		  match inter with
		    [] -> stmtErr(attr,
				  pr "switch: case label %s doesn't match any untreated states"
				    (B.id2c label))
		  | _ -> (Aux.subtract remaining_states conc_states,
			  inter@line_labels))
	      (remaining_states,[]) lbls in
	    (remaining_states,(List.rev local_line_labels)::line_labels))
      ((List.map (function (_,s) -> s) (!CS.state_env)),[])
      cases in
  let line_labels = List.rev line_labels in
  let all_states =
    List.map (function (CS.STATE(id,_,_,_)) -> id)
      (List.flatten line_labels) in
  if (Aux.is_set all_states)
  then
    (match Aux.subset_fail (!state_names) all_states with
      None ->
	(match Aux.subset_fail all_states (!state_names) with
	  None -> line_labels
	| Some(extra_state) ->
	    stmtErr(attr,
		    pr "switch: the case label %s is not a state name"
		      (B.id2c extra_state)))
    | Some(missing_state) ->
	stmtErr(attr,
		pr "switch: the state %s is not among the case labels"
		  (B.id2c missing_state)))
  else stmtErr(attr,"switch: state names cannot be repeated")
      
and type_loc env src_valid tgt_valid = function
    Ast.VAR(id,attr) ->
      (try
	let ((ty,isconst) as x) = lookup_id id env in
	(x,Ast.VAR(id,B.updty attr ty))
      with LookupErrException _ ->
	stmtErr(attr,pr "unknown location %s" (B.id2c(id))))
  | Ast.FIELD(exp,fld,attr) ->
      let (ty,exp1) = type_exp exp env None src_valid tgt_valid in
      let ((fldty,isconst) as x) =
	type_structure_field ty fld attr src_valid tgt_valid None in
      (match fldty with
	B.TIMER(_) -> stmtErr(attr,"cannot assign to a timer-typed field")
      |	_ -> ());
      (x,Ast.FIELD(exp1,fld,B.updty attr fldty))
  | _ -> raise (Error.Error("type: internal error"))
	
(* ------------------------ Ordering criteria ----------------------- *)
	
let rec verify_criteria criteria env enum_range_env =
  let pname = B.mkId(B.sched_select "process" "scheduler") in
  let fenv =
    let (_,fenv) = lookup_id pname (!struct_env) in
    List.map (function (id,vl) -> (id,(vl,true))) fenv in
  let flds = List.map (function (id,_) -> id) fenv in
  let enum_or_range id env =
    (try
      (match lookup_id id fenv with
	(B.ENUM(e_id) as ty,_)  -> (lookup_id e_id enum_range_env,ty)
      | (B.RANGE(r_id) as ty,_) -> (lookup_id r_id enum_range_env,ty)
      | _ ->
	  (nl_error(
	   pr "key criterion %s must have enum or range type"
	     (B.id2c id));
	   ((0,0),B.VOID)))
    with LookupErrException _ ->
      (nl_error(pr "invalid criterion %s" (B.id2c id)); ((0,0),B.VOID))) in
  let (_,l) =
    List.fold_left
      (function (non_key,prev) ->
	function
	    Ast.CRIT_ID(Ast.KEY(_,_),order,id,attr) ->
	      if non_key
	      then error(attr,
			 "key must only appear in a prefix of the "^
			 "ordering criteria");
	      let ((mn,mx),ty) = enum_or_range id fenv in
	      let newkey = Ast.KEY(mn,mx) in
	      (non_key, Ast.CRIT_ID(newkey,order,id,B.updty attr ty) :: prev)
	  | Ast.CRIT_ID(Ast.NOKEY,order,id,attr) as x ->
	      (try
		(match lookup_id id fenv with
		  (ty,_)  ->
		    let typed = Ast.CRIT_ID(Ast.NOKEY,order,id,B.updty attr ty)
		    in (true,typed::prev))
	      with LookupErrException _ ->
		(error(attr,pr "invalid criterion %s" (B.id2c id));
		 (true,x::prev)))
	  | Ast.CRIT_TEST(Ast.KEY(_,_),order,tst,thn,els,attr) ->
	      if non_key
	      then error(attr,
			 "key must only appear in a prefix of the "^
			 "ordering criteria");
	      let exp_env = fenv @ env in
	      let (_,tst1) = type_exp tst exp_env (Some B.BOOL) false B.BOTH
	  (* the following two lines are a bit redundant, but they initialize
	     the attr field correctly *)
	      and (_,thn1) = type_exp thn exp_env None false B.BOTH
	      and (_,els1) = type_exp els exp_env None false B.BOTH in
	      let cstorrange = function
		  Ast.INT(0,_) -> ((0,1),B.INT)
		| Ast.INT(1,_) -> ((0,1),B.INT)
		| Ast.VAR(id,attr) -> enum_or_range id exp_env
		| _ ->
		    (error(attr,
			   "key criterion must be 0, 1, or a variable of "^
			   "enum or range type");
		     ((0,0),B.VOID)) in
	      let ((m1,m2),ty) = cstorrange thn
	      and ((n1,n2),ty1) = cstorrange els in
	      if not (ty = ty1)
	      then error(attr,"inconsistent types");
	      let newkey =
		Ast.KEY((if m1 < n1 then m1 else n1),
			(if m2 > n2 then m2 else n2)) in
	      (non_key,
	       Ast.CRIT_TEST(newkey,order,tst1,thn1,els1,
			     (* ty doesn't matter here - type of thn1 used *)
			     B.updty attr ty)
	       :: prev)
	  | Ast.CRIT_TEST(Ast.NOKEY,order,tst,thn,els,attr) ->
	      let exp_env = fenv @ env in
	      let (_,tst1) = type_exp tst exp_env (Some B.BOOL) false B.BOTH
	      and (thty,thn1) =
		type_exp thn exp_env None (*(Some B.INT)*) false B.BOTH
	      and (elty,els1) =
		type_exp els exp_env None (*(Some B.INT)*) false B.BOTH in
	      let _ =
		rel_check (thty,elty) attr
		  (match order with
		    Ast.HIGHEST -> "highest"
		  | Ast.LOWEST -> "lowest")in
	      (true,Ast.CRIT_TEST(Ast.NOKEY,order,tst1,thn1,els1,attr)::prev))
      (false,[]) criteria in
  List.rev l
    
(* ------------------------------ Trace ----------------------------- *)
    
let get_id = function
    Ast.TRACEVAR(fldid,id,attr) -> (id,attr)
  | _ -> raise (Error.Error "internal error")
	
let verify_trace_expressions env ids =
  let pname = B.mkId(B.sched_select "process" "scheduler") in
  let fenv = let (_,fenv) = lookup_id pname (!struct_env) in fenv in
  List.map
    (function idexp ->
      let (id,attr) = get_id idexp in
      if Aux.member id (!state_names)
      then 
	(try
	  let (ty,_) = lookup_id id env in
	  Ast.TRACEVAR(B.mkId((B.id2c id)^"_data"),id,B.updty attr ty)
	with LookupErrException _ ->
	  Ast.TRACEQUEUE(B.mkId((B.id2c id)^"_data"),id,B.updty attr B.INT))
      else
	try
	  let (ty,_) = lookup_id id env in
	  Ast.TRACEVAR(id,id,B.updty attr ty)
	with
	  LookupErrException _ ->
	    (try
	      let ty = lookup_id id fenv in
	      Ast.TRACEFIELD(id,B.updty attr ty)
	    with
	      LookupErrException _ ->
		(error(attr,"unknown traced variable");
		 Ast.TRACEVAR(id,id,attr))))
    ids
    
let verify_trace env handlers = function
    Ast.NOTRACE -> Ast.NOTRACE
  | Ast.PRETRACE(count,info,attr) ->
      let seen_events = ref (None : Ast.event_name list option) in
      let seen_expressions = ref (None : Ast.classified_exp list option) in
      let seen_test = ref (None : Ast.expr option) in
      List.iter
	(function
	    Ast.TRACEEVENTS(events) ->
	      (match !seen_events with
		None ->
		  let event_names =
		    List.map (function Ast.EVENT(nm,_,_,_,_) -> nm) handlers in
		  List.iter
		    (function event ->
		      if not (List.exists (Ast.event_name_prefix event)
				event_names)
		      then error(attr,
				 pr "traced event %s not declared"
				   (Ast.event_name2c event)))
		    events;
		  seen_events := Some events;
	      |	Some _ ->
		  error(attr, "duplicate events declaration"))
	  | Ast.TRACEEXPS(ids) ->
	      (match !seen_expressions with
		None ->
		  let new_exps = verify_trace_expressions env ids in
		  seen_expressions := Some new_exps;
	      |	Some _ ->
		  error(attr, "duplicate expression declaration"))
	  | Ast.TRACETEST(exp) ->
	      (match !seen_test with
		None ->
		  let (_,exp) = type_exp exp env (Some B.BOOL) false B.BOTH in
		  seen_test := Some exp
	      |	Some _ ->
		  error(attr, "duplicate test declaration")))
	info;
      Ast.TRACE(count,
		(match !seen_events with None -> [] | Some x -> x),
		(match !seen_expressions with None -> [] | Some x -> x),
		!seen_test,
		attr)
  | _ -> raise (Error.Error "internal error")
	
(* -------------------------- Check attach -------------------------- *)
	
(* Check that all non-system process fields are initialized.  Check that
   there is no state change or assignment to globals before an error *)
	
let check_attach params procdecls stmt =
  let decls = List.map (function (Ast.VARDECL(_,id,_,attr)) -> id) params in
  let (Ast.VARDECL(_,id,_,attr)) = List.hd params in (* attached process *)
  let procdecls =
    Aux.option_filter
      (function
	  Ast.VARDECL(_,id,true,_) -> None
	| Ast.VARDECL(B.TIMER(_),id,false,_) -> None
	| Ast.VARDECL(_,id,false,_) -> Some id)
      procdecls in
  let rec loop state_change decls procdecls = function
      Ast.IF(tst,thn,els,a) ->
	let (state_change1,procdecls1) =
	  loop state_change decls procdecls thn in
	let (state_change2,procdecls2) =
	  loop state_change decls procdecls els in
	(state_change1 || state_change2, Aux.union procdecls1 procdecls2)
    | Ast.FOR_WRAPPER(_,stmt,a) -> raise (Error.Error "unexpected for wrapper")
    | Ast.FOR(id,_,_,_,stmt,a) ->
	let (state_change1,procdecls1) =
	  loop state_change (id::decls) procdecls stmt in
	(* two iterations should be enough, since the domain is boolean *)
	if not (state_change = state_change1)
	then loop state_change1 (id::decls) procdecls1 stmt
	else (state_change1,procdecls1)
    | Ast.SWITCH(_,cases,default,a) ->
	let (state_changes,procdeclss) =
	  List.split
	    (List.map
	       (function Ast.SEQ_CASE(_,_,stmt,_) ->
		 loop state_change decls procdecls stmt)
	       cases) in
	let (state_change2,procdecls2) =
	  (match default with
	    None -> (false, [])
	  | Some default -> loop state_change decls procdecls default) in
	(List.mem true (state_change2 :: state_changes),
	 List.fold_left Aux.union procdecls2 procdeclss)
    | Ast.SEQ(ldecls,stmts,a) ->
	List.fold_left
	  (function (state_change,procdecls) ->
	    function stm ->
	      let (state_change1,procdecls1) =
		loop state_change decls procdecls stm in
	      (state_change || state_change1,procdecls1))
	  (state_change,procdecls) stmts
    | Ast.RETURN(_,a) -> (false,procdecls)
    | Ast.MOVE(_,_,_,_,_,_,_,a) -> (true,procdecls)
    | Ast.MOVEFWD(_,_,_,a) -> (true,procdecls)
    | Ast.SAFEMOVEFWD(_,_,_,_,a) -> (true,procdecls)
    | Ast.ASSIGN(Ast.VAR(id1,_),_,_,a) when not(List.mem id1 decls) ->
	(true,procdecls)
    | Ast.ASSIGN(Ast.FIELD(Ast.VAR(id1,_),fld,_),_,_,a) when B.ideq(id,id1) ->
	let procdecls =
	  List.filter (function fld1 -> not (fld = fld1)) procdecls in
	(state_change,procdecls)
    | Ast.ERROR(_,a) ->
	if state_change
	then error(a,"error not allowed after state change in attach");
	(false,[])
    | _ -> (state_change,procdecls) in
  match loop false decls procdecls stmt with
    (_,[]) -> ()
  | (_,l) ->
      error(attr,
	    (Printf.sprintf
	       "attach does not set some fields of the process parameter %s\n"
	       (B.id2c id)) ^
	    (Printf.sprintf "missing fields are %s\n"
	       (Aux.set2c (List.map B.id2c l)))^
	    "note that the initialization must use the name of the\n"^
	    "process parameter explicitly\n")
	
(* ----------------------- Checking for admit ----------------------- *)
	
let rec check_no_return = function
    Ast.IF(exp,stmt1,stmt2,attr) ->
      check_no_return stmt1; check_no_return stmt2
	
  | Ast.FOR_WRAPPER(label,stmts,attr) -> raise (Error.Error "cannot occur")
	
  | Ast.FOR(id,state_ids,_,dir,stmt,attr) -> check_no_return stmt
	
  | Ast.SWITCH(exp,cases,default,attr) ->
      List.iter
	(function Ast.SEQ_CASE(pat,_,stmt,attr) -> check_no_return stmt)
	cases;
      (match default with
	None -> ()
      | Some s -> check_no_return s)
	    
  | Ast.SEQ(decls,stmts,attr) -> List.iter check_no_return stmts
	    
  | Ast.RETURN(exp,attr) -> error (attr, "return not allowed")
	    
  | Ast.SAFEMOVEFWD(exp,_,stm,state_end,attr) -> check_no_return stm
	    
  | Ast.ERROR(str,attr) -> error (attr, "error not allowed")
	    
  | s -> ()
	    
(* --------------------- Handlers and functions --------------------- *)
	    
(* There should be no overlap among the event names.  For the required
   names, either the name should be present exactly, or a star should be
   present on a shorter name. Star not required if all longer-named variants
   present. *)
	    
(* there should only be process.new.initial_process if default is true *)
	    
let rec check_events handlers =
  let declared_names = List.map (function Ast.EVENT(nm,_,_,_,_) -> nm) handlers
  and required_names = !Events.all_event_names in
  if not(Aux.is_set
	   (List.map (function Ast.EVENT_NAME(nm,star,_) -> (nm,star))
	      declared_names))
  then nl_error("some event handlers duplicated");
  let rec loop required skipped prev = function
      [] ->
	List.iter
	  (function nm ->
	    nl_error( "missing required name: "^(Ast.event_name2c nm)))
	  (required @ skipped)
    | (Ast.EVENT_NAME(dnms,dstar,_) as d)::ds ->
	(* skip over things subsumed by * *)
	if dstar
	then
	  let new_required = skip dnms required
	  and new_skipped = skip dnms skipped in
	  (* test if there was any need for the * case *)
	  if required = new_required && skipped = new_skipped
	  then nl_error("extra name provided: "^ (Ast.event_name2c d));
	  loop new_required new_skipped None ds
	else
	  (match required with
	    [] ->
	      nl_error("extra names provided: "^
		       (String.concat " " (List.map Ast.event_name2c (d::ds))));
	      Printf.printf "missing required names: %s\n"
		(String.concat " " (List.map Ast.event_name2c skipped))
	  | (Ast.EVENT_NAME(rnms,rstar,_) as r)::rs ->
	  (* test whether the required name and the defined name if any are the
	     same *)
	      if rnms = dnms && rstar = dstar
	      then loop rs skipped (Some r) ds
	      else
		(* the following test checks for the case where the * case
		   is not defined, but actually all of the subevents have been
		   handled *)
		if rstar &&
		  not(List.exists
			(function Ast.EVENT_NAME(rnms2,rstar2,_) ->
			  proper_prefix(rnms,rnms2))
			skipped) &&
		  (* check that the * thing is not the only instance of the
		     required name *)
		  (match prev with
		    None -> false
		  | Some (Ast.EVENT_NAME(rnms2,rstar2,_)) ->
		      proper_prefix(rnms,rnms2))
		then loop rs skipped (Some r) (d::ds)
		 (* in the final case, the required name is missing *)
		else loop rs (r::skipped) None (d::ds)) in
  loop
    (List.rev(List.sort compare required_names))
    [] None
    (List.rev(List.sort compare declared_names))
    
and proper_prefix = function
    ([],[]) -> false
  | ([],_) -> true
  | (_,[]) -> false
  | (x::xs,y::ys) -> B.ideq(x,y) && proper_prefix(xs,ys)
	
and skip nms = function
    [] -> []
  | (Ast.EVENT_NAME(rnms,rstar,_) as r)::rs ->
      if nms = rnms || proper_prefix(nms,rnms)
      then skip nms rs
      else r::rs
		
let type_handlers handlers var_env timers_with_targets
    persistent_timers star_timer_with_target =
  (*check_events handlers;*)
  in_handler := true;
  move_allowed := true;
  List.map
    (function
	(Ast.EVENT(nm,specd,(Ast.VARDECL(B.EVENT,id,imported,vattr) as param),
		   stmt,attr)) ->
	  in_persistent_timer_handler := false;
	  let snm = Ast.event_name2c nm in
	  let env =
	    match snm with
	      "bossa.schedule.*" -> var_env
	    | "bossa.schedule.block" -> var_env
	    | "preempt" -> var_env
	    | "unblock.timer.*" ->
		if star_timer_with_target
		then extend_env id (B.EVENT,true) var_env
		else var_env
	    | _ ->
		if Ast.event_name2c2 nm = Some "unblock.timer"
		then
		  begin
		    let var_env =
		      try
			let timer_id =
			  List.find
			    (function pt ->
			      snm = "unblock.timer."^(B.id2c pt))
			    persistent_timers in
			in_persistent_timer_handler := true;
			let (_,fldenv) = lookup_str "process" (!struct_env) in
			let ty = lookup_id timer_id fldenv in
			extend_env (B.mkId "local_data")
			  (B.NULL,true)
			  (extend_env (B.mkId "self")
			     (B.TIMER(B.PERSISTENT(None)),true)
			     (extend_env timer_id (ty,true(*const*)) var_env))
		      with Not_found -> var_env in
		    let twt_names =
		      List.map
			(function twt -> "unblock.timer."^(B.id2c twt))
			timers_with_targets in
		    if Aux.member snm twt_names
		    then extend_env id (B.EVENT,true) var_env
		    else var_env
		  end
		else
		  extend_env id (B.EVENT,true) var_env in
	  let (new_stmt,_) =
	    type_stmt stmt env None false B.BOTH false in
	  Ast.EVENT(nm, specd, param, new_stmt, attr)
      |	_ -> raise (Error.Error "internal error"))
    handlers
    
let check_params nm params attr =
  match (B.id2c nm,!B.schedtype) with
    ("attach",st) ->
      (match params with
	(Ast.VARDECL(ty,id,imported,attr)) :: rest ->
	  if not (ty = (B.sched_select B.PROCESS B.SCHEDULER))
	  then error(attr,"incorrect arguments to attach")
	  else
	    if st = B.VIRTSCHED
	    then
	      (List.iter
		 (function Ast.VARDECL(ty,id,imported,attr) ->
		   match ty with
		     B.INT -> ()
		   | B.ENUM(_)  -> ()
		   | B.RANGE(_)  -> ()
		   | B.BOOL  -> ()
		   | _ ->
		       error(attr,
			     pr "invalid parameter type %s" (B.type2c ty)))
		 rest;
	       if (List.length params) > 10
	       then error(attr,"at most 10 parameters to attach")
	       else ())
	    else ()
      |	_ -> error(attr,"incorrect arguments to attach"))
  | ("detach",B.PROCSCHED) ->
      (match params with
	[Ast.VARDECL(B.PROCESS,id,imported,attr)] -> ()
      |	_ -> error(attr,"incorrect arguments to detach"))
  | _ -> ()
	
let check_ret ret nm attr =
  match (B.id2c nm,!B.schedtype) with
    ("attach",B.PROCSCHED) ->
      if not(ret = B.VOID)
      then error(attr,"attach must have return type void");
      B.ENUM(B.mkId("SchedState"))
  | ("attach",B.VIRTSCHED) ->
      if not(ret = B.VOID)
      then error(attr,"attach must have return type void");
      ret
  | ("detach",B.PROCSCHED) ->
      if not(ret = B.VOID)
      then error(attr,"detach must have return type void");
      B.ENUM(B.mkId("SchedState"))
  | _ -> ret
	
	
let type_functions functions var_env proc_decls params_modifiable =
  in_handler := false;
  in_persistent_timer_handler := false;
  List.map
    (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) ->
      let snm = B.id2c nm in
      move_allowed := B.sched_select true (not(snm = "attach"));
      check_params nm params attr; (* only for attach/detach *)
      if snm = "attach" (* only for ifunctions *)
      then check_attach params proc_decls stmt;
      let ret = check_ret ret nm attr in (* only for attach/detach *)
      let (new_stmt,_) =
	type_stmt stmt
	  (List.fold_left
	     (function env ->
	       function Ast.VARDECL(ty,id,imported,attr) ->
		 extend_env id (ty,params_modifiable) env)
	     var_env params)
	  (Some(ret))
	  false B.BOTH false in
      Ast.FUNDEF(ret,nm,params,new_stmt,inl,attr))
    functions
    
let type_ifunctions functions var_env proc_decls =
  in_handler := false;
  in_persistent_timer_handler := false;
  let names =
    List.map
      (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) -> (B.id2c nm))
      functions in
  if (not (Aux.member "attach" names))
  then nl_error("Scheduler must define an attach function");
  if (not (Aux.member "detach" names))
  then nl_error("Scheduler must define a detach function");
  type_functions functions var_env proc_decls true
    
let type_aspects var_env proc_decls = function
    Ast.INITASPECTS(aspects) ->
      move_allowed := false;
      Ast.INITASPECTS(
      List.map
	(function
	    Ast.ASPECT(srcs,_,dsts,_,(Ast.VARDECL(ty,id,_,_) as x),
		       pos,body,fv,a) ->
	      let src_infos =
		Aux.union_list(Aux.flat_map CS.concretize srcs) in
	      let dst_infos =
		Aux.union_list(Aux.flat_map CS.concretize dsts) in
	      let new_env = extend_env id (ty,false) var_env in
	      let (new_stmt,_) =
		type_stmt body new_env None false B.BOTH false in
	      Ast.ASPECT(srcs,src_infos,dsts,dst_infos,x,pos,new_stmt,fv,a))
	aspects)
  | _ -> failwith "type: unexpected aspects"

(* Admit cannot have an event_typed parameter and cannot contain return
or error.  There are some restrictions on inlined functions that are checked
in inline.ml *)
let type_admit var_env proc_decls ifunctions = function
    None -> None
  | Some (Ast.ADMIT(vars,crit_params,crits,attach,detach,attr)) ->
      move_allowed := false;
      (* criteria parameters have to be a subset of the attach parameters
      and can contain no duplicates *)
      Ast.no_dup_params crit_params "admit criteria parameter";
      (try
	let Ast.FUNDEF(_,_,attach_params,_,_,_) =
	  List.find
	    (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) ->
	      B.id2c nm = "attach")
	  ifunctions in
	List.iter
	  (function Ast.VARDECL(crit_ty,crit_param,_,attr) ->
	    if not(List.exists
		     (function Ast.VARDECL(att_ty,att_param,_,_) ->
		       crit_ty = att_ty && B.ideq(crit_param,att_param))
		     attach_params)
	    then
	      error (attr,
		     (pr "parameter %s not found in attach params"
			(B.id2c crit_param))))
	  crit_params
      with Not_found -> error (attr, "missing attach function"));
      let params_modifiable = false in
      let mkenv params var_env =
	List.fold_left
	  (function env ->
	    function Ast.VARDECL(ty,id,imported,attr) ->
	      if ty = B.EVENT
	      then error (attr, "event parameter not allowed in admit");
	      extend_env id (ty,params_modifiable) env)
	  var_env params in
      let (var_env,new_vars) = type_defs vars var_env false B.BOTH in
      let (new_crits,_) =
	List.split
	  (List.map
	     (function crit ->
	       type_stmt crit (mkenv crit_params var_env) (Some B.BOOL)
		 false B.BOTH false)
	     crits) in
      let (new_attach,new_detach) =
	match (attach,detach) with
	  (None,None) as x -> x
	| (Some(att_prm,att),Some(det_prm,det)) ->
	    check_no_return att;
	    check_no_return det;
	    let (new_att,_) =
	      type_stmt att (mkenv [att_prm] var_env)
		None false B.BOTH false in
	    let (new_det,_) =
	      type_stmt det (mkenv [det_prm] var_env)
		None false B.BOTH false in
	    (Some(att_prm,new_att),Some(det_prm,new_det))
	| _ -> raise (Error.Error "type: unexpected admit") in
      Some (Ast.ADMIT(new_vars,crit_params,new_crits,new_attach,new_detach,
		      attr))
	  
(* - Add target, notarget, or persistent to unblock.timer handlers -- *)
	  
let rename_timer_handlers handlers timers_with_targets persistent_timers =
  List.map
    (function Ast.EVENT(nm,specd,param,stm,attr) ->
      let rename nm = Ast.EVENT(nm,specd,param,stm,attr) in
      match nm with
	Ast.EVENT_NAME([unblock;timer;something],star,attr) when
	  B.id2c unblock = "unblock" && B.id2c timer = "timer" ->
	    if List.mem something persistent_timers
	    then
	      rename
		(Ast.EVENT_NAME([unblock;timer;B.mkId "persistent";something],
				star,attr))
	    else if List.mem something timers_with_targets
	    then
	      rename
		(Ast.EVENT_NAME([unblock;timer;B.mkId "target";something],
				star,attr))
	    else
	      rename
		(Ast.EVENT_NAME([unblock;timer;B.mkId "notarget";something],
				star,attr))
      | _ -> rename nm)
    handlers
    
(* -------------------------- entry point --------------------------- *)
    
let type_check ast =
  Readwrite.readwrite ast; (* check that things are initd before read *)
  let ast = Rename.preprocess ast in
  let (Ast.SCHEDULER(nm,default,
		     cstdefs,enums,procdecls,fundecls,valdefs,states,
		     criteria,admit,trace,handlers,ifunctions,functions,
		       aspects,attr)) =
    Pre_type.preprocess ast in
  init_struct_env();
  init_fun_env();
  Ast.no_dup_valdefs (cstdefs@valdefs);
  let (cst_env,cstdefs1) = type_defs cstdefs [] false B.BOTH in
  let constenv = constant_values cstdefs in
  let (enum_env,enum_range_env) = typedef_env enums cst_env constenv in
  add_process_defn procdecls;
  add_fundecls fundecls;
  let state_env = type_states states enum_env in
  let (var_env,valdefs1) = type_defs valdefs state_env false B.BOTH in
  check_globals var_env;
  let criteria1 = verify_criteria criteria var_env enum_range_env in
  let trace1 = verify_trace var_env handlers trace in
  Events.init_event_env();
  let timers_with_targets =
    Aux.option_filter
      (function
	  Ast.VARDECL(B.TIMER(_),id,_,_) -> Some id
	| _ -> None)
      procdecls in
  let persistent_timers =
    Aux.option_filter
      (function
	  Ast.VARDECL(B.TIMER(B.PERSISTENT(_)),id,_,_) -> Some id
	| _ -> None)
      procdecls in
  let timers_without_targets =
    Aux.option_filter
      (function
	  Ast.SYSDEF(Ast.VARDECL(B.TIMER(_),id,_,_),_,_) -> Some id
	| _ -> None)
      valdefs in
  let handler_names =
    List.map
      (function Ast.EVENT(nm,_,_,_,_) -> nm)
      handlers in
  let star_timer_with_target =
    (* true if there is a handler for all of the timers declared as
       global variables *)
    Aux.set_equal
      (Aux.option_filter
	 (function
	     (* should reallycheck for unblock and timer *)
	     Ast.EVENT_NAME([_;_;x],false,_) ->
	       if Aux.member x timers_without_targets then Some x else None
	   | _ -> None)
	 handler_names)
      timers_without_targets in
  let handlers1 =
    type_handlers handlers var_env timers_with_targets
      persistent_timers star_timer_with_target in
  let handlers2 =
    rename_timer_handlers handlers1 timers_with_targets persistent_timers in
  let ifunctions1 = type_ifunctions ifunctions var_env procdecls in
  let functions1 = type_functions functions var_env procdecls false in
  let admit1 = type_admit var_env procdecls ifunctions1 admit in
  let aspects1 = type_aspects var_env procdecls aspects in
  let new_ast =
    Ast.SCHEDULER(nm,default,
		  cstdefs1,enums,procdecls,fundecls,valdefs1,states,criteria1,
		  admit1,
		  trace1,handlers2,ifunctions1,functions1,aspects1,attr) in
  Braces.braces(Admit.process(Inline.inline(Aspects.process new_ast)))
    
